package org.objectweb.eclipsespeedo.metadata;

import java.text.StringCharacterIterator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.IPath;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jface.dialogs.MessageDialog;
import org.objectweb.eclipsespeedo.EclipseSpeedoPlugin;
import org.objectweb.eclipsespeedo.IEclipseSpeedoConstants;

public class MetadataModelManager
{
	public static MetadataModelManager INSTANCE = new MetadataModelManager();

	protected void setPersistenceCapableSuperclass(
		EObject contexte,
		IType type,
		IType superType)
	{
		Resource resource = contexte.eResource();
		JDOClass jdoClass = getJDOClass(resource, type);

		if (jdoClass == null)
		{
			IFile metadataFile = findMetadataFile(type);
			if (metadataFile == null)
			{
				return;
			}
			jdoClass = getJDOClass(contexte, metadataFile, type);
		}

		if (jdoClass != null)
		{
			jdoClass.setPersistenceCapableSuperclass(
				superType.getFullyQualifiedName());
		}
	}

	protected void clearPersistenceCapcableSuperclass(
		EObject contexte,
		IType type)
	{
		Resource resource = contexte.eResource();
		JDOClass jdoClass = getJDOClass(resource, type);

		if (jdoClass == null)
		{
			IFile metadataFile = findMetadataFile(type);
			if (metadataFile == null)
			{
				return;
			}
			jdoClass = getJDOClass(contexte, metadataFile, type);
		}

		if (jdoClass != null)
		{
			jdoClass.setPersistenceCapableSuperclass(null);
		}
	}

	public IFile findMetadataFile(IType type)
	{
		IResource resource = type.getResource();
		if (resource == null)
		{
			return null;
		}
		IResource member = resource;
		Stack folderStack = new Stack();
		while (member != null)
		{
			folderStack.push(member);
			member = member.getParent();
		}

		IProject project = resource.getProject();
		IResource file = project.findMember(IEclipseSpeedoConstants.METADATA_FILE);
		if (file != null && file.exists() && hasJDOClass(file, type))
		{
			return (IFile) file;
		}

		IContainer parent = null;
		while (!folderStack.isEmpty())
		{
			member = (IResource) folderStack.pop();

			switch (member.getType())
			{
				case IResource.FOLDER :
				case IResource.PROJECT :
					IContainer container = (IContainer) member;
					file =
						container.findMember(
							IEclipseSpeedoConstants.PACKAGE_METADATA_FILE);
					if (file != null
						&& file.exists()
						&& hasJDOClass(file, type))
					{
						return (IFile) file;
					}
					if (parent != null)
					{
						String name = member.getName() + ".jdo";
						file = parent.findMember(name);
						if (file != null
							&& file.exists()
							&& hasJDOClass(file, type))
						{
							return (IFile) file;
						}
					}
					parent = container;
					break;
				case IResource.FILE :
					if (parent != null)
					{
						String elementName = member.getName();
						if (elementName.endsWith(".java"))
						{
							elementName =
								elementName.substring(
									0,
									elementName.lastIndexOf('.'));
						}
						String name = elementName + ".jdo";
						file = parent.findMember(name);
						if (file != null
							&& file.exists()
							&& hasJDOClass(file, type))
						{
							return (IFile) file;
						}
					}
					break;
			}
		}
		return null;
	}

	/**
	 * We got a jdo file, check in detail if this file contains the JDO class
	 * 
	 * @param resource
	 * @param type
	 * @return
	 */
	protected boolean hasJDOClass(IResource resource, IType type)
	{
		return hasJDOClass(getResource(resource, true), type);
	}

	public Resource getResource(IResource resource, boolean load)	{
		String path = "platform:/resource" + resource.getFullPath();
		ResourceSetImpl resourceSet = new ResourceSetImpl();
		return resourceSet.getResource(URI.createURI(path), load);
	}

	public Resource getResource(String fullPath, boolean load)	{
		String path = "platform:/resource" + fullPath;
		ResourceSetImpl resourceSet = new ResourceSetImpl();
		return resourceSet.getResource(URI.createURI(path), load);
	}

	/**
	 * Method getClassToEnhance. Get all class described in the metadata file
	 * 
	 * @return String
	 */
	public void collectAllClass(String filename, StringBuffer buffer)
	{
		Resource resource = getResource(filename, true);
		if (resource == null)
		{
			//
			// TODO
			// assert file no found
			return;
		}

		List contents = resource.getContents();
		if (contents.isEmpty())
		{
			//
			// TODO
			// assert empty file
			return;
		}

		JDOFile jdoFile = (JDOFile) contents.get(0);
		Iterator iterator = jdoFile.getPackages().iterator();
		while (iterator.hasNext())
		{
			JDOPackage jdOPackage = (JDOPackage) iterator.next();
			String packageName = jdOPackage.getName() + ".";

			Iterator classIterator = jdOPackage.getClasses().iterator();
			while (classIterator.hasNext())
			{
				JDOClass jdoClass = (JDOClass) classIterator.next();
				buffer.append(
					" "
						+ getFormatedString(packageName)
						+ jdoClass.getName()
						+ ".class");
			}
		}
	}

	private String getFormatedString(String s)
	{
		String fileSeparator = System.getProperty("file.separator");
		StringCharacterIterator ite = new StringCharacterIterator(s);
		String out = "";

		for (char c = ite.first();
			c != StringCharacterIterator.DONE;
			c = ite.next())
		{
			if (c == '.')
			{
				out = out + fileSeparator;
			}
			else
			{
				out = out + c;
			}
		}
		return out;
	}

	public IFile createMetadataFile(IJavaElement javaElement)
	{
		ResourceSetImpl resourceSet = new ResourceSetImpl();

		IFile file = null;
		IResource fileResource = javaElement.getResource();
		Resource resource = null;
		switch (javaElement.getElementType())
		{
			case IJavaElement.TYPE :
				{
					IType type = (IType) javaElement;

					String filePath =
						"platform:/resource"
							+ fileResource.getFullPath().removeFileExtension()
							+ ".jdo";

					resource = resourceSet.createResource(URI.createURI(filePath));
					IPackageFragment parent = type.getPackageFragment();

					JDOFile jdoFile = MetadataFactory.eINSTANCE.createJDOFile();
					resource.getContents().add(jdoFile);

					JDOPackage jdoPackage =	MetadataFactory.eINSTANCE.createJDOPackage();
					jdoFile.getPackages().add(jdoPackage);
					jdoPackage.setName(parent.getElementName());

					JDOClass jdoClass =
						MetadataFactory.eINSTANCE.createJDOClass();
					jdoPackage.getClasses().add(jdoClass);
					jdoClass.setName(type.getElementName());
				}

				break;
			case IJavaElement.PACKAGE_FRAGMENT :
				{
					IPackageFragment parent = (IPackageFragment) javaElement;
					IPath fileFullPath = fileResource.getFullPath();

					//
					// for JDO 1.0.1
					//

						//if (JDOProjectHelper.isJDO101Compliant(parent.getJavaProject().getProject())) {
					fileFullPath = fileFullPath.append(IEclipseSpeedoConstants.PACKAGE_METADATA_NAME);
						//}


					String filePath = "platform:/resource" + fileFullPath + ".jdo";
					resource =	resourceSet.createResource(URI.createURI(filePath));

					JDOFile jdoFile = MetadataFactory.eINSTANCE.createJDOFile();
					resource.getContents().add(jdoFile);

					JDOPackage jdoPackage =	MetadataFactory.eINSTANCE.createJDOPackage();
					jdoFile.getPackages().add(jdoPackage);
					jdoPackage.setName(parent.getElementName());
				}
				break;
			case IJavaElement.COMPILATION_UNIT :
				{
					ICompilationUnit type = (ICompilationUnit) javaElement;
					String filePath =
						"platform:/resource"
							+ fileResource.getFullPath().removeFileExtension()
							+ ".jdo";

					resource =
						resourceSet.createResource(URI.createURI(filePath));
					IPackageFragment parent =
						(IPackageFragment) type.getParent();

					JDOFile jdoFile = MetadataFactory.eINSTANCE.createJDOFile();
					resource.getContents().add(jdoFile);

					JDOPackage jdoPackage =
						MetadataFactory.eINSTANCE.createJDOPackage();
					jdoFile.getPackages().add(jdoPackage);
					jdoPackage.setName(parent.getElementName());

					JDOClass jdoClass =
						MetadataFactory.eINSTANCE.createJDOClass();
					jdoPackage.getClasses().add(jdoClass);
					String sclass = type.getElementName();
					int i = sclass.length() - 5;
					sclass = sclass.substring(0, i);

					jdoClass.setName(sclass);
				}
				break;
			case IJavaElement.PACKAGE_FRAGMENT_ROOT :
			default :
				return null;
		}

		try
		{
			HashMap options = new HashMap();
			options.put(XMLResource.OPTION_ENCODING, "UTF-8");
			resource.save(options);
			return getIResource(resource);
		}
		catch (Exception e)
		{
			MessageDialog.openError(
				EclipseSpeedoPlugin.getShell(),
				"Error",
				e.getMessage());
			e.printStackTrace();
		}
		return null;
	}

	public static IFile getIResource(Resource resource)
	{
		URI uri = resource.getURI();
		String path = uri.path();
		path = path.substring("/resource".length()); //$NON-NLS-1$
		IWorkspaceRoot workspaceRoot =
			EclipseSpeedoPlugin.getWorkspace().getRoot();
		return (IFile) workspaceRoot.findMember(path);
	}

	public boolean isPersistenceClass(JDOPackage jdoPackage, IType type)
	{
		if (jdoPackage.findClass(type.getElementName()) != null) {
			return true;
		}
		IResource resource = type.getResource();
		IContainer container = resource.getParent();
		Stack folderStack = new Stack();
		while (container != null)
		{
			folderStack.push(container);
			container = container.getParent();
		}

		IProject project = resource.getProject();
		IResource file = project.findMember(IEclipseSpeedoConstants.PACKAGE_METADATA_FILE);
		if (file != null && file.exists() && hasJDOClass(file, type)){
			return true;
		}

		IContainer parent = null;
		while (!folderStack.isEmpty())	{
			container = (IContainer) folderStack.pop();

			file =	container.findMember(IEclipseSpeedoConstants.PACKAGE_METADATA_FILE);
			if (file != null && file.exists() && hasJDOClass(file, type)){
				return true;
			}

			if (parent != null)	{
				String name = container.getName() + ".jdo";
				file = parent.findMember(name);
				if (file != null && file.exists() && hasJDOClass(file, type)){
					return true;
				}
			}
			parent = container;
		}
		return false;
	}

	/**
	 * We got a jdo file, check in detail if this file contains the JDO class
	 * 
	 * @param resource
	 * @param type
	 * @return
	 */
	protected boolean hasJDOClass(Resource resource, IType type){
		List contents = resource.getContents();
		if (!contents.isEmpty()) {
			JDOFile jdoFile = (JDOFile) contents.get(0);
			IPackageFragment packageFragment = type.getPackageFragment();
			JDOPackage jdoPackage =	jdoFile.findPackage(packageFragment.getElementName());
			if (jdoPackage == null)	{
				return false;
			}

			JDOClass jdoClass = jdoPackage.findClass(type.getElementName());
			if (jdoClass != null){
				return true;
			}
		}
		return false;
	}

	public JDOClass getJDOClass(EObject contexte, IFile file, IType type)
	{
		String path = "platform:/resource" + file.getFullPath();

		ResourceSet jdoResourceSet = contexte.eResource().getResourceSet();
		Resource jdoResource =
			jdoResourceSet.getResource(URI.createURI(path), true);
		jdoResource.setTrackingModification(
			contexte.eResource().isTrackingModification());

		return getJDOClass(jdoResource, type);
	}

	public JDOClass getJDOClass(Resource jdoResource, IType type)
	{
		List contents = jdoResource.getContents();
		if (!contents.isEmpty())
		{
			JDOFile jdoFile = (JDOFile) contents.get(0);
			IPackageFragment packageFragment = type.getPackageFragment();
			JDOPackage jdoPackage =
				jdoFile.findPackage(packageFragment.getElementName());
			if (jdoPackage == null)
			{
				return null;
			}

			JDOClass jdoClass = jdoPackage.findClass(type.getElementName());
			if (jdoClass != null)
			{
				return jdoClass;
			}
		}
		return null;
	}
}
