package org.objectweb.eclipsespeedo.metadata;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.objectweb.eclipsespeedo.EclipseSpeedoPlugin;
import org.objectweb.eclipsespeedo.IEclipseSpeedoConstants;
import org.objectweb.eclipsespeedo.jdoextensions.xml.JDOExtensionsConfiguration;
import org.objectweb.eclipsespeedo.metadata.initializers.InitializerRegistry;
import org.objectweb.eclipsespeedo.metadata.jdt.initializers.FieldInitializer;
import org.objectweb.eclipsespeedo.utils.JavaCodeHelper;
import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;

/**
 * This class handles the association between Java Element et JDO Metadata.
 * 
 * It is helper of MetadataEditor
 *  
 */
public class MetadataModelModifier
{
	/**
	 * The current package name
	 */
	private IPackageFragment javaPackage;

	/**
	 * The current class name
	 */
	private IType javaClass;

	/**
	 * the current field name
	 */
	private IField javaField;

	/**
	 * The original instance of the DOM Tree
	 */
	private Resource resource = null;

	/**
	 * Represent the current package node
	 */
	private JDOFile jdoFile = null;

	/**
	 * Represent the current package node
	 */
	private JDOPackage jdoPackage = null;

	/**
	 * Represent the current class node
	 */
	private JDOClass jdoClass = null;

	/**
	 * Represent the current attribute node
	 */
	private JDOField jdoField = null;

	private ArrayList listeners = new ArrayList();

	/**
	 * The constructor. Take as parameter, the metadata file.
	 * 
	 * @param metadata
	 *            The file that contains the JDO metadata.
	 */
	public MetadataModelModifier(Resource resource)	{
		this.resource = resource;

		List contents = resource.getContents();
		if (!contents.isEmpty()){
			jdoFile = (JDOFile) contents.get(0);
		}
	}

	public void cleanPackage() {
		if (jdoPackage == null)
			return;
		jdoFile.getPackages().remove(jdoPackage);
		jdoPackage = null;
	}
	
	public void cleanClass() {
		if (jdoClass == null)
			return;
		jdoPackage.getClasses().remove(jdoClass);
		jdoClass = null;
	}

	public void cleanField(){
		if (jdoField == null)
			return;
		jdoClass.getFields().remove(jdoField);
		jdoField = null;
	}

	protected void checkPackageNode()	{
		if (jdoPackage == null && javaPackage != null)	{
			String packageName = javaPackage.getElementName();
			jdoPackage = jdoFile.findPackage(packageName);
			if (jdoPackage == null)	{
				jdoPackage = MetadataFactory.eINSTANCE.createJDOPackage();
				jdoPackage.setName(packageName);
				jdoFile.getPackages().add(jdoPackage);
			}
		}
	}

	protected void checkClassNode()	{
		checkPackageNode();
		if (jdoClass == null && javaClass != null && jdoPackage != null){
			String className = javaClass.getElementName();
			jdoClass = jdoPackage.findClass(className);
			if (jdoClass == null){
				jdoClass = MetadataFactory.eINSTANCE.createJDOClass();
				jdoClass.setName(className);
				jdoPackage.getClasses().add(jdoClass);
			}
		}
	}

	protected void checkFieldNode() {
		checkClassNode();

		if (jdoField == null && javaField != null && jdoClass != null)	{
			String fieldName = javaField.getElementName();
			jdoField = jdoClass.findField(fieldName);
			if (jdoField == null) {
				jdoField = MetadataFactory.eINSTANCE.createJDOField();
				jdoField.setName(fieldName);
				jdoClass.getFields().add(jdoField);
			}
		}
	}

	/**
	 * Set the persistence of the current attribute
	 * 
	 * @param check
	 *            if true then is it checked and that means the attribute is
	 *            persistent if false the attribut is not persistent
	 */
	public void setFieldPersistence(boolean check)	{
		try	{
			// if check is true then we set the attribute as persistent
			// we make sure that all the currents node are setup for this
			// context
			if (check)	{
				checkFieldNode();
				jdoField.setPersistenceModifier(JDOPersistenceModifier.PERSISTENT_LITERAL);
				InitializerRegistry.initializerField(jdoField, javaField);
			}
			else {
				cleanField();
			}
		}
		catch (Exception e)	{
			e.printStackTrace();
		}
	}

	/** 
	 * Check if the field is an array, a collection aor a map and set the model
	 */
	public int getLevel(){
		IField field = javaField;
		IJavaProject javaProject = field.getJavaProject();
		IType type = (IType) field.getParent();
		try	{
			String signature = field.getTypeSignature();
			int count = Signature.getArrayCount(signature);
			// The field is an array if count>0
			if (count != 0)	{
				jdoField.setArray(MetadataFactory.eINSTANCE.createJDOArray());
				return JDOExtensionsConfiguration.FIELD;
			}
			String typeName = Signature.toString(signature);
			String [][] arrays = type.resolveType(typeName);
			if (arrays == null || arrays.length == 0){
				return JDOExtensionsConfiguration.FIELD;   // Type cannot be resolved
			}
			
			// ex: for object arays is {{java.lang},{Oblect}}
			String qualifierName = arrays[0][0] + "." + arrays[0][1];
						
			IType javaType = javaProject.findType(qualifierName);
			if (javaType != null){
				if (FieldInitializer.isSubTypeOf(javaType, "java.util.Collection"))	{
					return JDOExtensionsConfiguration.COLLECTION;
				}
				if (FieldInitializer.isSubTypeOf(javaType, "java.util.Map"))	{
					return JDOExtensionsConfiguration.MAP;
				}
			}
		}
		catch (JavaModelException e){return JDOExtensionsConfiguration.FIELD;}
		return JDOExtensionsConfiguration.FIELD;
	}




	final protected static void makeClassPersistent(
		JDOPackage jodPackage,
		IType javaClass,
		JDOClass jdoClass,
		IProgressMonitor monitor){
		/*
		 * if the superclass is persistent, set the
		 * "persistence-capable-superclass"
		 */
		IType superclass = JavaCodeHelper.getSuperclass(javaClass);
		if (superclass != null && MetadataModelManager.INSTANCE.isPersistenceClass(jodPackage, superclass))	{
			jdoClass.setPersistenceCapableSuperclass(superclass.getFullyQualifiedName());
		}

		/*
		 * If this class has some subclasses, set their attribute of
		 * "persistence-capable-superclass"
		 */

		//
		// TODO
		// Here the implementation is not performant since the JDT will find
		// all inheritence.
		// But we need just one level. In terms, wee should look for in the
		// MetadataResourceManager when it holds all metadata in memory
		//

		ITypeHierarchy hierarchy = null;

		try	{
			hierarchy = javaClass.newTypeHierarchy(monitor);
			IType[] subclasses = hierarchy.getSubclasses(javaClass);

			for (int i = 0; i < subclasses.length; i++)	{
				if (MetadataModelManager
					.INSTANCE
					.isPersistenceClass(jodPackage, subclasses[i]))	{
					MetadataModelManager
						.INSTANCE
						.setPersistenceCapableSuperclass(
						jodPackage,
						subclasses[i],
						javaClass);
				}
			}
		}
		catch (JavaModelException e)
		{
			e.printStackTrace();
		}

		InitializerRegistry.initializerClass(jdoClass, javaClass);
	}

	final protected static void makeClassTransient(
		JDOPackage jodPackage,
		IType javaClass,
		JDOClass jdoClass,
		IProgressMonitor monitor)
	{
		/*
		 * If this class has some subclasses, clear their attribute of
		 * "persistence-capable-superclass"
		 */

		//
		// TODO
		// Here the implementation is not performant since the JDT will find
		// all inheritence.
		// But we need just one level. In terms, wee should look for in the
		// MetadataResourceManager when it holds all metadata in memory
		//
		try	{
			ITypeHierarchy hierarchy = javaClass.newTypeHierarchy(null);
			IType[] subclasses = hierarchy.getSubclasses(javaClass);

			for (int i = 0; i < subclasses.length; i++)	{
				if (MetadataModelManager
					.INSTANCE
					.isPersistenceClass(jodPackage, subclasses[i]))
				{
					MetadataModelManager
						.INSTANCE
						.clearPersistenceCapcableSuperclass(
						jodPackage,
						subclasses[i]);
				}
			}
		}
		catch (JavaModelException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Set the persistence state of the current class
	 * 
	 * @param check
	 */
	public void setClassPersistence(boolean persistence)
	{
		try
		{
			// we make sure that all the currents node are setup for this
			// context
			checkClassNode();
			ProgressMonitorDialog dialog =	new ProgressMonitorDialog(EclipseSpeedoPlugin.getShell());
			try
			{
				dialog.open();
				if (persistence)
				{
					makeClassPersistent(
						jdoPackage,
						javaClass,
						jdoClass,
						dialog.getProgressMonitor());

					IField[] fields = javaClass.getFields();
					for (int i = 0; i < fields.length; i++)
					{
						IField field = (IField) fields[i];
						if (!Flags.isTransient(field.getFlags()))
						{
							String name = field.getElementName();
							JDOField _jdoField = jdoClass.findField(name);
							if (_jdoField == null)
							{
								_jdoField =
									MetadataFactory.eINSTANCE.createJDOField();
								_jdoField.setName(name);
								jdoClass.getFields().add(_jdoField);
								_jdoField.setPersistenceModifier(
									JDOPersistenceModifier.PERSISTENT_LITERAL);

								InitializerRegistry.initializerField(
									_jdoField,
									field);
							}
						}
					}
				}
				else
				{
					makeClassTransient(
						jdoPackage,
						javaClass,
						jdoClass,
						dialog.getProgressMonitor());
					jdoPackage.getClasses().remove(jdoClass);
					jdoClass = null;
				}
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
			finally
			{
				dialog.close();
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	public Vector getPackageExtension()
	{
		checkPackageNode();
		return new Vector(jdoPackage.getExtensions());
	}

	public Vector getClassExtension()
	{
		checkClassNode();

		return new Vector(jdoClass.getExtensions());
	}

	public Vector getFieldExtension()
	{
		checkFieldNode();

		return new Vector(jdoField.getExtensions());
	}

	public void setPackageExtension(Vector extension)
	{
		checkPackageNode();

		jdoPackage.getExtensions().clear();
		jdoPackage.getExtensions().addAll(extension);
	}

	public void setClassExtension(Vector extension)
	{
		checkClassNode();

		jdoClass.getExtensions().clear();
		jdoClass.getExtensions().addAll(extension);
	}

	public void setFieldExtension(Vector extension)
	{
		checkFieldNode();

		jdoField.getExtensions().clear();
		jdoField.getExtensions().addAll(extension);
	}

	final static protected void makeJavaElementPersistent(
		JDOPackage jdoPackage,
		IJavaElement element,
		IProgressMonitor monitor)
	{
		try
		{
			switch (element.getElementType())
			{
				case IJavaElement.CLASS_FILE :
					IClassFile classFile = (IClassFile) element;
					makeJavaElementPersistent(
						jdoPackage,
						classFile.getType(),
						monitor);
					break;
				case IJavaElement.TYPE :
					IType type = (IType) element;
					if (!type.isAnonymous())
					{
						String name = type.getElementName();
						JDOClass _jdoClass = jdoPackage.findClass(name);
						if (_jdoClass == null)
						{
							_jdoClass =
								MetadataFactory.eINSTANCE.createJDOClass();
							_jdoClass.setName(name);
							jdoPackage.getClasses().add(_jdoClass);
							makeClassPersistent(
								jdoPackage,
								type,
								_jdoClass,
								monitor);

							IField[] fields = type.getFields();
							for (int i = 0; i < fields.length; i++)
							{
								IField field = (IField) fields[i];
								if (!Flags.isTransient(field.getFlags()))
								{
									String fieldName = field.getElementName();
									JDOField _jdoField =
										_jdoClass.findField(fieldName);
									if (_jdoField == null)
									{
										_jdoField =
											MetadataFactory
												.eINSTANCE
												.createJDOField();
										_jdoField.setName(fieldName);
										_jdoClass.getFields().add(_jdoField);
										_jdoField.setPersistenceModifier(
											JDOPersistenceModifier
												.PERSISTENT_LITERAL);

										InitializerRegistry.initializerField(
											_jdoField,
											field);
									}
								}
							}

						}
					}
					break;
				case IJavaElement.COMPILATION_UNIT :
					ICompilationUnit unit = (ICompilationUnit) element;
					IType[] types = unit.getAllTypes();
					for (int i = 0; i < types.length; i++)
					{
						makeJavaElementPersistent(
							jdoPackage,
							types[i],
							monitor);
					}
			}
		}
		catch (JavaModelException e)
		{
			e.printStackTrace();
		}
	}

	final static protected void makeJavaElementTransient(
		JDOPackage jdoPackage,
		IJavaElement element,
		IProgressMonitor monitor)
	{
		try
		{
			switch (element.getElementType())
			{
				case IJavaElement.CLASS_FILE :
					IClassFile classFile = (IClassFile) element;
					makeJavaElementTransient(
						jdoPackage,
						classFile.getType(),
						monitor);
					break;
				case IJavaElement.TYPE :
					IType type = (IType) element;
					if (!type.isAnonymous())
					{
						String name = type.getElementName();
						JDOClass _jdoClass = jdoPackage.findClass(name);
						if (_jdoClass != null)
						{
							makeClassTransient(
								jdoPackage,
								type,
								_jdoClass,
								monitor);
							jdoPackage.getClasses().remove(_jdoClass);
						}
					}
					break;
				case IJavaElement.COMPILATION_UNIT :
					ICompilationUnit unit = (ICompilationUnit) element;
					IType[] types = unit.getAllTypes();
					for (int i = 0; i < types.length; i++)
					{
						makeJavaElementTransient(jdoPackage, types[i], monitor);
					}
			}
		}
		catch (JavaModelException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Set the persistence of the current package That means set/unset all
	 * classes in this packages interms of persistance
	 * 
	 * @param check
	 */
	public void setPackagePersistence(boolean check)
	{
		try
		{
			// we make sure that all the currents node are setup for this
			// context
			checkPackageNode();
			ProgressMonitorDialog dialog =
				new ProgressMonitorDialog(EclipseSpeedoPlugin.getShell());
			try
			{
				dialog.open();
				IProgressMonitor monitor = dialog.getProgressMonitor();
				if (check)
				{
					IJavaElement[] elements = javaPackage.getChildren();
					monitor.beginTask(
						"Set package content persistent",
						elements.length);
					for (int i = 0; i < elements.length; i++)
					{
						monitor.subTask(
							"Set "
								+ elements[i].getElementName()
								+ " persistent");
						makeJavaElementPersistent(
							jdoPackage,
							elements[i],
							monitor);
						monitor.worked(1);
					}

					InitializerRegistry.initializerPackage(
						jdoPackage,
						javaPackage);
				}
				else
				{
					IJavaElement[] elements = javaPackage.getChildren();
					monitor.beginTask(
						"Clean package content persistent",
						elements.length);
					for (int i = 0; i < elements.length; i++)
					{
						monitor.subTask(
							"Clean "
								+ elements[i].getElementName()
								+ " persistent");
						makeJavaElementTransient(
							jdoPackage,
							elements[i],
							monitor);
						monitor.worked(1);
					}

					jdoFile.getPackages().remove(jdoPackage);
					jdoPackage = null;
				}
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
			finally
			{
				dialog.close();
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * The persistent state of the current field
	 * 
	 * @return boolean
	 */
	public boolean isFieldPersistence()
	{
		boolean isClassPersistence = isClassPersistence();
		if (!isClassPersistence)
			return false;
		checkClassNode();

		if (jdoClass == null || javaField == null)
		{
			return false;
		}

		return jdoClass.findField(javaField.getElementName()) != null;
	}

	/**
	 * The persistent state of the current class
	 * 
	 * @return boolean
	 */
	public boolean isClassPersistence()
	{
		boolean isPackagePersistence = isPackagePersistence();
		if (!isPackagePersistence)
			return false;
		checkPackageNode();

		if (javaClass == null)
		{
			return false;
		}

		return jdoPackage.findClass(javaClass.getElementName()) != null;
	}

	/**
	 * the persistent state of the current package
	 * 
	 * @return boolean
	 */
	public boolean isPackagePersistence() {
		if (javaPackage == null) {
			return false;
		}
		JDOPackage findPackage = jdoFile.findPackage(javaPackage.getElementName());
		if (findPackage == null) {
			return false;
		}
		return !findPackage.isEmpty();
	}

	private void cleanEmptyAttribute(
		Element el,
		String attribute,
		String value)
	{
		if (value.equals(""))
		{
			if (el.hasAttribute(attribute))
			{
				el.removeAttribute(attribute);
			}
		}
	}

	public void changeFieldList(String listType)
	{
		checkFieldNode();

		if (listType != null && listType.length() != 0)
		{
			if (listType
				.equalsIgnoreCase(
					IEclipseSpeedoConstants.METADATA_DTD_ELEMENT_COLLECTION))
			{
				JDOCollection list =
					MetadataFactory.eINSTANCE.createJDOCollection();
				jdoField.setCollection(list);
			}

			//map selected
			else if (
				listType.equalsIgnoreCase(
					IEclipseSpeedoConstants.METADATA_DTD_ELEMENT_MAP))
			{
				JDOMap list = MetadataFactory.eINSTANCE.createJDOMap();
				jdoField.setMap(list);
			}

			//array selected
			else if (
				listType.equalsIgnoreCase(
					IEclipseSpeedoConstants.METADATA_DTD_ELEMENT_ARRAY))
			{
				JDOArray list = MetadataFactory.eINSTANCE.createJDOArray();
				jdoField.setArray(list);
			}
		}
		else
		{
			jdoField.setArray(null);
			jdoField.setMap(null);
			jdoField.setCollection(null);
		}
	}

	public static DocumentBuilder setJDOErrorHandle(DocumentBuilder db)
	{
		ErrorHandler errorHandler = new ErrorHandler()
		{
			public void warning(org.xml.sax.SAXParseException e)
			{
				EclipseSpeedoPlugin.getDefault().getLog().log(
					new Status(
						IStatus.WARNING,
						IEclipseSpeedoConstants.PLUGIN_ID,
						0,
						e.getMessage(),
						null));
			}
			public void error(org.xml.sax.SAXParseException e)
			{
				EclipseSpeedoPlugin.getDefault().getLog().log(
					new Status(
						IStatus.ERROR,
						IEclipseSpeedoConstants.PLUGIN_ID,
						0,
						e.getMessage(),
						null));
			}
			public void fatalError(org.xml.sax.SAXParseException e)
			{
				EclipseSpeedoPlugin.getDefault().getLog().log(
					new Status(
						IStatus.ERROR,
						IEclipseSpeedoConstants.PLUGIN_ID,
						0,
						e.getMessage(),
						null));
			}
		};
		db.setErrorHandler(errorHandler);
		return db;
	}

	/**
	 * When the user validate the metadata configuration we return the XML
	 * content of the orginal DOM instance
	 * 
	 * @return String
	 */
	public String getMetadataContent()
	{
		ByteArrayOutputStream buffer = new ByteArrayOutputStream();

		try
		{
			boolean modified = resource.isModified();
			resource.save(buffer, Collections.EMPTY_MAP);
			if (modified)
			{
				resource.setModified(modified);
			}
			return buffer.toString();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * @return String
	 */
	public String getJavaFieldName()
	{
		if (javaField == null)
			return "";
		return javaField.getElementName();
	}

	public IField getJavaField()
	{
		return this.javaField;
	}

	/**
	 * @return String
	 */
	public String getJavaClassName()
	{
		if (javaClass == null)
			return "";

		return javaClass.getTypeQualifiedName();
	}

	/**
	 * @return String
	 */
	public String getJavaPackageName()
	{
		if (javaPackage == null)
			return "";
		return javaPackage.getElementName();
	}

	/**
	 * Sets the currentClass.
	 * 
	 * @param currentClass
	 *            The currentClass to set
	 */
	public void setJavaClass(IType currentClass)
	{
		// if the class selected is the same we do nothing
		if (this.javaClass == currentClass)
		{
			javaField = null;
			return;
		}

		setJavaPackage(currentClass.getPackageFragment());

		// if the data is saved , we can change the current class
		//if (isDataValidate)
		this.javaClass = currentClass;

		jdoClass = null;
		jdoField = null;

		javaField = null;
	}

	/**
	 * Sets the currentPackage.
	 * 
	 * @param currentPackage
	 *            The currentPackage to set
	 */
	public void setJavaPackage(IPackageFragment currentPackage)
	{
		if (this.javaPackage == currentPackage)
		{
			return;
		}
		this.javaPackage = currentPackage;

		jdoPackage = null;
		jdoClass = null;
		jdoField = null;

		javaClass = null;
		javaField = null;
	}

	/**
	 * Sets the currentAttribute.
	 * 
	 * @param currentAttribute
	 *            The currentAttribute to set
	 */
	public void setJavaField(IField currentField)
	{
		if (this.javaField == currentField)
		{
			return;
		}

		IType type = (IType) currentField.getParent();

		setJavaClass(type);

		this.javaField = currentField;
		// the attribute has change, the attribute node is not valide anymore
		jdoField = null;
	}

	/**
	 * @return
	 */
	public IPackageFragment getJavaPackage()
	{
		return javaPackage;
	}

	/**
	 * @return
	 */
	public JDOClass getJdoClass()
	{
		checkClassNode();
		return jdoClass;
	}

	/**
	 * @return
	 */
	public JDOField getJdoField()
	{
		checkFieldNode();
		return jdoField;
	}

	/**
	 * @return
	 */
	public JDOPackage getJdoPackage()
	{
		checkPackageNode();
		return jdoPackage;
	}

	/**
	 * @return
	 */
	public Resource getResource()
	{
		return resource;
	}

}
