package org.objectweb.eclipsespeedo.utils;

import java.util.List;

import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.objectweb.eclipsespeedo.EclipseSpeedoPlugin;
import org.objectweb.eclipsespeedo.wizards.MainCreationWizardPage;

/**
 * This method is used in the MainCreation process.
 * It use the AST API to check and add dynamically java code.
 */
public class JavaCodeHelper {

	/**
	 * Return the MethodDeclaration (an ASTNode) with the name "methodName"
	 * This method don't care of the signature of the method
	 * It just return the first method or null if there isn't one with this name
	 * The aim is to find the "main" method
	 * @param methodArray a list a method
	 * @param methodName the name of the method
	 * @return
	 */
	public static MethodDeclaration getMethod(MethodDeclaration[] methodArray,	String methodName) {

		for (int j = 0; j < methodArray.length; j++) {
			MethodDeclaration tmp = (MethodDeclaration) methodArray[j];
			if (tmp.getName().toString().equals(methodName)){
				return tmp;
			}
		}
		// we return null otherwise
		return null;

	}

	/**
	 * Create a main method in a type (a class)
	 * @param type the class
	 * @return
	 */
	public static MethodDeclaration createMainMethod(TypeDeclaration type)
	{
		AST ast = type.getAST();
		MethodDeclaration mainMethod = ast.newMethodDeclaration();

		//The name
		mainMethod.setName(ast.newSimpleName("main"));

		// the signature
		// return type : void
		mainMethod.setReturnType(ast.newPrimitiveType(PrimitiveType.VOID));
		// modifiers :  publis static 
		mainMethod.setModifiers(Modifier.PUBLIC | Modifier.STATIC);

		//Parameter : String[] args
		List params = mainMethod.parameters();
		SingleVariableDeclaration var = ast.newSingleVariableDeclaration();
		var.setName(ast.newSimpleName("args"));
		var.setType(
			ast.newArrayType(
				ast.newSimpleType(ast.newName(new String[] { "String" }))));
		params.add(var);

		//then we create a body
		Block block = ast.newBlock();
		mainMethod.setBody(block);
		// we add it to the class
		type.bodyDeclarations().add(mainMethod);
		return mainMethod;

	}

	/**
	 * This method is specific to a jdo main class.  it takes a method which will be generally a
	 * main method. It also take a set of properties.
	 * Those properties configure the persistentManagerFactory , like password , main factory class ....
	 * This method take care of the existing java code. 
	 * If there isn't a Property it add it, if some properties exist they a replace, if the PersistentManagerFactory 
	 * or the Persistent aren't "created" in the code this method just do it. 
	 * @param method
	 * @param prop
	 * @param emptyBlock
	 * return a boolean indicating whether it is necessary or not to add io and utils imports
	 */
	public static boolean adjustPropertiesAndManagers(MethodDeclaration method, MainCreationWizardPage page){

		if (method == null) return false;
		Block block = method.getBody();
		// We get the list of statements
		List statements = block.statements();
		boolean res = false;
		
		// We use those String to check if some instances exist or not
		// if they do not exit they stay null
		String propertiesObjectName = null;
		String persistenceManagerFactoryObjectName = null;
		String persistenceManagerObjectName = null;
		String transactionName = null;
		int position = 0;   // position = la position de la declaration potentielle (Properties)
		
		for (int k = 0; k < statements.size(); k++)	{
			Statement statementTmp = (Statement) statements.get(k);
				if (statementTmp.getNodeType() == ASTNode.VARIABLE_DECLARATION_STATEMENT) {
				// get the instance declaration
				VariableDeclarationStatement varState =	(VariableDeclarationStatement) statementTmp;
					//We look for the type of the propertie object
				Type typeVar = varState.getType();
				try {
					
					if (typeVar.isSimpleType()){
						SimpleType simType = (SimpleType) typeVar;
						//	we look for a Properties Object
						if (simType.resolveBinding().getQualifiedName().equals("java.util.Properties"))	{
							List frag = varState.fragments();
							if (frag.size() > 0) {
								for (int i = 0; i < frag.size(); i++) {
									VariableDeclarationFragment varFrag = (VariableDeclarationFragment) frag.get(0);
									position = k+1;
									// we store the Properties variable identifier
									propertiesObjectName =	varFrag.getName().getIdentifier();
								}
							}
						}
						//	we look for a PersistenceManagerFactory
						if (simType.resolveBinding().getQualifiedName().equals("javax.jdo.PersistenceManagerFactory")){
							List frag = varState.fragments();
							if (frag.size() > 0){
								for (int i = 0; i < frag.size(); i++){
									VariableDeclarationFragment varFrag = (VariableDeclarationFragment) frag.get(0);
									persistenceManagerFactoryObjectName = varFrag.getName().getIdentifier();
								}
							}
						}
						// we look for a PersistenceManager
						if (simType.resolveBinding().getQualifiedName().equals("javax.jdo.PersistenceManager"))	{
							List frag = varState.fragments();
							if (frag.size() > 0){
								for (int i = 0; i < frag.size(); i++){
									VariableDeclarationFragment varFrag = (VariableDeclarationFragment) frag.get(0);
									persistenceManagerObjectName = varFrag.getName().getIdentifier();
								}
							}
						}
						// we look for a Transaction 
						if (simType.resolveBinding().getQualifiedName().equals("javax.jdo.Transaction")) {
							List frag = varState.fragments();
							if (frag.size() > 0) {
								for (int i = 0; i < frag.size(); i++) {
									VariableDeclarationFragment varFrag = (VariableDeclarationFragment) frag.get(0);
									transactionName = varFrag.getName().getIdentifier();
								}
							}
						}
					}	
				}
				catch (Exception e){
					UITools.selectionError("An Error occured during source code inspection:\n" +
											"   Impossible to resolve certain types in your java file",EclipseSpeedoPlugin.getDefault().getWorkbench().getActiveWorkbenchWindow());
					return false; 
				}
			}
		}

		//Here we only have in prop, properties that weren't in the code 
		// so we add those
		//
		AST ast = block.getAST();

		// if there isn't a properties object , we create it !!!
 
		if (!page.hasProperties()&&(propertiesObjectName==null)) {
			propertiesObjectName = "properties";
			res = true;
			statements.add(position++, createObject(ast, "Properties", "Properties", propertiesObjectName, null));
			statements.add(position++, makePropertieLoadInvocationStatement(ast, propertiesObjectName, "enhancer.properties"));
		}
		else if (!page.hasProperties()&&(propertiesObjectName!=null)) {
			res = true;
			statements.add(position++, makePropertieLoadInvocationStatement(ast, propertiesObjectName, "enhancer.properties"));
		}
		//Now that properties are created we create the PersistentManager familly
		if (persistenceManagerFactoryObjectName == null){
			persistenceManagerFactoryObjectName = "pmfactory";
			statements.add(position++, createFromStaticObject(ast, "PersistenceManagerFactory", persistenceManagerFactoryObjectName, "JDOHelper",	"getPersistenceManagerFactory",	new Expression[] { ast.newSimpleName(propertiesObjectName)}));
		}
		if (persistenceManagerObjectName == null)	{
			persistenceManagerObjectName = "pm";
			statements.add(position++, invokeAndGetResult(ast, "PersistenceManager", persistenceManagerObjectName, persistenceManagerFactoryObjectName,	"getPersistenceManager", null));
		}
		if (transactionName == null){
			transactionName = "tx";
			statements.add(position++,invokeAndGetResult(ast, "Transaction", transactionName,	persistenceManagerObjectName, "currentTransaction", null));
		}
		return res; 
	}

	public static IPackageFragment getJavaPackage(IJavaElement element) {
		while (element != null && (element.getElementType() != IJavaElement.PACKAGE_FRAGMENT)) {
			element = element.getParent();
		}
		if (element == null) {
			return null;
		}
		return (IPackageFragment) element;
	}

	/**
	 * Create the "propertiesObjectName.setProperties(key,value)" java code
	 * @param ast
	 * @param propertiesObjectName
	 * @param key
	 * @param value
	 * @return
	 */
	private static Statement makePropertieLoadInvocationStatement(AST ast, String propertiesObjectName,	String fileName) {
		
		TryStatement tryStmt = ast.newTryStatement();
		// a new method invocation	p.load(new FileInputStream("speedo.properties"));
		MethodInvocation mi = ast.newMethodInvocation();
		SimpleName propertiesObject = ast.newSimpleName(propertiesObjectName);
		mi.setExpression(propertiesObject);
		ClassInstanceCreation cic = ast.newClassInstanceCreation();
		cic.setName(ast.newSimpleName("FileInputStream"));
		StringLiteral filename = ast.newStringLiteral();
		filename.setLiteralValue("enhancer.properties");
		cic.arguments().add((Expression) filename);
		mi.setName(ast.newSimpleName("load"));
		mi.arguments().add(cic);
		
		// making a block from method invocation
		Block block = ast.newBlock();
		block.statements().add(ast.newExpressionStatement(mi));
		
		// Creating catch clauses
		CatchClause cc1 = ast.newCatchClause();
		SingleVariableDeclaration e1 = ast.newSingleVariableDeclaration();
		e1.setName(ast.newSimpleName("e1"));
		SimpleType exceptionType = ast.newSimpleType(ast.newSimpleName("FileNotFoundException"));
		e1.setType(exceptionType);
		cc1.setException(e1);
		Block bcc1 = ast.newBlock();
		MethodInvocation mi1 = ast.newMethodInvocation();
		mi1.setExpression(ast.newSimpleName("e1"));
		mi1.setName(ast.newSimpleName("printStackTrace"));
		bcc1.statements().add(ast.newExpressionStatement(mi1));
		cc1.setBody(bcc1);
		
		CatchClause cc2 = ast.newCatchClause();
		SingleVariableDeclaration e2 = ast.newSingleVariableDeclaration();
		e2.setName(ast.newSimpleName("e2"));
		exceptionType = ast.newSimpleType(ast.newSimpleName("IOException"));
		e2.setType(exceptionType);
		cc2.setException(e2);
		Block bcc2 = ast.newBlock();
		MethodInvocation mi2 = ast.newMethodInvocation();
		mi2.setExpression(ast.newSimpleName("e2"));
		mi2.setName(ast.newSimpleName("printStackTrace"));
		bcc2.statements().add(ast.newExpressionStatement(mi2));
		cc2.setBody(bcc2);
		
		tryStmt.catchClauses().add(cc1);
		tryStmt.catchClauses().add(cc2);
		tryStmt.setBody(block);
		return tryStmt;
	}

	/**
	 *  Create the "refObjectType resultVariableName = instanceInvokeObjectName.methodName(params)" java code
	 * @param ast
	 * @param refObjectType
	 * @param resultVariableName
	 * @param instanceInvokeObjectName
	 * @param methodName
	 * @param params
	 * @return
	 */
	private static Statement invokeAndGetResult(
		AST ast,
		String refObjectType,
		String resultVariableName,
		String instanceInvokeObjectName,
		String methodName,
		Expression[] params)
	{

		MethodInvocation methodInvoke =
			makeInvocation(ast, instanceInvokeObjectName, methodName, params);
		VariableDeclarationFragment varFrag =
			ast.newVariableDeclarationFragment();
		varFrag.setInitializer(methodInvoke);
		varFrag.setName(ast.newSimpleName(resultVariableName));
		VariableDeclarationStatement varState =
			ast.newVariableDeclarationStatement(varFrag);
		Type type =
			ast.newSimpleType(ast.newName(new String[] { refObjectType }));
		varState.setType(type);

		return varState;

	}

	/**
	 *  Create the "instanceObjectName.methodName(param)" java code
	 * @param ast
	 * @param instanceObjectName
	 * @param methodName
	 * @param params
	 * @return MethodInvocation
	 */
	private static MethodInvocation makeInvocation(
		AST ast,
		String instanceObjectName,
		String methodName,
		Expression[] params)
	{
		MethodInvocation staticInvoc = ast.newMethodInvocation();
		SimpleName factoryObject = ast.newSimpleName(instanceObjectName);
		staticInvoc.setExpression(factoryObject);
		staticInvoc.setName(ast.newSimpleName(methodName));
		if (params != null)
		{
			List args = staticInvoc.arguments();
			for (int i = 0; i < params.length; i++)
			{
				args.add(params[i]);
			}

		}

		return staticInvoc;

	}

	/**
	 * Create the java code : " refObjectType variableName = new qualObjectType(constructorParams)" 
	 * @param ast
	 * @param refObjectType
	 * @param qualObjectType
	 * @param variableName
	 * @param constructorParams
	 * @return
	 */
	private static Statement createObject(
		AST ast,
		String refObjectType,
		String qualObjectType,
		String variableName,
		Expression[] constructorParams)
	{
		//Variable declaration fragment 
		VariableDeclarationFragment varFrag =
			ast.newVariableDeclarationFragment();
		// The name of the variable
		SimpleName propertiesObject = ast.newSimpleName(variableName);
		varFrag.setName(propertiesObject);
		// the instance declaration : "new ..."
		ClassInstanceCreation instanceCreation = ast.newClassInstanceCreation();
		SimpleName snProperties = ast.newSimpleName(qualObjectType);
		//we set the qualification of the instanciation
		instanceCreation.setName(snProperties);
		// Then we setup the parameters
		if (constructorParams != null)
		{
			List args = instanceCreation.arguments();
			for (int i = 0; i < constructorParams.length; i++)
			{
				args.add(constructorParams[i]);
			}

		}

		// we set up the initializer ,wich is the rigth side of this statement
		varFrag.setInitializer(instanceCreation);
		VariableDeclarationStatement varState =
			ast.newVariableDeclarationStatement(varFrag);
		Type type =
			ast.newSimpleType(ast.newName(new String[] { refObjectType }));
		//We setup the "left side" wich is the reference type
		varState.setType(type);
		//varState.
		return varState;
	}

	/**
	 * Same as makeInvocation
	 * @param ast
	 * @param refObjectType
	 * @param variableName
	 * @param factoryObjectName
	 * @param staticMethodName
	 * @param params
	 * @return
	 */
	private static Statement createFromStaticObject(
		AST ast,
		String refObjectType,
		String variableName,
		String factoryObjectName,
		String staticMethodName,
		Expression[] params)
	{

		VariableDeclarationFragment varFrag =
			ast.newVariableDeclarationFragment();

		//the variable name
		SimpleName propertiesObject = ast.newSimpleName(variableName);
		varFrag.setName(propertiesObject);

		//the rigth side
		MethodInvocation staticInvoc =
			makeInvocation(ast, factoryObjectName, staticMethodName, params);
		// here the staticInvoc is an Expression
		varFrag.setInitializer(staticInvoc);
		VariableDeclarationStatement varState =
			ast.newVariableDeclarationStatement(varFrag);
		Type type =
			ast.newSimpleType(ast.newName(new String[] { refObjectType }));
		varState.setType(type);

		return varState;
	}

	public static IType getSuperclass(IType type)
	{
		try
		{
			String superclassName = type.getSuperclassName();
			if (superclassName == null)
				return null;
			String[][] resolveTypes = type.resolveType(superclassName);
			if (resolveTypes == null)
			{
				return null;
			}
			IJavaProject javaProject = type.getJavaProject();
			return javaProject.findType(resolveTypes[0][0], resolveTypes[0][1]);
		}
		catch (JavaModelException e)
		{
		}
		return null;
	}

}
