/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;

import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Loggable;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.speedo.mapper.rdb.JDBCMapperAttributes;
import org.objectweb.speedo.query.api.QueryManagerAttribute;
import org.objectweb.speedo.workingset.lib.SpeedoTransaction;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.mapper.api.JormFactoryAttributes;
import org.objectweb.speedo.api.TransactionListener;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.cache.api.CacheAttributeController;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.mapper.rdb.lib.ConnectionSpecJDBC;

import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOUserException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import java.util.Arrays;
import java.util.Properties;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * This class is a client helper which permits to create a new speedo
 * PersistentManagerFactory. Due to the JDO constraint the persistence manager
 * factory implementation must have a public empty contructor. But the use of
 * the fractal component (www.objectweb.org/fractal) needs to use a
 * fractal implementation to initialize components. Then this class is an
 * implementation of the PersistentManagerFactory interface which delegates all
 * calls on a persistent manager factory delegate. The constructor of this class
 * make the components creation and their initialization via Julia (fractal
 * implementation).
 *
 * @author S.Chassande-Barrioz
 */
public class Speedo implements PersistenceManagerFactory, ProxyManagerFactory {

    public final static String CONNECTION_SPEC_FACTORY =
        "org.objectweb.speedo.connectionSpecFactory";

    private final static String LOGGER_NAME = SpeedoProperties.LOGGER_NAME + ".init";
    private final static String SPEEDO_TEMPLATE = "org.objectweb.speedo.Speedo";
    private final static String OPTIMISTIC_CONCURRENCY_TEMPLATE =
		"org.objectweb.perseus.persistence.concurrency.POptimisticConcurrencyManager";
    private final static String PESSIMISTIC_CONCURRENCY_TEMPLATE =
		"org.objectweb.perseus.persistence.concurrency.PPessimisticConcurrencyManager";
    private final static String PM_POOL_PATH = "proxy-manager-pool";
    private final static String MEMORY_CACHE_PATH = "tpm.cache-manager";
    public final static String QUERY_CACHE_PATH = "compiled-query-cache";
    public final static String QUERY_MANAGER = "query-manager";
    public final static String CONNECTION_POOL_PATH = "mapper.pool";
    public final static String MONOLOG_FACTORY_PATH = "monolog-factory";
    public final static String PRIMITIVE_MAPPER_PATH = "mapper.mapper";
    public final static String JORM_FACTORY_PATH = "mapper.jorm-factory";
    public final static String PMF_PATH = "proxy-manager-factory";


    private static Map INSTANCES = new HashMap();
    private final static String FRACTAL_PROVIDER = "fractal.provider";
    private final static String DEFAULT_FRACTAL_PROVIDER
            = "org.objectweb.fractal.julia.Julia";

    private final static String JULIA_LOADER = "julia.loader";
    private final static String DEFAULT_JULIA_LOADER
            = "org.objectweb.fractal.julia.loader.DynamicLoader";

    private final static String JULIA_CONFIG = "julia.config";
    private final static String DEFAULT_JULIA_CONFIG = "julia.cfg";
    
    private static Factory factory = null;
    private static Component speedoTemplate = null;

    // method called by the JDOHelper.getPersistenceManagerFactory method
    public static PersistenceManagerFactory getPersistenceManagerFactory(
            Properties props) throws Throwable {
        Object instance;
        synchronized (INSTANCES) {
            instance = INSTANCES.get(props);
            if (instance == null) {
                instance = new Speedo(props);
                INSTANCES.put(props, instance);
            }
        }
        return (PersistenceManagerFactory) instance;
    }

    public static void clearInstances() {
        INSTANCES.clear();
    }

    /**
     * The persistence manager factory delegate
     */
    private PersistenceManagerFactory delegate = null;
	ProxyManagerFactory pmfComponent = null;
	private Logger logger = null;
	private Component speedo = null;
	boolean isPropertiesInitialized = false;

	public Speedo() throws Throwable {
		getSpeedoComponent();
		isPropertiesInitialized = false;
	}
    /**
     * It creates and initializes a real SpeedoProxyManagerFactory with
     * Julia (Fractal implementation).
     */
    public Speedo(Properties props) throws Throwable {
		try {
			init(getSpeedoComponent(), props);
			isPropertiesInitialized = true;
		} catch(Throwable e) {
			System.err.println("Error during the instanciation of the Speedo " +
					"persistence manager factory:");
			e.printStackTrace(System.err);
			throw e;
		}
	}

	public ProxyManagerFactory getPMFComponent() throws Exception {
		return pmfComponent;
	}

  private Factory getADLFactory () throws Exception {
    if (factory == null) {
      factory = FactoryFactory.getFactory(FactoryFactory.FRACTAL_BACKEND);
    }
    return factory;
  }
  
	private Component getSpeedoComponent() throws Throwable {
		
		if (speedo == null) { 
			//use Julia as Fractal implementation
			System.setProperty(FRACTAL_PROVIDER, DEFAULT_FRACTAL_PROVIDER);
			System.setProperty(JULIA_LOADER, DEFAULT_JULIA_LOADER);
			System.setProperty(JULIA_CONFIG, DEFAULT_JULIA_CONFIG);
	
			//Allocate a Logger through the LoggerFactory component

			Monolog.initialize();

			LoggerFactory loggerFactory = Monolog.monologFactory;
			logger = loggerFactory.getLogger(LOGGER_NAME);
	
			//instanciate the Speedo component
			try {
				if (speedoTemplate == null) {
          Map ctxt = new HashMap();
          ctxt.put("template", "true");
					speedoTemplate = (Component)getADLFactory().newComponent(
              SPEEDO_TEMPLATE, ctxt);
				}
        //speedo = (Component)getADLFactory().newComponent(SPEEDO_TEMPLATE, null);
				speedo = Fractal.getFactory(speedoTemplate).newFcInstance();
			} catch (Throwable e) {
				System.err.println("Error during the template loading or the " +
						"component instanciation '" + SPEEDO_TEMPLATE + "': ");
				throw e;
			}
			Fractal.getNameController(speedo).setFcName("org.objectweb.speedo");
			delegate = (PersistenceManagerFactory) speedo.getFcInterface("persistence-manager-factory");	
		}
		return speedo;
	}
	
	public void init(Component speedo, Properties pmfProps) throws Throwable {
				

		//Manage properties of the PersistenceManagerFactory
		//--------------------------------------------------
		Properties props = new Properties(pmfProps);
		String strval;

		props.remove(SpeedoProperties.JDO_PERSISTENCE_MANAGER_FACTORY_CLASS);

		//Debug Mode
		strval = props.getProperty(SpeedoProperties.DEBUG, null);
		props.remove(SpeedoProperties.DEBUG);
		if (strval != null) {
			Debug.ON = Boolean.valueOf(strval).booleanValue();
			logger.log(BasicLevel.INFO, SpeedoProperties.DEBUG + ": " + Debug.ON);
		}

		configurePool(getSubComponent(speedo, PM_POOL_PATH),
				SpeedoProperties.PM_POOL_MIN, "PersistenceManager pool: min size: ",
				SpeedoProperties.PM_POOL_MAX, "PersistenceManager pool: max size: ",
				SpeedoProperties.PM_POOL_TTL, "PersistenceManager pool: ttl: ",
				SpeedoProperties.PM_POOL_TIMEOUT, "PersistenceManager pool: timeout: ",
				props);
		configureMemoryCache(speedo, props);
		configureQueryCache(speedo, props);

		boolean useConnectionFactory = configureMapper(speedo, props);

		//deprecated properties
		final String TM_NAME_old = "org.objectweb.perseus.connector.ra.jdo.TMName";
		strval = props.getProperty(TM_NAME_old, "");
		if (strval.length()>0) {
			logger.log(BasicLevel.WARN, "Property " + TM_NAME_old
					+ " is deprecated, you must use " + SpeedoProperties.TM_NAME);
			props.remove(TM_NAME_old);
			pmfProps.remove(TM_NAME_old);
			if (props.getProperty(SpeedoProperties.TM_NAME) == null) {
				props.setProperty(SpeedoProperties.TM_NAME, strval);
				pmfProps.setProperty(SpeedoProperties.TM_NAME, strval);
			}
		}

		// Transaction managed by an application server
		strval = props.getProperty(SpeedoProperties.MANAGED, "");
		if (strval.length()>0) {
			boolean b = isTrue(strval);
			if (!useConnectionFactory && b) {
				String msg = "In a managed environnement, a connection factory is required";
				logger.log(BasicLevel.ERROR, msg);
				throw new JDOUserException(msg);
			}
			logger.log(BasicLevel.INFO, SpeedoProperties.MANAGED + ": Speedo is used "
					+ (b
					? "in a managed environnement"
					: "in stand alone"));

			//TransactionManager JNDI name
			strval = props.getProperty(SpeedoProperties.TM_NAME, "");
			props.remove(SpeedoProperties.TM_NAME);
			logger.log(BasicLevel.INFO, SpeedoProperties.TM_NAME + ": "
					+ (strval.length()>0
					? strval
					:  ": no value specified, use a well known application server"));
		} else {
			props.remove(SpeedoProperties.TM_NAME);
		}

		//Listener of Transaction life cycle (statistic)
		strval = props.getProperty(SpeedoProperties.TX_LISTENER, "");
		if (strval.length()>0) {
			try {
				SpeedoTransaction.txListener = (TransactionListener)
						Class.forName(strval).newInstance();
				if (SpeedoTransaction.txListener instanceof Loggable) {
					((Loggable) SpeedoTransaction.txListener)
							.setLoggerFactory(Monolog.monologFactory);
				}
				logger.log(BasicLevel.INFO, "Transaction listener: " + strval);
			} catch (Exception e) {
				logger.log(BasicLevel.WARN,
						"Impossible to instanciate the TransactionListener specified '"
						+ strval + "': ", e);
			}
		}

		configurePrefetching(speedo, props);
		configureConcurrencyManager(speedo, props);

		//managed other properties
        for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
            Map.Entry me = (Map.Entry) it.next();
            String key = (String) me.getKey();
            String value = (String) me.getValue();
			boolean knownProperty = true;
            if (SpeedoProperties.JDO_OPTION_RETAIN_VALUES.equals(key)) {
                delegate.setRetainValues(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_RESTORE_VALUES.equals(key)) {
                delegate.setRestoreValues(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_IGNORE_CACHE.equals(key)) {
                delegate.setIgnoreCache(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_READ.equals(key)) {
                delegate.setNontransactionalRead(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_NON_TRANSACTIONAL_WRITE.equals(key)) {
                delegate.setNontransactionalWrite(Boolean.valueOf(value).booleanValue());

            } else if (SpeedoProperties.JDO_OPTION_MULTITREADED.equals(key)) {
                delegate.setMultithreaded(Boolean.valueOf(value).booleanValue());

            } else {
				knownProperty = false;
            }
			if (knownProperty) {
				logger.log(BasicLevel.INFO, key + ": " + value);
				it.remove();
			}
        }

		//Warn the unknwon properties
		for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			String key = (String) me.getKey();
			String value = (String) me.getValue();
			logger.log(BasicLevel.WARN, "The (" + key + ", " + value
				+ ") property is not managed");
		}

		delegate.getProperties().putAll(pmfProps);
        Fractal.getLifeCycleController(speedo).startFc();
		pmfComponent = (ProxyManagerFactory) getSubComponent(
				((Interface) delegate).getFcItfOwner(), PMF_PATH)
				.getFcInterface("proxy-manager-factory");
    }

    private Component getSubComponent(Component parent,
                                      String path) throws Exception {
		if (logger != null) {
			logger.log(BasicLevel.DEBUG, "path: " + path);
		}
        StringTokenizer st = new StringTokenizer(path, ".", false);
        Component res = parent;
        while(st.hasMoreTokens()) {
            String commponenentname = st.nextToken();
            Component[] children = Fractal.getContentController(res)
                    .getFcSubComponents();
            int i = 0;
			if (logger != null) {
				logger.log(BasicLevel.DEBUG, "search: " + commponenentname);
			}
            while(i<children.length &&
                    !Fractal.getNameController(children[i])
                        .getFcName().equals(commponenentname)) {
				if (logger != null) {
                	logger.log(BasicLevel.DEBUG, "current: " + ((NameController) children[i].getFcInterface("name-controller")).getFcName());
				}
				i++;
            }
            if (i<children.length) {
				if (logger != null) {
					logger.log(BasicLevel.DEBUG, "found: " + ((NameController) children[i].getFcInterface("name-controller")).getFcName());
				}
                res = children[i];
            } else {
				if (logger != null) {
					logger.log(BasicLevel.DEBUG, "not found");
				}
                return null;
            }
        }
        return res;
    }

    private void configurePool(Component pool,
							   String minProp, String minMsg,
							   String maxProp, String maxMsg,
							   String ttlProp, String ttlMsg,
							   String timeoutProp, String timeoutMsg,
							   Properties props) throws Exception {
		String strval = null;
		int poolMinSize = -2;
		int poolMaxSize = -2;
		long poolTTL = -2;
		long poolTimeout = -2;
		boolean conf = false;
		if (maxProp != null) {
			conf = true;
			strval = props.getProperty(maxProp, "");
			if (strval.length() > 0) {
				props.remove(maxProp);
				if (strval.equalsIgnoreCase("nolimit")) {
//TODO: use a constant for an unlimited pool max size
					poolMaxSize = -1;
				} else {
					poolMaxSize = Integer.parseInt(strval);
				}
				logger.log(BasicLevel.INFO, maxMsg + strval);
			}
		}
		if (minProp != null) {
			conf = true;
			strval = props.getProperty(minProp, "");
			if (strval.length() > 0) {
				props.remove(minProp);
				poolMinSize = Integer.parseInt(strval);
				logger.log(BasicLevel.INFO, minMsg + strval);
			}
		}
		if (ttlProp != null) {
			conf = true;
			strval = props.getProperty(ttlProp, "");
			if (strval.length() > 0) {
				props.remove(ttlProp);
				try {
					poolTTL = Long.parseLong(strval);
				} catch (NumberFormatException e) {
					poolTTL = -1;
					strval = "NOTTL";
				}
				logger.log(BasicLevel.INFO, ttlMsg + strval);
			}
		}
		if (timeoutProp != null) {
			conf = true;
			strval = props.getProperty(timeoutProp, "");
			if (strval.length() > 0) {
				props.remove(timeoutProp);
				try {
					poolTimeout = Long.parseLong(strval);
				} catch (NumberFormatException e) {
					poolTimeout = -1;
					strval = "WAIT";
				}
				logger.log(BasicLevel.INFO, timeoutMsg + strval);
			}
		}
		if (conf) {
      PoolAttributes poolAttr;
      try {
        poolAttr = (PoolAttributes)
					Fractal.getAttributeController(pool);
      } catch (Exception e) {
        Component[] children = Fractal.getContentController(speedo).getFcSubComponents();
        String[] strs= new String[children.length];
        for (int i = 0; i < strs.length; ++i) {
          strs[i] = Fractal.getNameController(children[i]).getFcName();
        }
        throw new RuntimeException("TOTO " + Arrays.asList(strs));
      }
			if (poolMaxSize > -2) {
				poolAttr.setMaxSize(poolMaxSize);
			}
			if (poolMinSize > -2) {
				poolAttr.setMinSize(poolMinSize);
			}
			if (poolTTL > -2) {
				poolAttr.setTTL(poolTTL);
			}
			if (poolTimeout > -2) {
				poolAttr.setTimeout(poolTimeout);
			}
		}
	}

	/**
	 * It configures a cache manager.
	 * @param composite_cmC the composite component containing the primitive
	 * cache manager and the primitive replacement manager.
	 * @param maxSize is the maximal cache size.
	 * (-1 => nolimit, -2 => no config)
	 * @param maxMemSize is the maximal memory size.
	 * (-1 => nolimit, -2 => no config)
	 * @param autoCleanSize is the string describing the auto clean replacement
	 * size (absolute value: "150", relative value: "7%", no config: null)
	 * @param autoCleanThreshold is the string describing the threshold since
	 * the cache must try to free resource wrt the autocleanSize
	 * (absolute value: "150", relative value: "7%", no config: null)
	 * @param policy is the replacement policy
	 * (null => default, "LRU", "MRU", "FIFO")
	 * @throws Exception if it is noo possible to configure the specified cache
	 * manager.
	 */
	private void configureCacheManager(Component composite_cmC,
									   int maxSize,
									   int maxMemSize,
	                                   String autoCleanSize,
									   String autoCleanThreshold,
									   String policy) throws Exception {
		Component cmC = getSubComponent(composite_cmC, "cache-manager");
		CacheAttributeController cacheAttr = (CacheAttributeController)
			Fractal.getAttributeController(cmC);
		if (maxSize > -2) {
			cacheAttr.setMaxObjects(maxSize);
		}
		if (maxMemSize > -2) {
			cacheAttr.setMemorySize(maxMemSize);
		}
		if (autoCleanSize != null) {
			cacheAttr.setAutoCleanSize(autoCleanSize);
		}
		if (autoCleanThreshold != null) {
			cacheAttr.setAutoCleanThreshold(autoCleanThreshold);
		}
		if (policy != null) {
			String tempName = null;
			if (SpeedoProperties.CACHE_REPLCAEMENT_LRU.equalsIgnoreCase(policy)) {
			} else if (SpeedoProperties.CACHE_REPLCAEMENT_MRU.equalsIgnoreCase(policy)) {
				tempName = "org.objectweb.perseus.cache.replacement.lib.MRUReplacementManager";
			} else if (SpeedoProperties.CACHE_REPLCAEMENT_FIFO.equalsIgnoreCase(policy)) {
				tempName = "org.objectweb.perseus.cache.replacement.lib.FIFOReplacementManager";
			} else {
				logger.log(BasicLevel.ERROR, new SpeedoException(
						"Unmanaged cache replacement policy: " + policy));
			}
			if (tempName != null) {
				//unbind the rm from the composite
				Fractal.getBindingController(composite_cmC).unbindFc("replacement-manager");
				//unbind the rm from the primitive cm
				Fractal.getBindingController(cmC).unbindFc("replacement-manager");
				//unbind the primitive cm from the old rm
				Fractal.getBindingController(
						getSubComponent(composite_cmC, "replacement-manager"))
						.unbindFc("unbind-manager");

				//remove the old rm from the composite
				Fractal.getContentController(composite_cmC).removeFcSubComponent(
						getSubComponent(composite_cmC, "replacement-manager"));

				//instanciate the new component
				Component new_rm = (Component)getADLFactory().newComponent(tempName, null);
				Fractal.getNameController(new_rm).setFcName("replacement-manager");

				//Add the new rm in the composite
				Fractal.getContentController(composite_cmC).addFcSubComponent(new_rm);

				Object rm = new_rm.getFcInterface("replacement-manager");
				//bind the rm to the composite
				Fractal.getBindingController(composite_cmC).bindFc("replacement-manager", rm);
				//bind the rm to the primitive cm
				Fractal.getBindingController(cmC).bindFc("replacement-manager", rm);
				//bind the primitive cm to the new rm
				Fractal.getBindingController(new_rm).bindFc("unbind-manager",
						cmC.getFcInterface("unbind-manager"));
			}
		}
	}

	/**
	 * Configures the compiled query cache.
	 * @param speedo is the speedo composite component
	 * @param props is the speedo initialisation properties
	 */
	private void configureQueryCache(Component speedo, Properties props)
			throws Exception {
		int cacheSize = -2;
		String autoCleanSize = null;
		String autoCleanThreshold = null;
		String policy = null;
		Component cm  = getSubComponent(speedo, QUERY_CACHE_PATH);

		//configure the pool of compiled query
		String strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_CACHE_SIZE, "");
		if (strval.length()>0) {
			props.remove(SpeedoProperties.COMPILED_QUERY_CACHE_SIZE);
			if (strval.equalsIgnoreCase("nolimit")) {
				cacheSize = CacheAttributeController.NO_LIMIT;
			} else {
				cacheSize = Integer.parseInt(strval);
			}
			logger.log(BasicLevel.INFO, "Compiled query cache: max cache size: " + strval);
		}
		//configure the cache auto clean size
		strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_SIZE, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_SIZE);
			autoCleanSize = strval;
			logger.log(BasicLevel.INFO, "Compiled query cache: auto replacement size: " + strval);
		}

		//configure the cache auto clean size
		strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_THRESHOLD, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.COMPILED_QUERY_CACHE_AUTO_CLEAN_THRESHOLD);
			autoCleanThreshold = strval;
			logger.log(BasicLevel.INFO, "Compiled query cache: auto replacement threshold: " + strval);
		}

		//configure the replacement policy of the cache (Replacement manager)
		strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_CACHE_POLICY, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.COMPILED_QUERY_CACHE_POLICY);
			policy = strval;
			logger.log(BasicLevel.INFO, "Compiled query cache: replacement policy: " + strval);
		}

		configureCacheManager(cm, cacheSize, -2, autoCleanSize,
		        autoCleanThreshold, policy);
	}

	/**
	 * Configures the memory cache.
	 * @param speedo is the speedo composite component
	 * @param props is the speedo initialisation properties
	 */
	private void configureMemoryCache(Component speedo, Properties props)
			throws Exception {
		int cacheSize = -2;
		String autoCleanSize = null;
		String autoCleanThreshold = null;
		String policy = null;
		//configure the cache size
		String strval = props.getProperty(SpeedoProperties.CACHE_SIZE, "");
		Component cm  = getSubComponent(speedo, MEMORY_CACHE_PATH);
		if (strval.length()>0) {
			props.remove(SpeedoProperties.CACHE_SIZE);
			if (strval.equalsIgnoreCase("nolimit")) {
				cacheSize = CacheAttributeController.NO_LIMIT;
			} else {
				cacheSize = Integer.parseInt(strval);
			}
			logger.log(BasicLevel.INFO, "Persistent objects cache: max size: " + strval);
		}

		//configure the cache auto clean size
		strval = props.getProperty(SpeedoProperties.CACHE_AUTO_CLEAN_SIZE, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.CACHE_AUTO_CLEAN_SIZE);
			autoCleanSize = strval;
			logger.log(BasicLevel.INFO, "Persistent objects cache: auto replacement size: " + strval);
		}

		//configure the cache auto clean size
		strval = props.getProperty(SpeedoProperties.CACHE_AUTO_CLEAN_THRESHOLD, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.CACHE_AUTO_CLEAN_THRESHOLD);
			autoCleanThreshold = strval;
			logger.log(BasicLevel.INFO, "Persistent objects cache: auto replacement threshold: " + strval);
		}

		//configure the replacement policy of the cache (Replacement manager)
		strval = props.getProperty(SpeedoProperties.CACHE_REPLCAEMENT, "");
		if (strval.length() > 0) {
			props.remove(SpeedoProperties.CACHE_REPLCAEMENT);
			policy = strval;
			logger.log(BasicLevel.INFO, "Persistent objects cache: replacement policy: " + strval);
		}
		configureCacheManager(cm, cacheSize, -2, autoCleanSize,
		        autoCleanThreshold, policy);
	}


	private void configureConcurrencyManager(Component speedo,
											 Properties props)
			throws Exception {
		String strval = props.getProperty(SpeedoProperties.JDO_OPTION_OPTIMISTIC, "false");
		boolean b = isTrue(strval);
		logger.log(BasicLevel.INFO, "Transaction mode: "
				+ (b ? "OPTIMISTIC" : "PESSIMISTIC"));
		if (!b) {
			return;
		}
		String tempName = (b
				? OPTIMISTIC_CONCURRENCY_TEMPLATE
				: PESSIMISTIC_CONCURRENCY_TEMPLATE);
		//fetch the composite TPM
		Component composite_tpm = getSubComponent(
				speedo, "tpm");
		ContentController tpmCC = Fractal.getContentController(composite_tpm);
		//fetch the primitive concurrency manager
		BindingController tpmBC = Fractal.getBindingController(
				getSubComponent(composite_tpm, "transactional-persistence-manager"));
		tpmBC.unbindFc("concurrency-manager");

		//Remove the old concurrency manager
		Component cm = getSubComponent(composite_tpm, "concurrency-manager");
		BindingController cmBC = Fractal.getBindingController(cm);
		cmBC.unbindFc("state-manager");
		cmBC.unbindFc("storage-manager");
		cmBC.unbindFc("dependency-graph");
		tpmCC.removeFcSubComponent(cm);


		//instanciate the new ConcurrencyManager
		cm = (Component)getADLFactory().newComponent(tempName, null);
		//Add the new concurrency manager into the composite
		tpmCC.addFcSubComponent(cm);
		//bind the server interface of the new concurrency manager
		cmBC = Fractal.getBindingController(cm);
		cmBC.bindFc("state-manager",
				tpmCC.getFcInternalInterface("state-manager"));
		cmBC.bindFc("storage-manager",
				tpmCC.getFcInternalInterface("storage-manager"));
		cmBC.bindFc("dependency-graph",
				getSubComponent(composite_tpm, "dependency-graph").getFcInterface("dependency-graph"));

		//bind the client interface of the new ConcurrencyManager
		tpmBC.bindFc("concurrency-manager", cm.getFcInterface("concurrency-manager"));
	}

	private void configurePrefetching(Component speedo,
									  Properties props) throws Throwable {
		//Data prefetching on queries
		Component qm = getSubComponent(speedo, QUERY_MANAGER);
		QueryManagerAttribute qma = ((QueryManagerAttribute) Fractal.getAttributeController(qm));
		boolean prefetchOnQuery = qma.getPrefetchActivatedOnQuery();
		boolean prefetchOnExtent = qma.getPrefetchActivatedOnExtent();

		String strval = props.getProperty(SpeedoProperties.PREFETCH, "");
		if (strval.length()>0) {
			prefetchOnQuery = !isFalse(strval);
			prefetchOnExtent = prefetchOnQuery;
		}
		strval = props.getProperty(SpeedoProperties.PREFETCH_ON_QUERY, "");
		if (strval.length()>0) {
			prefetchOnQuery = !isFalse(strval);
		}
		strval = props.getProperty(SpeedoProperties.PREFETCH_ON_EXTENT, "");
		if (strval.length()>0) {
			prefetchOnExtent = !isFalse(strval);
		}
		qma.setPrefetchActivatedOnQuery(prefetchOnQuery);
		qma.setPrefetchActivatedOnExtent(prefetchOnExtent);
		logger.log(BasicLevel.INFO, "Prefetching: on query: "
				+ (prefetchOnQuery ? "ON" : "OFF") + ", on extent: "
				+ (prefetchOnExtent ? "ON" : "OFF"));
	}

	/**
	 * Configures the mapper component
	 * @param speedo is the Speedo component
	 * @param props is the properties of Speedo
	 * @return true if a connection factory is used, false if the JDBC driver is
	 * used directly.
	 * @throws Throwable if an error or an exception occurs
	 */
	private boolean configureMapper(Component speedo,
								 Properties props) throws Throwable {
		boolean useConnectionFactory;
		String mapperName = props.getProperty(SpeedoProperties.MAPPER_NAME, "rdb.automatic");
		props.remove(SpeedoProperties.MAPPER_NAME);

		Component component = getSubComponent(speedo, PRIMITIVE_MAPPER_PATH);
		String strval = null;
		//Choose the mapper in according to the connection factory (JCA or JDBC)
		String cfName = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME, null);
		props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_FACTORY_NAME);
		useConnectionFactory = cfName != null && cfName.length() > 0;
		if (useConnectionFactory) {
			Object cf = null;
			//There is a connection factory
			try {
				InitialContext ic = new InitialContext();
				cf = ic.lookup(cfName);
			} catch (NamingException e) {
				throw new SpeedoException(
						"Problem to get the connection factory in JNDI ("
						+ cfName + ")", e);
			}
			if (cf == null) {
				throw new SpeedoException(
						"No connection factory registered in JNDI with the name "
						+ cfName);
			} else if (cf instanceof ConnectionSpecJDBC) {
				// Use the JDBC Mapper ==> nothing to do
			} else if (cf instanceof javax.sql.DataSource) {
				// Use the JDBC Mapper ==> nothing to do
			} else if (cf instanceof javax.resource.cci.ConnectionFactory) {
				System.out.println("The connection is JCA compliant: " + cf);
				throw new SpeedoException("JCA datasource no yet supported");
			} else {
				throw new SpeedoException("The connection factory registered "
						+ "in JNDI is not supported by Speedo " + cfName + " => "
						+ cf);
			}

			//Assign the connection factory to the mapper
			logger.log(BasicLevel.INFO, "Data source access through a connection factory found in JNDI:"
					+ "\n\tjndi name: " + cfName
					+ "\n\tconnection factory found: " + cf);
			PMapper mapper = (PMapper) component.getFcInterface("mapper");
			mapper.setMapperName(mapperName);
			mapper.setConnectionFactory(cf);

		} else {
			JDBCMapperAttributes jdbcma = (JDBCMapperAttributes)
					Fractal.getAttributeController(component);

			//deprecated properties
			strval = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old, "");
			if (strval.length()>0) {
				logger.log(BasicLevel.WARN, "Property " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old
						+ " is deprecated, you must use " + SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
				props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME_old);
				
				if (props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME) == null) {
					props.setProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, strval);
				}
			}
			String jdbcDriverCN = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_DRIVER_NAME);
			String jdbcUser = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_USER_NAME);
			String jdbcUrl = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_URL, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_URL);
			String jdbcPass = props.getProperty(SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD, null);
			props.remove(SpeedoProperties.JDO_OPTION_CONNECTION_PASSWORD);
			logger.log(BasicLevel.INFO, "Data source access through a jdbc driver:"
					+ "\n\tdriver= " + jdbcDriverCN
					+ "\n\turl= " + jdbcUrl
					+ "\n\tuser= " + jdbcUser);
			jdbcma.setMapperName(mapperName);
			jdbcma.setDriverClassName(jdbcDriverCN);
			jdbcma.setURL(jdbcUrl);
			jdbcma.setUserName(jdbcUser);
			jdbcma.setPassword(jdbcPass);
			jdbcma.setPoolConnection(true);

			//configure the pool of connection
			configurePool(getSubComponent(speedo, CONNECTION_POOL_PATH),
					SpeedoProperties.CONNECTION_POOL_MIN, "Connection pool: min size: ",
					SpeedoProperties.CONNECTION_POOL_MAX, "Connection pool: max size: ",
					SpeedoProperties.CONNECTION_POOL_TTL, "Connection pool: ttl: ",
					SpeedoProperties.CONNECTION_POOL_TIMEOUT, "Connection pool: timeout: ",
					props);
		}

		//initialize the JormFactory with the mappingStructureRule attribute
		byte mappingStructureRule = JormFactoryAttributes.CREATE_IF_REQUIRED;
		strval = props.getProperty(SpeedoProperties.MAPPING_STRUCTURE, "");
		if (strval.length()>0) {
			props.remove(SpeedoProperties.MAPPING_STRUCTURE);
			if (SpeedoProperties.MAPPING_STRUCTURE_DN.equals(strval)) {
				mappingStructureRule = JormFactoryAttributes.DO_NOTHING;
			} else if (SpeedoProperties.MAPPING_STRUCTURE_DD.equals(strval)) {
				mappingStructureRule = JormFactoryAttributes.DELETE_DATA;
			} else if (SpeedoProperties.MAPPING_STRUCTURE_CIR.equals(strval)) {
				mappingStructureRule = JormFactoryAttributes.CREATE_IF_REQUIRED;
			} else if (SpeedoProperties.MAPPING_STRUCTURE_FC.equals(strval)) {
				mappingStructureRule = JormFactoryAttributes.FORCE_CREATE;
			} else {
				logger.log(BasicLevel.WARN, "Unexpected value for the property '"
						+ SpeedoProperties.MAPPING_STRUCTURE
						+ "', found: '" + strval + "', expected ["
						+ SpeedoProperties.MAPPING_STRUCTURE_DN
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_DD
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_CIR
						+ " | " + SpeedoProperties.MAPPING_STRUCTURE_FC
						+ "]");
			}
		}
		component = getSubComponent(speedo, JORM_FACTORY_PATH);
		JormFactoryAttributes jfa = (JormFactoryAttributes)
                    Fractal.getAttributeController(component);
		jfa.setMappingStructureRule(mappingStructureRule);
		logger.log(BasicLevel.INFO, "Mapping Structure treatment: "
				+ getMappingStructureString(mappingStructureRule));

		return useConnectionFactory;
	}

    private boolean isFalse(String val) {
        return "false".equalsIgnoreCase(val)
                || "off".equalsIgnoreCase(val)
                || "no".equalsIgnoreCase(val)
                || "non".equalsIgnoreCase(val)
                || "n".equalsIgnoreCase(val);
    }

    private boolean isTrue(String val) {
        return "true".equalsIgnoreCase(val)
                || "on".equalsIgnoreCase(val)
                || "yes".equalsIgnoreCase(val)
                || "oui".equalsIgnoreCase(val)
                || "y".equalsIgnoreCase(val)
                || "o".equalsIgnoreCase(val);
    }

	private String getMappingStructureString(byte b) {
		switch(b) {
		case JormFactoryAttributes.DO_NOTHING:
			return SpeedoProperties.MAPPING_STRUCTURE_DN;
		case JormFactoryAttributes.DELETE_DATA:
			return SpeedoProperties.MAPPING_STRUCTURE_DD;
		case JormFactoryAttributes.FORCE_CREATE:
			return SpeedoProperties.MAPPING_STRUCTURE_FC;
		case JormFactoryAttributes.CREATE_IF_REQUIRED:
			return SpeedoProperties.MAPPING_STRUCTURE_CIR;
		default:
			return "UNKNOWN";
		}
	}

	public Object getConcurrencyManagerComponent() throws Exception {
		return getSubComponent(speedo, "tpm.concurrency-manager");
	}
	
	public DependencyGraph getDependencyGraph() throws Exception {
	    Component cmC = (Component) getConcurrencyManagerComponent();
	    Object dg = Fractal.getBindingController(cmC).lookupFc("dependency-graph");
	    return (DependencyGraph) dg;
	}
	
    // IMPLEMENTATION OF THE PersistenceManagerFactory INTERFACE //
    //-----------------------------------------------------------//

    public PersistenceManager getPersistenceManager() {
    	if (!isPropertiesInitialized) {
    		try {
    			init(getSpeedoComponent(), getProperties());
    		} catch(Throwable e) {
    			System.err.println("Error during the initilization of the Speedo " +
    			"persistence manager factory:");
    			e.printStackTrace(System.err);
    			return null;
    		}    		
    	}
    	
        return delegate.getPersistenceManager();
    }

    public PersistenceManager getPersistenceManager(String s, String s1) {
    	if (!isPropertiesInitialized) {
    		try {
    			init(getSpeedoComponent(), getProperties());
    		} catch(Throwable e) {
    			System.err.println("Error during the initilization of the Speedo " +
    			"persistence manager factory:");
    			e.printStackTrace(System.err);
    			return null;
    		}    		
    	}
    	return delegate.getPersistenceManager(s, s1);
    }

    public void setConnectionUserName(String s) {
        delegate.setConnectionUserName(s);
    }

    public String getConnectionUserName() {
        return delegate.getConnectionUserName();
    }

    public void setConnectionPassword(String s) {
        delegate.setConnectionPassword(s);
    }

    public void setConnectionURL(String s) {
        delegate.setConnectionURL(s);
    }

    public String getConnectionURL() {
        return delegate.getConnectionURL();
    }

    public void setConnectionDriverName(String s) {
        delegate.setConnectionDriverName(s);
    }

    public String getConnectionDriverName() {
        return delegate.getConnectionDriverName();
    }

    public void setConnectionFactoryName(String s) {
        delegate.setConnectionFactoryName(s);
    }

    public String getConnectionFactoryName() {
        return delegate.getConnectionFactoryName();
    }

    public void setConnectionFactory(Object o) {
        delegate.setConnectionFactory(o);
    }

    public Object getConnectionFactory() {
        return delegate.getConnectionFactory();
    }

    public void setConnectionFactory2Name(String s) {
        delegate.setConnectionFactory2Name(s);
    }

    public String getConnectionFactory2Name() {
        return delegate.getConnectionFactory2Name();
    }

    public void setConnectionFactory2(Object o) {
        delegate.setConnectionFactory2(o);
    }

    public Object getConnectionFactory2() {
        return delegate.getConnectionFactory2();
    }

    public void setMultithreaded(boolean b) {
        delegate.setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        return delegate.getMultithreaded();
    }

    public void setOptimistic(boolean b) {
        delegate.setOptimistic(b);
    }

    public boolean getOptimistic() {
        return delegate.getOptimistic();
    }

    public void setRetainValues(boolean b) {
        delegate.setRetainValues(b);
    }

    public boolean getRetainValues() {
        return delegate.getRetainValues();
    }

    public void setRestoreValues(boolean b) {
        delegate.setRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return delegate.getRestoreValues();
    }

    public void setNontransactionalRead(boolean b) {
        delegate.setNontransactionalRead(b);
    }

    public boolean getNontransactionalRead() {
        return delegate.getNontransactionalRead();
    }

    public void setNontransactionalWrite(boolean b) {
        delegate.setNontransactionalWrite(b);
    }

    public boolean getNontransactionalWrite() {
        return delegate.getNontransactionalWrite();
    }

    public void setIgnoreCache(boolean b) {
        delegate.setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        return delegate.getIgnoreCache();
    }

    public Properties getProperties() {
        return delegate.getProperties();
    }

    public Collection supportedOptions() {
        return delegate.supportedOptions();
    }

    public PersistenceManagerFactory getDelegate() {
        return delegate;
    }

	public ProxyManager lookup() {
		return pmfComponent.lookup();
	}

	public void unbindPM() {
		pmfComponent.unbindPM();
	}

	public void bindPM2Thread(ProxyManager pm) {
		pmfComponent.bindPM2Thread(pm);
	}

	public void proxyManagerClosed(ProxyManager pr) {
		pmfComponent.proxyManagerClosed(pr);
	}

	public void close() {
		pmfComponent.close();
	}
}
