/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Attribute;
import org.objectweb.util.monolog.api.Logger;

import java.util.Map;
import java.util.HashMap;
import java.util.List;

/**
 * Creates several copies of a method.
 */
public class CodeDuplicator extends LoggedClass implements CodeVisitor {

    /**
     * The code visitor used to generate the copies.
     */
    private CodeVisitor[] cvs;

    /**
     * A map associating original labels to new labels, for each copy. Indeed
     * labels can be used only inside a single method. It is therefore necessary
     * to create copies of the original labels for each duplicated method.
     */
    private Map[] labels;

    public CodeDuplicator(Logger logger) {
        super(logger);
    }

    /**
     * Constructs a {@link CodeDuplicator}.
     *
     * @param cvs a list of code visitors, to be used to generate the copies.
     */
    public CodeDuplicator(List cvs) {
        this.cvs = (CodeVisitor[]) cvs.toArray(new CodeVisitor[cvs.size()]);
        this.labels = new Map[this.cvs.length];
    }

    /**
     * Constructs a {@link CodeDuplicator}.
     *
     * @param cvs an array of code visitors, to be used to generate the copies.
     */
    public CodeDuplicator(CodeVisitor[] cvs) {
        this.cvs = cvs;
        this.labels = new Map[cvs.length];
    }

    // IMPLEMENTATION OF THE CodeVisitor INTERFACE //
    // --------------------------------------------//

    public void visitInsn(final int opcode) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitInsn(opcode);
        }
    }

    public void visitIntInsn(final int opcode, final int operand) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitIntInsn(opcode, operand);
        }
    }

    public void visitVarInsn(final int opcode, final int var) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitVarInsn(opcode, var);
        }
    }

    public void visitTypeInsn(final int opcode, final String desc) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitTypeInsn(opcode, desc);
        }
    }

    public void visitFieldInsn(final int opcode,
                               final String owner,
                               final String name,
                               final String desc) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitFieldInsn(opcode, owner, name, desc);
        }
    }

    public void visitMethodInsn(final int opcode,
                                final String owner,
                                final String name,
                                final String desc) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitMethodInsn(opcode, owner, name, desc);
        }
    }

    public void visitJumpInsn(final int opcode, final Label label) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitJumpInsn(opcode, getLabel(label, i));
        }
    }

    public void visitLabel(final Label label) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitLabel(getLabel(label, i));
        }
    }

    public void visitLdcInsn(final Object cst) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitLdcInsn(cst);
        }
    }

    public void visitIincInsn(final int var, final int increment) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitIincInsn(var, increment);
        }
    }

    public void visitTableSwitchInsn(final int min,
                                     final int max,
                                     final Label dflt,
                                     final Label labels[]) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitTableSwitchInsn(min,
                    max, getLabel(dflt, i), getLabels(labels, i));
        }
    }

    public void visitLookupSwitchInsn(final Label dflt,
                                      final int keys[],
                                      final Label labels[]) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitLookupSwitchInsn(getLabel(dflt, i),
                    keys, getLabels(labels, i));
        }
    }

    public void visitMultiANewArrayInsn(final String desc, final int dims) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitMultiANewArrayInsn(desc, dims);
        }
    }

    public void visitTryCatchBlock(final Label start,
                                   final Label end,
                                   final Label handler,
                                   final String type) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitTryCatchBlock(getLabel(start, i),
                    getLabel(end, i), getLabel(handler, i), type);
        }
    }

    public void visitMaxs(final int maxStack, final int maxLocals) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitMaxs(maxStack, maxLocals);
        }
    }

    public void visitLocalVariable(final String name,
                                   final String desc,
                                   final Label start,
                                   final Label end,
                                   final int index) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitLocalVariable(name,
                    desc, getLabel(start, i), getLabel(end, i), index);
        }
    }

    public void visitLineNumber(final int line, final Label start) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitLineNumber(line, getLabel(start, i));
        }
    }

    public void visitAttribute(Attribute attribute) {
        for (int i = 0; i < cvs.length; ++i) {
            cvs[i].visitAttribute(attribute);
        }
    }

    // OTHER METHODS : LABELS MANAGEMENT //
    // ----------------------------------//

    /**
     * Returns the label corresponding to the given label for the i-th copy.
     *
     * @param label a label of the method that is visited by this visitor
     * @param i index of the copy for which the new label must be computed
     * @return the label corresponding to the given label for the i-th copy.
     */
    private Label getLabel(final Label label, final int i) {
        if (labels[i] == null) {
            labels[i] = new HashMap();
        }
        Label l = (Label) labels[i].get(label);
        if (l == null) {
            l = new Label();
            labels[i].put(label, l);
        }
        return l;
    }

    /**
     * Returns the labels corresponding to the given labels for the i-th copy.
     *
     * @param labels some labels of the method that is visited by this visitor
     * @param i index of the copy for which the new label must be computed
     * @return the labels corresponding to the given labels for the i-th copy.
     */
    private Label[] getLabels(final Label[] labels, final int i) {
        Label[] l = new Label[labels.length];
        for (int j = 0; j < l.length; ++j) {
            l[j] = getLabel(labels[j], i);
        }
        return l;
    }
}
