/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.generation.generator.proxy;

import org.apache.velocity.app.Velocity;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.generation.generator.api.SpeedoGenerationException;
import org.objectweb.speedo.generation.generator.lib.SpeedoGenerator;
import org.objectweb.speedo.mapper.lib.Object2StringSerializer;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.speedo.tools.StringReplace;
import org.objectweb.util.monolog.wrapper.velocity.VelocityLogger;

import java.io.File;
import java.io.FileWriter;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

/**
 * This class is used to generate the file proxy.
 * This file will represent the persistent instance.
 * @author	S. Chassande-Barrioz
 */
public class ProxyGenerator
	extends SpeedoGenerator {

    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".generation.generator.proxy";

    public final static String TEMPLATE_NAME = TEMPLATE_DIR + ".proxy.Proxy";

    // IMPLEMENTATION OF THE GeneratorComponent INTERFACE //
    //----------------------------------------------------//

    public boolean init() throws SpeedoException {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
	    return !scp.getXmldescriptor().isEmpty();
    }


    // IMPLEMENTATION OF THE VelocityGenerator INTERFACE //
    //---------------------------------------------------//

    /**
     * This method generates the new file.
     *
     * @param sClass is the SpeedoClass
     * @param fileName name of the new file.
     * @exception org.objectweb.speedo.generation.generator.api.SpeedoGenerationException if there is a problem during writing
     * the new file.
     */
    public void generate(SpeedoClass sClass, String fileName)
            throws SpeedoException {
        //EnhancerTool.logger = logger;
        if (templateFN == null)
            templateFN = TEMPLATE_NAME.replace('.', fs) + ".vm";
        try {
            if (!ve.templateExists(templateFN)) {
                templateFN = templateFN.replace(fs, '/');
                if (!ve.templateExists(templateFN)) {
                    throw new SpeedoException("The Speedo template " + templateFN
                        + " is not availlable in the classpath");
                }
            }
            Context ctx = getContext(sClass);
            FileWriter fw = new FileWriter(fileName);
            ve.getTemplate(templateFN).merge(ctx, fw);
            ve.setProperty(Velocity.RUNTIME_LOG_LOGSYSTEM, new VelocityLogger(logger));
            fw.flush();
            fw.close();
        } catch (Exception e) {
            throw new SpeedoGenerationException(
                    "Error during the generation of the file " + fileName, e);
        }
    }

    // PRIVATE METHODS //
    //-----------------//

    /**
     * This method initialises the Velocity context.
     *
     * @param jdoClass : the jdoClass which represents the class to generate.
     * @return a Velocity context.
     */
    protected Context getContext(SpeedoClass jdoClass) throws SpeedoException {
		Context ctx = super.getContext(jdoClass);

        //class declaration
        ctx.put("signatureClass", jdoClass.signature);
        ctx.put("serialversionuid", jdoClass.VersionUID + "L");

        // Jorm config
		Properties classProperties = new Properties();
		getJormConfig(jdoClass, classProperties);
		String xmlFileName = jdoClass.jdoPackage.jdoXMLDescriptor.xmlFile;
		xmlFileName = StringReplace.replaceChar('/', '.', xmlFileName);
		xmlFileName = StringReplace.replaceChar(fs, '.', xmlFileName);
		classProperties.setProperty(
			Object2StringSerializer.JDO_FILE_NAME_PROP,xmlFileName);
		//Build the array of String with the calculated Map
		String[] jormconfigInit = new String[classProperties.size()];
		int i=0;
		for (Iterator it = classProperties.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			StringBuffer sb = new StringBuffer();
			sb.append("classProperties.setProperty(\"");
			sb.append((String) me.getKey());
			sb.append("\", \"");
			sb.append((String) me.getValue());
			sb.append("\");");
			jormconfigInit[i] = sb.toString();
			i++;
		}
		ctx.put("jormconfig", jormconfigInit);

        //Define the 'pnameHints' attribute.
        String pnamehints = scp.nmf.getNamingManager(jdoClass).getPNameHints(
                        jdoClass, getClassNameDef(jdoClass.jormclass));
        ctx.put("pnameHints", pnamehints);
        return ctx;
    }

	private GenClassMapping getGenClassMapping(GenClassRef gcr,
                                 Class clazz) throws SpeedoException {
		String gcid = gcr.getGenClassId();
		GenClassMapping gcm = getMapping(clazz).getGenClassMapping(gcid);
		if (gcm == null) {
			Class _clazz = clazz;
			Collection superclasses = _clazz.getSuperClasses();
			while (gcm == null && !superclasses.isEmpty()) {
				_clazz = (Class) superclasses.iterator().next();
				gcm = getMapping(_clazz).getGenClassMapping(gcid);
				superclasses = _clazz.getSuperClasses();
			}
		}
		if (gcm == null) {
			throw new SpeedoException(
				"No GenClassMapping found for the field " + gcid);
		}
		return gcm;
	}

    private NameDef getRefNameDef(GenClassRef gcr,
                                 Class clazz) throws SpeedoException {
		return (NameDef) getGenClassMapping(gcr, clazz)
			.getIdentifierMapping().getLinkedMO();
    }

    private NameDef getElemNameDef(GenClassRef gcr,
                                 Class clazz) throws SpeedoException {
		return (NameDef) getGenClassMapping(gcr, clazz)
                .getReferenceMapping().getLinkedMO();
    }

	private ReferenceMapping getReferenceMapping(ClassRef cr,
                                 Class clazz) throws SpeedoException {
		String rid = cr.getName();
		ReferenceMapping rm = getMapping(clazz).getClassMapping()
			.getReferenceMapping(rid);
		if (rm == null) {
			Class _clazz = clazz;
			Collection superclasses = _clazz.getSuperClasses();
			while (rm == null && !superclasses.isEmpty()) {
				_clazz = (Class) superclasses.iterator().next();
				rm = getMapping(_clazz).getClassMapping()
					.getReferenceMapping(rid);
				superclasses = _clazz.getSuperClasses();
			}
		}
		if (rm == null) {
			throw new SpeedoException(
				"No ReferenceMapping found for the field " + rid);
		}
		return rm;

	}
    private NameDef getRefNameDef(ClassRef cr,
                                 Class clazz) throws SpeedoException {
        return (NameDef) getReferenceMapping(cr, clazz).getLinkedMO();
    }


    private void jormConfigs(NameDef nd,
							 SpeedoClass targetClass,
							 MetaObject sourceMO,
							 String key,
							 Properties jormProp) throws SpeedoException {
        scp.nmf.getNamingManager(targetClass)
                .getJormNamingConfig(nd, targetClass, sourceMO, key, jormProp);
    }

    private void getJormConfig(SpeedoClass jdoClass, Properties classProperties) throws SpeedoException {
        //Build the map
        // Put the binder class name of the current persistent class
        String id = JormPathHelper.getPath(jdoClass.jormclass);
        NameDef nd = getClassNameDef(jdoClass.jormclass);
        jormConfigs(nd, jdoClass, jdoClass.jormclass, id, classProperties);

        //Put the binder class name for each reference field
        SpeedoXMLDescriptor xml = jdoClass.jdoPackage.jdoXMLDescriptor;
        for (Iterator it = jdoClass.jormclass.getAllFields().iterator(); it.hasNext();) {
            Object o = it.next();
            if (o instanceof ClassRef) {
                nd = getRefNameDef((ClassRef) o, jdoClass.jormclass);
                id = JormPathHelper.getPath(((ClassRef) o));
                SpeedoClass tsc = xml.getSpeedoClass(((ClassRef) o).getMOClass().getFQName(), true);
                jormConfigs(nd, tsc, jdoClass.jormclass, id, classProperties);
            } else if (o instanceof GenClassRef) {
                GenClassRef gcr = (GenClassRef) o;
                nd = getRefNameDef(gcr, jdoClass.jormclass);
                id = JormPathHelper.getPath((GenClassRef) o, false);
                jormConfigs(nd, jdoClass, gcr, id, classProperties);
                while (gcr.isGenClassRef()) {
                    gcr = gcr.getGenClassRef();
                    nd = getElemNameDef(gcr, jdoClass.jormclass);
                    id = JormPathHelper.getPath(gcr, false);
                    jormConfigs(nd, jdoClass, gcr, id, classProperties);
                }
                if (gcr.isClassRef()) {
                    ClassRef cr = gcr.getClassRef();
                    nd = getElemNameDef(gcr, jdoClass.jormclass);
                    id = JormPathHelper.getPath(cr);
                    SpeedoClass tsc = xml.getSpeedoClass(cr.getMOClass().getFQName(), true);
                    jormConfigs(nd, tsc, cr, id, classProperties);
                }
            }
        }
    }


}
