/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.jorm.rdb;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbExternalTable;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbGenClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbJoin;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbTable;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CommonClassMapping;
import org.objectweb.jorm.metainfo.api.GenClass;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.ParentClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.lib.BasicClassMapping;
import org.objectweb.jorm.metainfo.lib.BasicScalarField;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.generation.jorm.JormMIMappingBuilder;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.metadata.SpeedoElement;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * This class is an implementation of the JormMIMappingBuilder for the mapper
 * rdb and its sub mappers. It defines a simple mapping where each persistent
 * objects (generic class included) are mapped over one table.
 *
 * @author S.Chassande-Barrioz
 */
public class OneTableRdbJormIMappingBuilder
        implements JormMIMappingBuilder, SpeedoProperties, Loggable {

    private final static String CLASS_MAPPING_RULE_NAME = "multi-table";
    private final static String GENCLASS_MAPPING_RULE_NAME = "to-table";

    private Logger logger;
	private boolean debug = false;

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    public Logger getLogger() {
        return logger;
    }

    public LoggerFactory getLoggerFactory() {
        return null;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
		debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
    }

    public void setLoggerFactory(LoggerFactory loggerFactory) {
        if (logger == null && loggerFactory != null) {
            setLogger(loggerFactory.getLogger(
                    SpeedoProperties.LOGGER_NAME
				+ ".generation.jorm.mimappingbuilder"));
        }
    }

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    /**
     * It builds a RdbClassMultiMapping and creates the main table.
     * The name of the table is built with the short class name.
     * @param clazz is the Jorm meta object representing the class which the
     * ClassMapping must be built.
     * @param mapping is the Mapping instance which will host the ClassMapping.
     * @return the ClassMapping instance built by the method (never null).
     * @throws org.objectweb.jorm.api.PException if it is not possible to build the ClassMapping
     */
    public ClassMapping createClassMapping(Class clazz,
                                           SpeedoClass sc,
                                           Mapping mapping)
            throws PException, SpeedoException {
        RdbClassMultiMapping rcm = new RdbClassMultiMapping(
                CLASS_MAPPING_RULE_NAME, clazz, mapping);
        mapping.setClassMapping(rcm);
	    rcm.createRdbTable(getTableName(sc));
        return rcm;
    }

	public void finalizeClassMapping(SpeedoClass sc,
	                                 String projectName,
	                                 String mapperName)
	    throws PException, SpeedoException {
		RdbClassMultiMapping rcm = (RdbClassMultiMapping) sc.jormclass
		    .getClassProject(projectName).getMapping(mapperName).getClassMapping();
		if (sc.superClassName != null) {
			String ruleName;
			String im = sc.getExtensionValueByKey(SpeedoProperties.INHERITANCE_MAPPING);
			if (im == null) {
				throw new SpeedoException(
				    "No inheritance mapping defined for the class '" + sc.getFQName() +"'");
			} else if (SpeedoProperties.INHERITANCE_FILTERED_MAPPING.equals(im)) {

			    ruleName = RdbClassMapping.MAP_NEW_FIELDS_TO_EXTENDED_STRUCTURES;
			} else if (SpeedoProperties.INHERITANCE_HORIZONTAL_MAPPING.equals(im)) {
				ruleName = RdbClassMapping.REMAP_FIELDS_TO_NEW_STRUCTURES;
			} else if (SpeedoProperties.INHERITANCE_VERTICAL_MAPPING.equals(im)) {
				ruleName = RdbClassMapping.MAP_NEW_FIELDS_TO_ADDED_STRUCTURES;
				//TODO: support vertical inheritance mapping
				//TODO: link the main table to the one of the parent, and create an external table
				throw new SpeedoException("The vertical mapping for the inheritance is not supported currently");
			} else {
				throw new SpeedoException("Inheritance mapping type is unknown ("
				    + im + ") for the class '" + sc.getFQName() +"'");
			}
			SpeedoClass parent = sc.getSpeedoClassFromContext(sc.superClassName);
			rcm.createParentClassMapping(ruleName, sc.jormclass.getSuperClass(parent.getFQName()));
			//Add inter dependencies between the parent and its child
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Add dependencies between "
					+ parent.getFQName() + " and " + sc.getFQName());
			}
			rcm.addDependency(parent.getFQName());
			parent.jormclass.getClassProject(projectName)
				.getMapping(mapperName).getClassMapping().addDependency(sc.getFQName());
		}
	}

    /**
     * It builds a RdbClassMultiMapping and creates the table.
     * the table is build with the generic class identifier ("A/f1" ==> "A_f1")
     * @param gcr is the Jorm meta object representing the gen class which the
     * GenClassMapping must be built.
     * @param mapping is the Mapping instance which will host the GenClassMapping.
     * @return the GenClassMapping instance built by the method (never null).
     * @throws org.objectweb.jorm.api.PException if it is not possible to build the GenClassMapping
     */
    public GenClassMapping createGenClassMapping(GenClassRef gcr,
                                                 SpeedoField sf,
                                                 Mapping mapping)
            throws PException, SpeedoException {
        RdbGenClassMapping rgcm = new RdbGenClassMapping(
                GENCLASS_MAPPING_RULE_NAME, gcr, mapping);
        mapping.addGenClassMapping(gcr.getGenClassId(), rgcm);
        SpeedoExtension se = sf.getExtension(VENDOR_NAME, JOIN_TABLE);
        if (se != null) {
            rgcm.createRdbTable(se.value);
            se = sf.getExtensionByKey(REVERSE_FIELD);
            if (se != null) {
                String cn = null;
                if (sf.jdoTuple instanceof SpeedoCollection) {
                    cn = (String) ((SpeedoCollection) sf.jdoTuple).elementType;
                } else if (sf.jdoTuple instanceof SpeedoMap) {
					cn = (String) ((SpeedoMap) sf.jdoTuple).valueType;
				}
                SpeedoClass sc = sf.jdoClass.jdoPackage.jdoXMLDescriptor.smi.
	                getSpeedoClass(cn, sf.jdoClass.jdoPackage);
                SpeedoField sf2 = (SpeedoField) sc.jdoField.get(se.value);
                Class c = ((Manager) gcr.getParent().getParent().getParent())
                        .getClass(sc.getFQName());
                TypedElement te = c.getTypedElement(sf2.name);
                if (te instanceof GenClassRef) {
                    rgcm.getRdbTable().setReadOnly(true);
                    logger.log(BasicLevel.DEBUG, "field '" + sf.name + "' is read only");
                }
                //optimisation: if there is a reverse field then we suppose that
            }
        } else if (gcr.isClassRef()
                && sf.getExtension(VENDOR_NAME, SOURCE_FK) != null
                && sf.getExtension(VENDOR_NAME, TARGET_FK) == null) {
            // The gen class is mapped over the table of the targeted class
            SpeedoClass tsc = sf.jdoClass.jdoPackage.jdoXMLDescriptor
                    .getSpeedoClass(gcr.getClassRef().getMOClass().getFQName(), true);
            se = (tsc == null ? null : tsc.getExtensionByKey(SQL_NAME));
            Manager mgr = getManager(gcr);
            Class tclazz = mgr.getClass(tsc.getFQName());
            RdbClassMultiMapping trcm = (RdbClassMultiMapping)
            tclazz.getClassProject(mapping.getClassMapping().getProjectName())
                .getMapping(mapping.getMapperName()).getClassMapping();
            RdbTable t = rgcm.createRdbTable(
                    (se == null ? getTableName(tsc) : se.value));
            // The GC table and the table of the targeted class are colocated
            t.setColocated(true);
            se = sf.getExtensionByKey(REVERSE_FIELD);
            if (se != null) {
                logger.log(BasicLevel.DEBUG, "REVERSE-FIELD(" + sf.name + "):" + se.value);
                //optimisation: if there is a reverse field then we suppose that
                // the user will do the coherence.
                t.setReadOnly(true);
            } else {
                logger.log(BasicLevel.DEBUG, "No REVERSE-FIELD(" + sf.name + ")");
                trcm.getRdbTable().setColocated(true);
                trcm.getRdbTable().setColocatedMaster(true);
            }
            trcm.addDependency(sf.jdoClass.getFQName());
            mapping.getClassMapping().addDependency(tclazz.getFQName());
        } else {
            //default mapping: join table
            rgcm.createRdbTable(getTableName(gcr));
        }
        return rgcm;
    }

    /**
     * The column name is built with the field name.
     * @param pe is the Jorm meta object representing a primitive field
     * @param cm is the MappingStructure which will host the mapping of the
     * field
     * @return a PrimitiveElementMapping corresponding to the given primitive
     * field.
     * @throws org.objectweb.jorm.api.PException if it is not possible to build the mapping of the
     * primitive field.
     */
    public PrimitiveElementMapping createFieldMapping(PrimitiveElement pe,
                                                      SpeedoElement selem,
                                                      CommonClassMapping cm)
            throws PException {
        RdbTable t;
		boolean ingc = false;
		String extensionColName = SQL_NAME;
		String extensionColType = SQL_TYPE;
        if (cm instanceof RdbClassMultiMapping) {
            t = ((RdbClassMultiMapping) cm).getRdbTable();
        } else if (cm instanceof RdbGenClassMapping) {
			ingc = true;
			GenClassRef gcr = (GenClassRef) pe.getParent();
			if (gcr.getPrimitiveElement() == pe) {
				extensionColName = ELEMENT;
				extensionColType = ELEMENT_TYPE;
			} else {
				Iterator it = gcr.getIndexFields().iterator();
				if (it.hasNext() && pe == it.next()) {
					extensionColName = INDEX;
					extensionColType = INDEX_TYPE;
				}
			}
            t = ((RdbGenClassMapping) cm).getRdbTable();
        } else {
            new Exception().printStackTrace();
            throw new PException("Unmanaged HelperClassMapping: " + cm);
        }
	    String colName = getColumnName(pe);
		String colType = null;
        if (selem != null) {
	        if (selem instanceof SpeedoClass) {
				//Search the mapping of the inherited field (horizontal mapping).
		        extensionColName = pe.getName() + '-' + extensionColName;
		        extensionColType = pe.getName() + '-' + extensionColType;
	        } else if (selem instanceof SpeedoField
					&& ((SpeedoField) selem).jdoTuple instanceof SpeedoMap
					&& selem.getExtensionValueByKey(KEY_FIELD) != null
					) {

			}
			SpeedoExtension seColName = selem.getExtensionByKey(extensionColName);
			if (seColName != null) {
				colName = seColName.value;
			}
			SpeedoExtension seColType = selem.getExtensionByKey(extensionColType);
			if (seColType != null) {
				colType = seColType.value;
			}
        }
        return t.createPrimitiveElementMapping(pe,colName, colType, false);
    }

    /**
     * It creates the mapping of a name def (reference or identifier).
     * @param cm is the MappingStructure which will host the mapping of the
     * fields of the name def
     * @param nd is the NameDef which must be mapped.
     * @throws org.objectweb.jorm.api.PException if it is not possible to build the mapping of the
     * field.
     */
    public void createNameDefMapping(CommonClassMapping cm,
                                     NameDef nd,
                                     SpeedoClass sc,
                                     boolean isIdentifier,
                                     boolean isInGenClass)
            throws PException, SpeedoException {
		if (debug) {
			logger.log(BasicLevel.DEBUG, "RdbMapping: createNameDefMapping "
				+ "\n\tnd=" + nd
				+ "\n\tsc=" + (sc == null ? null : sc.getFQName())
				+ "\n\tisIdentifier=" + isIdentifier
				+ "\n\tisInGenClass=" + isInGenClass
				);
		}
        if (!isInGenClass) {
            Class clazz;
            if (isIdentifier) {
                //This is the identifier of the class
                clazz = (Class) nd.getParent();
                if (nd.isFieldName()) {
                    PrimitiveElement pe = (PrimitiveElement)
                            clazz.getTypedElement(nd.getFieldName());
                    if (pe == null) {
                        throw new SpeedoException("impossible to map the fields of the namedef of the metaobject '" + JormPathHelper.getPath(clazz) + "'.");
                    }
                    createFieldMapping(pe, null, cm);
                } else {
                    Iterator it = nd.getNameRef().getProjection().values().iterator();
                    while(it.hasNext()) {
                        PrimitiveElement pe = (PrimitiveElement)
                            clazz.getTypedElement((String) it.next());
                        if (pe == null) {
                            throw new SpeedoException("impossible to map the fields of the namedef of the metaobject '" + JormPathHelper.getPath(clazz) + "'.");
                        }
                        createFieldMapping(pe, null, cm);
                    }
                }
            } else {
                MetaObject mo = nd.getParent();
                if (mo instanceof GenClassRef) {
                    NameDef classNd = (NameDef) cm.getIdentifierMapping().getLinkedMO();
                    if ((nd.isFieldName() && classNd.isFieldName() && nd.getFieldName().equals(classNd.getFieldName()))
                        || (nd.isNameRef() && classNd.isNameRef()
                            && classNd.getNameRef().getCompositeName().equals(nd.getNameRef().getCompositeName())
                            && classNd.getNameRef().getProjection().equals(nd.getNameRef().getProjection())) ) {
                        // The reference of the gen class is mapped over the source
                        // class identifier, then the field of the identifier are
                        // already mapped.
                        logger.log(BasicLevel.DEBUG, "Mapping of the GCR ref '"
                                + ((GenClassRef) mo).getName()
                                + "' already defined");
                    } else {
                        //The reference of the gen class is mapped over fields
                        // which do not composed the source class identifier
                        // then the mapping of these field must be created.
                        createGenClassRefNameDef(sc, (GenClassRef) mo, cm, nd);
                    }
                } else {
                    //This is a reference to a persistent object since a class.
                    createClassRefNameDef(sc, (ClassRef) mo, cm, nd);
                }
            }
        } else {
            if (isIdentifier) {
                creaGenClassRefIdNameDef(nd, sc, (RdbGenClassMapping) cm);
            } else {
                creaGenClassRefRefNameDef(nd, sc, (RdbGenClassMapping) cm);
            }
        }
    }


    // PRIVATE METHODS //
    //-----------------//
    private void createGenClassRefNameDef(SpeedoClass sc,
                                       GenClassRef gcr,
                                       CommonClassMapping cm,
                                       NameDef gcrnd)
            throws SpeedoException, PException {
        logger.log(BasicLevel.DEBUG, "Defining the mapping of the GCR ref '"
                + gcr.getName() + "' ...");
        //The ClassMapping of the current class.
        RdbClassMultiMapping rcm = (RdbClassMultiMapping) cm;
        if (gcrnd.isFieldName()) {
            String fn = gcrnd.getFieldName();
            PrimitiveElement pe = (PrimitiveElement)
                    sc.jormclass.getTypedElement(fn);
            if (pe == null) {
                pe = sc.jormclass.getHiddenField(fn);
            }
            if (pe == null) {
                throw new SpeedoException("No field '" + fn
                        + "' found on the class '" + sc.getFQName() + "'");
            }
            logger.log(BasicLevel.DEBUG, "Map the PE '" + pe.getName()
                    + "' on the column '" + pe.getName() + "'.");
            rcm.getRdbTable().createPrimitiveElementMapping(pe, pe.getName());
        } else if (gcrnd.isNameRef()) {
            Iterator itClaFns = gcrnd.getNameRef().getProjection().values().iterator();
            while(itClaFns.hasNext()) {
                String fn = (String) itClaFns.next();
                PrimitiveElement pe = (PrimitiveElement)
                        sc.jormclass.getTypedElement(fn);
                if (pe == null) {
                    pe = sc.jormclass.getHiddenField(fn);
                }
                if (pe == null) {
                    throw new SpeedoException("No field '" + fn
                            + "' found on the class '" + sc.getFQName() + "'");
                }
                logger.log(BasicLevel.DEBUG, "Map the PE '" + pe.getName()
                        + "' on the column '" + pe.getName() + "'.");
                rcm.getRdbTable().createPrimitiveElementMapping(pe, pe.getName());
            }
        }
    }

    /**
     * Mappes a ClassRef over sql columns. In according to the speedo extension
     * a local foreign key (default case) or a join with the referenced table
     * is added for the reference mapping.
     * @param sc is the SpeedoClass hosting the reference
     * @param cr is the jorm meta object representing the reference
     * @param cm is the ClassMapping of the class hosting the reference
     * @param crnd is the name def of the reference.
     * @throws org.objectweb.speedo.api.SpeedoException if the meta information is not complete
     */
    private void createClassRefNameDef(SpeedoClass sc,
                                       ClassRef cr,
                                       CommonClassMapping cm,
                                       NameDef crnd)
            throws SpeedoException, PException {
        logger.log(BasicLevel.DEBUG, "Create the name def of the ClassRef: "
                + JormPathHelper.getPath(cr));
        //The ClassMapping of the current class.
        RdbClassMultiMapping rcm = (RdbClassMultiMapping) cm;
        //The Speedo Meta object of the referenced class
        SpeedoClass tsc = sc.jdoPackage.jdoXMLDescriptor
            .getSpeedoClass(cr.getMOClass().getFQName(), true);
        //The ClassMapping of the referenced class.
        RdbClassMultiMapping trcm = (RdbClassMultiMapping)
                tsc.jormclass.getClassProject(cm.getProjectName())
                .getMapping(cm.getMapperName()).getClassMapping();
        //The Speedo meta object mathcing to the ClassRef
        SpeedoField sf = (SpeedoField) sc.jdoField.get(cr.getName());
        if (sf == null) {
            throw new SpeedoException("impossible to map the namedef of the field '"
                    + JormPathHelper.getPath(cr) + "': no SpeedoField '"
                    + cr.getName()+ "' on the SpeedoClass " + sc.getFQName());
        }

        //The name def of the referenced class
        NameDef tclassnd = (NameDef)
                trcm.getIdentifierMapping().getLinkedMO();

        SpeedoExtension se = sf.getExtension(VENDOR_NAME, SOURCE_FK);
        String sfk = (se == null ? null : se.value);
        if (sfk != null) {
            //foreign key in the external table
			if (tsc.superClassName != null) {
				throw new PException(
					"Jorm does not support to store a polymorphic reference '"
					+ sc.getFQName() + "." + cr.getName()+ "'with a " + SOURCE_FK);
			}
            se = tsc.getExtensionByKey(SQL_NAME);
            String tn = (se == null ? getTableName(sc) : se.value);
            logger.log(BasicLevel.DEBUG, "  - External table: " + tn);
            RdbExternalTable exttable = rcm.createRdbExternalTable(tn);
            if (sf.getExtensionByKey(REVERSE_FIELD) != null) {
                exttable.setReadOnly(true);
            }
            // The external table and the table of the targeted class are
            // colocated
            exttable.setColocated(true);
            trcm.getRdbTable().setColocated(true);
            trcm.getRdbTable().setColocatedMaster(true);
            trcm.addDependency(sc.getFQName());
            cm.addDependency(tsc.getFQName());

            //optimisation: if there is a reverse field then we suppose that
            // the user will do the coherence.
            se = sf.getExtension(VENDOR_NAME, REVERSE_FIELD);
            logger.log(BasicLevel.DEBUG, "  - readonly: " + (se != null));
            exttable.setReadOnly(se != null);

            //Define the join to the external table
            RdbJoin j = exttable.createRdbJoin(cr.getName());
            Iterator it = getPk2Fk(sfk).entrySet().iterator();
            while(it.hasNext()) {
                Map.Entry me = (Map.Entry) it.next();
                logger.log(BasicLevel.DEBUG, "  - join: "
                        + me.getKey() + " == " + me.getValue());
                j.addJoinColumnNames((String) me.getKey(),
                        (String) me.getValue());
            }
            //map the name def over pk field of the referenced class
            if (tclassnd.isFieldName()) {
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(tclassnd.getFieldName(), true);
                PrimitiveElement pe = (PrimitiveElement)
                        sc.jormclass.getTypedElement(crnd.getFieldName());
                exttable.createPrimitiveElementMapping(pe, pem.getName(), null, false, j);
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                it = crnd.getNameRef().getProjection().entrySet().iterator();
                while(it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe = (PrimitiveElement)
                            sc.jormclass.getTypedElement((String) me.getValue());
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        trcm.getPrimitiveElementMapping(
                                (String) tclassndproj.get(me.getKey()), true);
                    exttable.createPrimitiveElementMapping(pe, pem.getName(), null, false, j);
                }
            }
        } else {
            //local foreign keys
            RdbTable table = rcm.getRdbTable();
            logger.log(BasicLevel.DEBUG, "  - main table: " + table.getName());

            se = sf.getExtension(VENDOR_NAME, TARGET_FK);
            String tfk = (se == null ? null : se.value);
            Map pk2fk = null;
            if (tfk != null) {
                pk2fk = getPk2Fk(tfk);
            } else {
                //default mapping of a reference: local foreign keys
                // The column name is equal to the field name.
                logger.log(BasicLevel.DEBUG, "  - default mapping");
                pk2fk = new HashMap();
                if (crnd.isFieldName()) {
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            trcm.getPrimitiveElementMapping(tclassnd.getFieldName(), true);
                    pk2fk.put(pem.getName(), crnd.getFieldName());
                    logger.log(BasicLevel.DEBUG, "  - pk2fk: " + pem.getName() + " == " + crnd.getFieldName());
                } else {
                    Iterator it = crnd.getNameRef().getProjection().entrySet().iterator();
                    Map tclassndproj = tclassnd.getNameRef().getProjection();
                    while (it.hasNext()) {
                        Map.Entry me = (Map.Entry) it.next();
                        RdbPrimitiveElementMapping pem = getPEMOfField(
							tsc, cm.getProjectName(), cm.getMapperName(), (String)
						tclassndproj.get(me.getKey()));
                        pk2fk.put(pem.getName(), me.getValue());
                    }
                }
            }
            //Map the name def of the reference over new local columns
            if (crnd.isFieldName()) {
                //find the pk column name
                RdbPrimitiveElementMapping pem = getPEMOfField(
					tsc, cm.getProjectName(), cm.getMapperName(), tclassnd.getFieldName());

                //find the field in the class used in by this name def
                PrimitiveElement pe = (PrimitiveElement)
                        sc.jormclass.getTypedElement(crnd.getFieldName());
                //find the fk column name
                String colname = (String) pk2fk.get(pem.getName());
                //check if the fk column name is already used
                PrimitiveElementMapping pem2 =
                        table.getPrimitiveElementMappingByCol(colname);
                if (pem2 == null) {
					// create it
					if (colname == null) {
						throw new SpeedoException("No column name defined for a foreign key: "
							+ "\n\tfield name=" + pe.getName()
							+ "\n\tpk column name=" + pem.getName());
					}
                    table.createPrimitiveElementMapping(pe, colname, pem.getType(), false);
                } else {
                    //it exists then assign the real primitive element to the
                    // the name def
                	logger.log(BasicLevel.INFO, "pas collection:\n pe.getParent():" + pe.getParent().toString()
                			+ "\n pe.getName():" + pe.getName() + "\n pem2.getLinkedMO():" + pem2.getLinkedMO());
                    ((Class) pe.getParent()).removeTypedElement(pe.getName());
                    pe = (PrimitiveElement) pem2.getLinkedMO();
                    crnd.setFieldName(pe.getName());
                }
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = crnd.getNameRef().getProjection().entrySet().iterator();
                while(it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    //find the field in the class used in by this name def
                    PrimitiveElement pe = (PrimitiveElement)
                            sc.jormclass.getTypedElement((String) me.getValue());
                    //find the pk column
                    RdbPrimitiveElementMapping pem = getPEMOfField(
						tsc, cm.getProjectName(), cm.getMapperName(),
						(String) tclassndproj.get(me.getKey()));
                    //find the fk column name
                    String colname = (String) pk2fk.get(pem.getName());
                    //check if the fk column name is already used
                    PrimitiveElementMapping pem2 =
                            table.getPrimitiveElementMappingByCol(colname);
                    if (pem2 == null) {
                        // create it
						if (colname == null) {
							throw new SpeedoException("No column name defined for a foreign key: "
								+ "\n\tfield name=" + pe.getName()
								+ "\n\tpk column name=" + pem.getName());
						}
                        table.createPrimitiveElementMapping(pe, colname, pem.getType(), false);
                    } else {
                        //it exists then assign the real primitive element to the
                        // the name def
                        ((Class) pe.getParent()).removeTypedElement(pe.getName());
                        pe = (PrimitiveElement) pem2.getLinkedMO();
                        me.setValue(pe.getName());
                    }
                }
            }
        }
    }

	private RdbPrimitiveElementMapping getPEMOfField(SpeedoClass tsc,
									  String projectName,
									  String mapperName,
									  String fieldName) throws PException {
		RdbClassMultiMapping trcm = (RdbClassMultiMapping)
			tsc.jormclass.getClassProject(projectName)
			.getMapping(mapperName).getClassMapping();
		RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
			trcm.getPrimitiveElementMapping(fieldName, true);
		if (pem == null) {
			if (tsc.superClassName == null) {
				throw new PException("No mapping found for the field '"
					+ fieldName + "' for the class '" + tsc.getFQName() + "'.");

			} else {
				//the mapping of this field is inherited from parent
				return getPEMOfField(
					tsc.getSpeedoClassFromContext(tsc.superClassName),
					projectName, mapperName, fieldName);
			}
		} else {
			return pem;
		}
	}


    /**
     * Mappes the identifier of a GenClassRef over sql columns. In according
     * to the speedo extensions a join table is used (default case) or not.
     *
     * The columms match to the foreign key of the souce class. The default
     * column names of the join table are the name of the hidden field of the
     * genclass.
     *
     * @param sc is the SpeedoClass hosting the reference
     * @param nd is the name def of the gen class identifier
     * @param sc is the Speedo meta object hosting the field which references
     * the generic class.
     * @param rgcm is the jorm meta object representing the mapping of the gen
     * class.
     * @throws org.objectweb.jorm.api.PException
     */
    private void creaGenClassRefIdNameDef(NameDef nd,
                                     SpeedoClass sc,
                                     RdbGenClassMapping rgcm)
            throws PException {
        MetaObject mo = nd.getParent();
        RdbTable table = rgcm.getRdbTable();
        GenClassRef gcr = (GenClassRef) mo;
        SpeedoField sf = (SpeedoField) sc.jdoField.get(gcr.getName());
        RdbClassMultiMapping srcm = (RdbClassMultiMapping)
                ((Class) gcr.getParent())
                .getClassProject(rgcm.getProjectName())
                .getMapping(rgcm.getMapperName()).getClassMapping();
        NameDef sclassnd = (NameDef) srcm.getIdentifierMapping().getLinkedMO();
        SpeedoExtension se = sf.getExtension(VENDOR_NAME, SOURCE_FK);
        Map pk2fk = (se == null
                ? null // default mapping
                : getPk2Fk(se.value));
        if (nd.isFieldName()) {
            PrimitiveElement pe = gcr.getHiddenField(nd.getFieldName());
            String colname = null;
            if (pk2fk != null) {
                RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                        srcm.getPrimitiveElementMapping(sclassnd.getFieldName(), true);
                colname = (String) pk2fk.get(pem.getName());
				if (colname == null) {
					logger.log(BasicLevel.WARN, "No " + SOURCE_FK
						+ " defined for the primary column " + pem.getName()
						+ " of the field " + gcr.getName()
						+ " of the class " + sc.getFQName()
						+ ", use the default mapping");
				}
            }
			if (colname == null) {
				// default mapping
				colname = pe.getName();
			}
			//check if the column name is already used
            PrimitiveElementMapping pem2 =
                    table.getPrimitiveElementMappingByCol(colname);
            if (pem2 == null) {
				// create it
                table.createPrimitiveElementMapping(pe, colname);
            } else {
                //it exists then assign the real primitive element to the
                // the name def
            	((GenClassRef) pe.getParent()).removeTypedElement(pe.getName());
                pe = (PrimitiveElement) pem2.getLinkedMO();
                nd.setFieldName(pe.getName());
            }
        } else {
            Map sclassndproj = sclassnd.getNameRef().getProjection();
            Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry me = (Map.Entry) it.next();
                PrimitiveElement pe =
                        gcr.getHiddenField((String) me.getValue());
                String colname = null;
                if (pk2fk != null) {
                    RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                            srcm.getPrimitiveElementMapping(
                            		(String) sclassndproj.get(me.getKey()), true);
                    colname = (String) pk2fk.get(pem.getName());
					if (colname == null) {
						logger.log(BasicLevel.WARN, "No " + SOURCE_FK
							+ " defined for the primary column " + pem.getName()
							+ " of the field " + gcr.getName()
							+ " of the class " + sc.getFQName()
							+ ", use the default mapping");
					}
                }
				if (colname == null) {
					// default mapping
					colname = pe.getName();
				}
				//check if the column name is already used
                PrimitiveElementMapping pem2 =
                        table.getPrimitiveElementMappingByCol(colname);
                if (pem2 == null) {
                    // create it
                    table.createPrimitiveElementMapping(pe, colname);
                } else {
                    //it exists then assign the real primitive element to the
                    // the name def
                    ((GenClassRef) pe.getParent()).removeTypedElement(pe.getName());
                    pe = (PrimitiveElement) pem2.getLinkedMO();
                    me.setValue(pe.getName());
                }
            }
        }
    }


    /**
     * Mappes the element of a GenClassRef over sql columns. In according
     * to the speedo extensions a join table is used (default case) or not.
     *
     * The element of the gen class is a Gen class (not supported) or
     * a classRef. In all case the parent of this element is the GenClassRef
     *
     * @param sc is the SpeedoClass hosting the reference
     * @param nd is the name def of the gen class element
     * @param sc is the Speedo meta object hosting the field which references
     * the generic class.
     * @param rgcm is the jorm meta object representing the mapping of the gen
     * class.
     * @throws org.objectweb.jorm.api.PException
     */
    private void creaGenClassRefRefNameDef(NameDef nd,
                                     SpeedoClass sc,
                                     RdbGenClassMapping rgcm)
            throws PException {
        MetaObject mo = nd.getParent();
        RdbTable table = rgcm.getRdbTable();
        GenClassRef gcr = (GenClassRef) mo.getParent();
        ClassRef cr = (ClassRef) mo;
        SpeedoField sf = (SpeedoField) sc.jdoField.get(gcr.getName());
		SpeedoClass tsc = sc.getSpeedoClassFromContext(cr.getMOClass().getFQName());
        RdbClassMultiMapping trcm = (RdbClassMultiMapping)
                cr.getMOClass().getClassProject(rgcm.getProjectName())
                .getMapping(rgcm.getMapperName()).getClassMapping();
        NameDef tclassnd = (NameDef) trcm.getIdentifierMapping().getLinkedMO();
        if (isColocatedGCTable(sf)) {
            //The columms match to the primary key of the target class
            if (nd.isFieldName()) {
                PrimitiveElement pe = gcr.getHiddenField(nd.getFieldName());
                RdbPrimitiveElementMapping pem = getPEMOfField(tsc,
					rgcm.getProjectName(),
					rgcm.getMapperName(),
					tclassnd.getFieldName());
                //We must use the same column name
                table.createPrimitiveElementMapping(pe, pem.getName());
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe =
                            gcr.getHiddenField((String) me.getValue());
                    RdbPrimitiveElementMapping pem = getPEMOfField(tsc,
						rgcm.getProjectName(),
						rgcm.getMapperName(),
						(String) tclassndproj.get(me.getKey()));
                    //We must use the same column name
                    table.createPrimitiveElementMapping(pe, pem.getName());
                }
            }
        } else {
            //The columns are foreign key of the primary key of the
            // target class.
            SpeedoExtension se = sf.getExtension(VENDOR_NAME, TARGET_FK);
            Map pk2fk = (se == null
                    ? null // default mapping
                    : getPk2Fk(se.value));
            if (nd.isFieldName()) {
                PrimitiveElement pe =
                        gcr.getHiddenField(nd.getFieldName());
                String colname;
                if (pk2fk == null) {
                    // default mapping
                    colname = pe.getName();
                } else {
                    RdbPrimitiveElementMapping pem = getPEMOfField(tsc,
						rgcm.getProjectName(),
						rgcm.getMapperName(),
						tclassnd.getFieldName());
                    colname = (String) pk2fk.get(pem.getName());
                }
                table.createPrimitiveElementMapping(pe, colname);
            } else {
                Map tclassndproj = tclassnd.getNameRef().getProjection();
                Iterator it = nd.getNameRef().getProjection().entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry me = (Map.Entry) it.next();
                    PrimitiveElement pe =
                            gcr.getHiddenField((String) me.getValue());
                    String colname;
                    if (pk2fk == null) {
                        // default mapping
                        colname = pe.getName();
                    } else {
                        RdbPrimitiveElementMapping pem = getPEMOfField(tsc,
							rgcm.getProjectName(),
							rgcm.getMapperName(),
							(String) tclassndproj.get(me.getKey()));
                        colname = (String) pk2fk.get(pem.getName());
                    }
                    table.createPrimitiveElementMapping(pe, colname);
                }
            }
        }
    }

    /**
     * Calculates the main table name of a Class. First it looks in the meta
     * information to find if it is not already defined. If it is not defined
     * then the table name is built with the class name (not fully qualified).
     */
    private String getTableName(SpeedoClass sc) throws SpeedoException {
		SpeedoExtension se = sc.getExtensionByKey(SQL_NAME);
		if (se != null) {
			return se.value;
		}
		if (sc.superClassName != null) {
			se = sc.getExtensionByKey(INHERITANCE_MAPPING);
			if (se == null) {
				throw new SpeedoException("No inheritance mapping defined for the class "
						+ sc.getFQName() + " in the jdo file "
						+ sc.getJDOFileName());
			}
			if (INHERITANCE_FILTERED_MAPPING.equals(se.value)) {
				return getTableName(sc.getSpeedoClassFromContext(sc.superClassName));
			}
		}

		return sc.name;
    }

    private String getTableName(GenClassRef gcr) {
        String t = gcr.getGenClassId();
        int idx = t.lastIndexOf('.');
        if (idx != -1) {
            t = t.substring(idx + 1);
        }
        return t.replace(JormPathHelper.SEP, '_');
    }

    private String getColumnName(PrimitiveElement pe) {
        String c = pe.getName();
        int idx = c.lastIndexOf('.');
        if (idx != -1) {
            c = c.substring(idx +1);
        }
        return c;
    }

    private Map getPk2Fk(String fks) {
        Map res = new HashMap();
        StringTokenizer st = new StringTokenizer(fks, "=,:;/", false);
        String pk = null;
        while(st.hasMoreTokens()) {
            String tok = st.nextToken();
            tok = tok.trim();
            if (pk == null) {
                pk = tok;
            } else {
                res.put(pk, tok);
                pk = null;
            }
        }
        return res;
    }

    private Manager getManager(MetaObject mo) {
        MetaObject m = mo;
        while(m != null && !(m instanceof Manager)) {
            m = m.getParent();
        }
        return (Manager) m;
    }

    private boolean isColocatedGCTable(SpeedoField sf) {
        return sf.getExtension(VENDOR_NAME, JOIN_TABLE) == null
                &&  sf.getExtension(VENDOR_NAME, TARGET_FK) == null
                &&  sf.getExtension(VENDOR_NAME, SOURCE_FK) != null;
    }
}
