/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.mivisitor;

import org.objectweb.speedo.metadata.SpeedoPackage;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.metadata.SpeedoElement;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.List;
import java.util.Collections;
import java.util.ArrayList;

/**
 * This class checks if the specified speedo extensions in .jdo file are supported.
 * A warning (logging) is print for the unknown extensions.
 *
 * @author S.Chassande-Barrioz
 */
public class ExtensionVerifier
	extends AbstractMetaInfoVisitor
	implements SpeedoProperties {

	protected String getLoggerName() {
		return super.getLoggerName() + ".extensionverifier";
	}

	public void visitPackage(SpeedoPackage sp) throws SpeedoException {
		super.visitPackage(sp);
		StringBuffer sb = new StringBuffer();
		sb.append("package ");
		sb.append(sp.name);
		sb.append(" in the .jdo file ");
		sb.append(sp.jdoXMLDescriptor.xmlFile);
		checkSpeedoElement(sp, EXTENSION_PACKAGE_IDX, sb.toString());
	}

	public void visitClass(SpeedoClass sc) throws SpeedoException {
		super.visitClass(sc);
		StringBuffer sb = new StringBuffer();
		sb.append("class ");
		sb.append(sc.getFQName());
		sb.append(" in the .jdo file ");
		sb.append(sc.getJDOFileName());
		checkSpeedoElement(sc, EXTENSION_CLASS_IDX, sb.toString());
	}

	public void visitField(SpeedoField sf) throws SpeedoException {
		super.visitField(sf);
		StringBuffer sb = new StringBuffer();
		sb.append("field ");
		sb.append(sf.name);
		sb.append("of class ");
		sb.append(sf.jdoClass.getFQName());
		sb.append(" in the .jdo file ");
		sb.append(sf.jdoClass.getJDOFileName());
		checkSpeedoElement(sf, EXTENSION_FIELD_IDX, sb.toString());
		if (sf.jdoTuple != null) {
			if (sf.jdoTuple instanceof SpeedoMap) {
				sb.insert(0, "map ");
				checkSpeedoElement(sf.jdoTuple, EXTENSION_MAP_IDX, sb.toString());
			} else if (sf.jdoTuple instanceof SpeedoCollection) {
				sb.insert(0, "collection ");
				checkSpeedoElement(sf.jdoTuple, EXTENSION_COLLECTION_IDX, sb.toString());
			}
		}
	}

	private void checkSpeedoElement(SpeedoElement element,
									int nodeIdx,
									String nodeName) {
		List exts = getSpeedoExtension(element);
		String unexpected = getUnexpectedExtensionStr(exts, nodeIdx);
		if (unexpected != null) {
			logger.log(BasicLevel.WARN,
				"Unsupported speedo extension found for the "
				+ nodeName + unexpected);
		}
	}

	private String getUnexpectedExtensionStr(List exts, int node) {
		StringBuffer sb = new StringBuffer();
		String sep = "";
		boolean unexpectedFound = false;
		List supported = (List) SUPPORTED_EXTENSION.get(node);
		for(int i=0; i<exts.size(); i++) {
			SpeedoExtension se = (SpeedoExtension) exts.get(i);
			boolean uf = supported.isEmpty() || !supported.contains(se.key);
			if (uf) {
				sb.append(sep);
				sep = ", ";
				unexpectedFound = true;
				sb.append("(");
				sb.append(se.key);
				sb.append("=>");
				sb.append(se.value);
				sb.append(")");
			}
		}
		if (unexpectedFound) {
			sb.insert(0, "{");
			sb.append("}");
			return sb.toString();
		} else {
			return null;
		}
	}


	private List getSpeedoExtension(SpeedoElement element) {
		List exts = element.jdoExtension;
		if(exts.isEmpty()) {
			return Collections.EMPTY_LIST;
		}
		ArrayList result = new ArrayList(exts.size());
		for(int i=0; i<exts.size(); i++) {
			SpeedoExtension se = (SpeedoExtension) exts.get(i);
			if (VENDOR_NAME.equals(se.vendorName)) {
				result.add(se);
			}
		}
		return result;
	}
}
