/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.mivisitor;

import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.speedo.metadata.SpeedoArray;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.generation.enhancer.Util;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.asm.Type;

/**
 * Add the reverse extensions on the forgetten side. In addition the mapping
 * is added too.
 *
 * @author S.Chassande-Barrioz
 */
public class ReverseFieldAdder extends AbstractMetaInfoVisitor {

    public ReverseFieldAdder() {
    }

    public ReverseFieldAdder(MetaInfoVisitor mim) {
        super(mim);
    }

    protected String getLoggerName() {
        return LOGGER_NAME + ".reverseField";
    }

    public void visitExtension(SpeedoExtension se) throws SpeedoException {
        debug = logger.isLoggable(BasicLevel.DEBUG);
        if (!SpeedoProperties.VENDOR_NAME.equals(se.vendorName)
                || !SpeedoProperties.REVERSE_FIELD.equals(se.key)) {
            super.visitExtension(se);
            return;
        }

        //check that the extension REVERSE_FIELD is specified for a SpeedoField
        if (!(se.jdoElement instanceof SpeedoField)) {
            SpeedoClass sc = getSpeedoClass(se.jdoElement);
            logger.log(BasicLevel.ERROR, "You have specified the reverse field ('"
                    + se.value + "') in other thing than a field tag ('"
                    + se.jdoElement.toString() + "')" +
                    (sc == null
                    ? ""
                    : " in class '" + sc.getFQName() + "' of the descriptor '"
                    + sc.jdoPackage.jdoXMLDescriptor.xmlFile));
            super.visitExtension(se);
            return;
        }

        SpeedoField sf = (SpeedoField) se.jdoElement;
        SpeedoField reverse = getReverseField(sf,se.value);
        if (reverse == null) {
            throwError("No reverse field '" + se.value + "' found for the", sf);
            super.visitExtension(se);
            return;
        }
        SpeedoExtension _se = reverse.getExtension(
                SpeedoProperties.VENDOR_NAME, SpeedoProperties.REVERSE_FIELD);
        if (_se == null) {
            //define the extension REVERSE_FIELD on the reverse field
            _se = new SpeedoExtension(SpeedoProperties.VENDOR_NAME,
                    SpeedoProperties.REVERSE_FIELD,
                    sf.name, reverse);
            logger.log(BasicLevel.DEBUG, "add the extension("
                    + SpeedoProperties.REVERSE_FIELD + ", " + sf.name
                    +") on the field '" + reverse.name + "'");
            reverse.addExtension(_se);
        }

        // compute relationType
        if (sf.jdoTuple != null) {
            if (reverse.jdoTuple != null) {
                sf.relationType = SpeedoField.MANY_MANY_RELATION;
                reverse.relationType = SpeedoField.MANY_MANY_RELATION;
            } else {
                sf.relationType = SpeedoField.ONE_MANY_RELATION;
                reverse.relationType = SpeedoField.MANY_ONE_RELATION;
            }
        } else {
            if (reverse.jdoTuple != null) {
                sf.relationType = SpeedoField.MANY_ONE_RELATION;
                reverse.relationType = SpeedoField.ONE_MANY_RELATION;
            } else {
                sf.relationType = SpeedoField.ONE_ONE_RELATION;
                reverse.relationType = SpeedoField.ONE_ONE_RELATION;
            }
        }

        checkExtension(sf, se, reverse,
                SpeedoProperties.TARGET_FK, SpeedoProperties.SOURCE_FK);
        checkExtension(sf, se, reverse,
                SpeedoProperties.SOURCE_FK, SpeedoProperties.TARGET_FK);
        checkExtension(sf, se, reverse,
                SpeedoProperties.JOIN_TABLE, SpeedoProperties.JOIN_TABLE);
        super.visitExtension(se);
    }

    /**
     * Check if a SpeedoField has a specified SpeedoExtension ('ext1'). If yes
     * the extension 'ext2' is add on the reverse field with the same value.
     * If the meta data is malformed then the visit continue to the next
     * MetaDatavisitor.
     * @param sf is the SpeedoField containing the SpeedoExtension 'ext1'
     * @param se the parameter to the visitor
     * @param reverse the SpeedoField which must contains at the end the options
     * ext2 if the extension ext1 is defined on the field 'sf'
     * @param ext1 a name of a SpeedoExtension (@see SpeedoProperties)
     * @param ext2 a name of a SpeedoExtension (@see SpeedoProperties)
     */
    private SpeedoExtension checkExtension(SpeedoField sf,
                                SpeedoExtension se,
                                SpeedoField reverse,
                                String ext1,
                                String ext2) throws SpeedoException {
        SpeedoExtension _se = sf.getExtension(SpeedoProperties.VENDOR_NAME, ext1);
        if (_se != null) {
            SpeedoExtension __se = reverse.getExtension(SpeedoProperties.VENDOR_NAME, ext2);
            if (__se == null) {
                //define the extension SOURCE_FK on the reverse field
                __se = new SpeedoExtension(SpeedoProperties.VENDOR_NAME, ext2, _se.value, reverse);
                logger.log(BasicLevel.DEBUG, "add the extension(" + ext2
                        + ", " + _se.value +") on the field '"
                        + reverse.name + "'");
                reverse.addExtension(__se);
            } else if (!__se.value.equals(_se.value)) {
                throwError("A bad mapping has been specified on the reverse field '"
                        + se.value + "' for the ", sf);
                super.visitExtension(se);
                return _se;
            } else if (debug) {
                SpeedoClass sc1 = getSpeedoClass(sf);
                SpeedoClass sc2 = getSpeedoClass(reverse);
                logger.log(BasicLevel.DEBUG, "You could remove one of the following extensions:"
                        + "\n- .jdo: '" + sc1.jdoPackage.jdoXMLDescriptor.xmlFile
                        + "', class: '" + sc1.getFQName()
                        + "', field: '" + sf.name
                        + "',  extension: '" + ext1 + "'"

                        + "\n- .jdo: '" + sc2.jdoPackage.jdoXMLDescriptor.xmlFile
                        + "', class: '" + sc2.getFQName()
                        + "', field: '" + reverse.name
                        + "',  extension: '" + ext2 + "'"
                        );
            }
        }
        return _se;
    }

    /**
     * Retrieves the SpeedoField instance which is the reverse of a given
     * SpeedoField. The name of reverse field is specified by 'rfn'.
     * @param sf is the SpeedoField which the reverse SpeedoField must be retrieved
     * @param rfn the name of the reverse field
     */
    public static SpeedoField getReverseField(SpeedoField sf, String rfn)  throws SpeedoException {
        String className = null;
        if (sf.jdoTuple != null) {
            if (sf.jdoTuple instanceof SpeedoCollection) {
                className = (String) ((SpeedoCollection) sf.jdoTuple).elementType;
            } else if (sf.jdoTuple instanceof SpeedoMap) {
                className = (String) ((SpeedoMap) sf.jdoTuple).valueType;
            } else if (sf.jdoTuple instanceof SpeedoArray) {
                className = Util.type(Type.getType(sf.desc));
            }
        } else {
            className = Util.type(Type.getType(sf.desc));
        }
        SpeedoClass sc = sf.jdoClass.jdoPackage
                .jdoXMLDescriptor.getSpeedoClass(className, true);
        if (sc == null) {
            sc = sf.jdoClass.jdoPackage.jdoXMLDescriptor.getSpeedoClass(
                    getSpeedoClass(sf).jdoPackage.name + "." + className, true);
        }
        if (sc == null) {
            throwError("No persistent class '" + className + "' found for managing the ", sf);
            return null;
        }
        return (SpeedoField) sc.jdoField.get(rfn);
    }

    /**
     * Log a pretty error message with the context (field, class, xml desc)
     * @param msg the additionnal message
     * @param sf the field which is managed
     */
    public static void throwError(String msg, SpeedoField sf) throws SpeedoException {
        throw new SpeedoException(msg + " field '" + sf.name
                + "' of the class '" + sf.jdoClass.getFQName()
                + "' in the descriptor '"
                + sf.jdoClass.jdoPackage.jdoXMLDescriptor.xmlFile + "'");
    }


}
