/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.generation.parser;

import org.apache.xerces.parsers.DOMParser;
import org.objectweb.jorm.xml2mi.lib.DTDResolver;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.generation.api.SpeedoXMLError;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.generation.lib.AbstractGeneratorComponent;
import org.objectweb.speedo.metadata.SpeedoArray;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoCollection;
import org.objectweb.speedo.metadata.SpeedoElement;
import org.objectweb.speedo.metadata.SpeedoExtension;
import org.objectweb.speedo.metadata.SpeedoFetchGroup;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoIdentity;
import org.objectweb.speedo.metadata.SpeedoMap;
import org.objectweb.speedo.metadata.SpeedoModifier;
import org.objectweb.speedo.metadata.SpeedoNullValue;
import org.objectweb.speedo.metadata.SpeedoPackage;
import org.objectweb.speedo.metadata.SpeedoTuple;
import org.objectweb.speedo.metadata.SpeedoVersion;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.util.monolog.api.BasicLevel;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.Iterator;
import java.io.File;

/**
 * @author S.Chassande-Barrioz
 */
public class JdoParser extends AbstractGeneratorComponent {

    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".generation.parser";

    /**
     * The parser instance is reused at each process method call
     */
    private DOMParser parser = null;


    // IMPLEMENTATION OF THE GeneratorComponent INTERFACE //
    //----------------------------------------------------//

    public boolean init() throws SpeedoException {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
        if (scp.jdo.isEmpty())
            return false;
        DTDResolver resolver = new DTDResolver(scp.dtdLocations, logger);
        //  Create a Xerces DOM Parser
        parser = new DOMParser();
        try {
            parser.setProperty(
                    "http://apache.org/xml/properties/internal/entity-resolver",
                    resolver);
        } catch (Exception e) {
            throw new SpeedoException(
                    "Impossible to assign the entity-resolver property", e);
        }
	    return true;
    }

    public void process() throws SpeedoException {
        if (scp.jdo.isEmpty())
            return;
        logger.log(BasicLevel.DEBUG, scp.jdo.size() + " files to parse");
        for (Iterator it = scp.jdo.iterator(); it.hasNext();) {
            String xmlFileName = (String) it.next();
            logger.log(BasicLevel.DEBUG, "Parse " + xmlFileName);
            scp.getXmldescriptor().put(xmlFileName, createObjectModel(xmlFileName));
        }

        for (Iterator itDesc = scp.getXmldescriptor().values().iterator(); itDesc.hasNext();) {
            SpeedoXMLDescriptor desc = (SpeedoXMLDescriptor) itDesc.next();
            for (Iterator itPack = desc.jdoPackage.values().iterator(); itPack.hasNext();) {
                SpeedoPackage sp = (SpeedoPackage) itPack.next();
                for (Iterator itclass = sp.jdoClass.values().iterator(); itclass.hasNext();) {
                    SpeedoClass jdoclass = (SpeedoClass) itclass.next();
                    jdoclass.computeFieldNumbers();
                }
            }
        }
    }


    // PRIVATE METHODS //
    //-----------------//

    /**
     * This method constructs an object descriptor for a XML file ".jdo".
     * It returns the new object descriptor.
     * If the file isn't well formed, a SpeedoXMLError is thrown.
     * @param xmlFile file name.
     * @return object descriptor for this file
     * @exception SpeedoException if the XML file doesn't follow jdo dtd.
     */
    private SpeedoXMLDescriptor createObjectModel(String xmlFile)
            throws SpeedoException {
        try {
            // parse the document
            parser.parse(scp.jdoDir + File.separator + xmlFile);
            Document document = parser.getDocument();

            //create a descriptor for the document
            SpeedoXMLDescriptor desc = new SpeedoXMLDescriptor(scp.smi);
            desc.xmlFile = xmlFile;
            //fill and return the file descriptor
            return (SpeedoXMLDescriptor) treatDocument(document, desc);
        } catch (SpeedoException e) {
            throw e;
        } catch (Exception e) {
            throw new SpeedoException("Error with the construction of file "
                    + scp.jdoDir + File.separator + xmlFile, e);
        }
    }

    /**
     * This methods creates the object descriptor for describing a tag of the XML file.
     * It returns the descriptor for the root tag.
     * It throws an exception if an attribute is missing.
     * @param node node which will be described.
     * @param o default descriptor to return.
     * @return node descriptor.
     * @exception SpeedoException if a compulsory attribute is missing for
     * the current tag.
     */
    private Object treatNode(Node node, Object o) throws SpeedoException {
        // type of node
        int type = node.getNodeType();
        if (type != Node.ELEMENT_NODE)
            return o;
        //tag name
        String tag = node.getNodeName();
        Node n = null;
        logger.log(BasicLevel.DEBUG, "Parse tag: " + tag);

        //tag jdo
        if (tag.equals("jdo")) {
            //there is no descriptor to create
            return o;

        } else if (tag.equals("package")) { //tag package
            SpeedoPackage p = new SpeedoPackage();

            //attribute name (complusory)
            n = node.getAttributes().getNamedItem("name");
            if (n == null)
                throw new SpeedoXMLError("Attribute name for tag package requested.");
            p.name = n.getNodeValue();

            logger.log(BasicLevel.DEBUG, "New package: name= " + p.name);
            return p;

        } else if (tag.equals("class")) { //tag class
            SpeedoClass c = new SpeedoClass();

            //attribute name (compulsory)
            n = node.getAttributes().getNamedItem("name");
            if (n == null)
                throw new SpeedoXMLError("Attribute name for tag class requested.");
            c.name = n.getNodeValue();

            //attribute identity-type
            n = node.getAttributes().getNamedItem("identity-type");
            if (n != null)
                c.identityType = SpeedoIdentity.toByte(n.getNodeValue());
			  
            //attribute object-id
            n = node.getAttributes().getNamedItem("objectid-class");
            if (n != null) {
                c.objectidClass = n.getNodeValue();
                c.identityType = SpeedoIdentity.USER_ID;
            }

            //attribute requires-extent
            n = node.getAttributes().getNamedItem("requires-extent");
            if (n != null)
                c.requiresExtent = Boolean.valueOf(n.getNodeValue()).booleanValue();

			//attribute detachable
			n = node.getAttributes().getNamedItem("detachable");
			if (n != null)
				c.isDetachable = Boolean.valueOf(n.getNodeValue()).booleanValue();
				  
            //attribute persistence-capable-superclass
            n = node.getAttributes().getNamedItem("persistence-capable-superclass");
            if (n != null)
                c.superClassName = n.getNodeValue();

            logger.log(BasicLevel.DEBUG, "New class: "
                    + "name=" + c.name
                    + "  / id = " + c.identityType
                    + "  / idClass = " + c.objectidClass
                    + "  / requiresExtent = " + c.requiresExtent
                    + "  / detachable = " + c.isDetachable
            );
            return c;

        } else if (tag.equals("field")) { //tag field
            SpeedoField f = new SpeedoField();

            //attribut name (compulsory)
            n = node.getAttributes().getNamedItem("name");
            if (n == null)
                throw new SpeedoXMLError("Attribute name for tag field requested.");
            String name = n.getNodeValue();
            if(name.indexOf(".") != -1){
            	
            }
            f.name = n.getNodeValue();

            //attribute persistence-modifier
            n = node.getAttributes().getNamedItem("persistence-modifier");
            if (n != null)
                f.persistenceModifier = SpeedoModifier.toByte(n.getNodeValue());

            //attribute primary-key
            n = node.getAttributes().getNamedItem("primary-key");
            if (n != null)
                f.primaryKey = Boolean.valueOf(n.getNodeValue()).booleanValue();

            //attribute null-value
            n = node.getAttributes().getNamedItem("null-value");
            if (n != null) {
                f.nullValue = SpeedoNullValue.toByte(n.getNodeValue());
			}

            //attribute default-fetch-group
            n = node.getAttributes().getNamedItem("default-fetch-group");
            if (n != null)
                f.defaultFetchGroup = Boolean.valueOf(n.getNodeValue()).booleanValue();

			//attribute fetch-group
			n = node.getAttributes().getNamedItem("fetch-group");
			if (n != null)
				f.fetchGroup = n.getNodeValue();
				
			//attribute depth
			n = node.getAttributes().getNamedItem("depth");
			if (n != null)
				f.depth = Integer.valueOf(n.getNodeValue()).intValue();

            //attribute embbedded
            n = node.getAttributes().getNamedItem("embedded");
            if (n != null)
                f.embedded = Boolean.valueOf(n.getNodeValue()).booleanValue();

            logger.log(BasicLevel.DEBUG, "New field: "
                    + "name=" + f.name
                    + "  / persistenceModifier = " + f.persistenceModifier
                    + "  / primaryKey = " + f.primaryKey
                    + "  / nullValue = " + f.nullValue
                    + "  / defaultFetchGroup = " + f.defaultFetchGroup
                    + "  / embedded = " + f.embedded
					+ "  / depth = " + f.depth
            );
            return f;

        } else if (tag.equals("collection")) { //tag collection
            SpeedoCollection co = new SpeedoCollection();

            //attribute element-type
            n = node.getAttributes().getNamedItem("element-type");
            if (n != null)
                co.elementType = n.getNodeValue();

            //attribute embedded-element
            n = node.getAttributes().getNamedItem("embedded-element");
            if (n != null)
                co.embeddedElement = Boolean.valueOf(n.getNodeValue()).booleanValue();

            logger.log(BasicLevel.DEBUG, "New collection: "
                    + "elementType=" + co.elementType
                    + "  / embeddedElement = " + co.embeddedElement
            );
            return co;

        } else if (tag.equals("map")) { //tag map
            SpeedoMap m = new SpeedoMap();

            //attribut key-type
            n = node.getAttributes().getNamedItem("key-type");
            if (n != null)
                m.keyType = n.getNodeValue();

            //attribut embedded-key
            n = node.getAttributes().getNamedItem("embedded-key");
            if (n != null)
                m.embeddedKey = Boolean.valueOf(n.getNodeValue()).booleanValue();

            //attribute value-type
            n = node.getAttributes().getNamedItem("value-type");
            if (n != null)
                m.valueType = n.getNodeValue();

            //attribute embedded-value
            n = node.getAttributes().getNamedItem("embedded-value");
            if (n != null)
                m.embeddedValue = Boolean.valueOf(n.getNodeValue()).booleanValue();

            logger.log(BasicLevel.DEBUG, "New map: "
                    + "keyType=" + m.keyType
                    + "  / embeddedKey = " + m.embeddedKey
                    + "  / valueType = " + m.valueType
                    + "  / embeddedValue = " + m.embeddedValue
            );
            return m;

        } else if (tag.equals("array")) { //tag array
            SpeedoArray a = new SpeedoArray();

            //attribute embedded-element
            n = node.getAttributes().getNamedItem("embedded-element");
            if (n != null)
                a.embeddedElement = Boolean.valueOf(n.getNodeValue()).booleanValue();

            logger.log(BasicLevel.DEBUG, "New array: "
                    + "embeddedElement=" + a.embeddedElement
            );
            return a;
		} else if (tag.equals("version")) { //tag version
		   SpeedoVersion v = new SpeedoVersion();

			//attribute embedded-element
			n = node.getAttributes().getNamedItem("strategy");
			v.strategy = SpeedoVersion.toByte(n.getNodeValue());
			
			logger.log(BasicLevel.DEBUG, "New version: "
					   + "strategy=" + v.strategy);   
			return v;
        } else if (tag.equals("extension")) { //tag extension
            SpeedoExtension e = new SpeedoExtension();

            //attribute vendor-name
            n = node.getAttributes().getNamedItem("vendor-name");
            e.vendorName = n.getNodeValue();

            //attribute key
            n = node.getAttributes().getNamedItem("key");
            if (n != null)
                e.key = n.getNodeValue();

            //attribute value
            n = node.getAttributes().getNamedItem("value");
            if (n != null)
                e.value = n.getNodeValue();

            return e;
        } else if  (tag.equals("fetch-group")) { //tag fetch-group
        	SpeedoFetchGroup fg = new SpeedoFetchGroup();
        	
        	//attribute name
			n = node.getAttributes().getNamedItem("name");
			fg.name = n.getNodeValue();
			
			//attribute post-load
			n = node.getAttributes().getNamedItem("post-load");
			if (n != null){
				fg.postLoad = Boolean.valueOf(n.getNodeValue()).booleanValue();
			}
			else{
				if(fg.name.equals("default"))
					fg.postLoad = true;
				else
					fg.postLoad = false;
			}
			return fg;
		} else
            return o;
    }

    /**
     * Create the XML file descriptor.
     * It throws an exception if the XML file doesn't follow the jdo dtd.
     * @param node root node of the descriptor which will be return.
     * @param o default descriptor to return.
     * @return descriptor.
     */
    private Object treatDocument(Node node, Object o) throws SpeedoException {
        //o is descriptor of the node root
        o = treatNode(node, o);
        //list of child node
        NodeList children = node.getChildNodes();

        //if the node has a child
        if (children != null) {
            //for each child
            for (int i = 0; i < children.getLength(); i++) {
                //create a descriptor for the child
                Object fils = treatDocument(children.item(i), o);
                //add this descriptor to the parent descriptor
                o = add(o, fils, children.item(i).getNodeName());
            }
        }
        return o;
    }

    /**
     * This methods adds an element into a descriptor.
     * @param graph descriptor.
     * @param newElem object to add.
     * @param tag object type.
     * @return the new descriptor.
     * @exception SpeedoException if a field is described twice.
     */
    private Object add(Object graph, Object newElem, String tag)
            throws SpeedoException {

        if (tag.equals("package")) { //the new element is a package
            logger.log(BasicLevel.DEBUG, "add package '" + ((SpeedoPackage) newElem).name + "'");
            ((SpeedoXMLDescriptor) graph).add(newElem, true, logger);

        } else if (tag.equals("class")) { //the new element is a class
            logger.log(BasicLevel.DEBUG, "add class '" + ((SpeedoClass) newElem).name + "'");
            ((SpeedoPackage) graph).add(newElem, true, logger);

        } else if (tag.equals("field")) { //the new element is a field
        	if(graph instanceof SpeedoClass){
            	if (logger.isLoggable(BasicLevel.DEBUG)) {
                	logger.log(BasicLevel.DEBUG,
                    	    "add field '" + ((SpeedoField) newElem).name
                        	+ "' into the class '" + ((SpeedoClass) graph).name +"'");
            	}
            	((SpeedoClass) graph).add(newElem, true, logger);
        	}
        	else{
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
						"add field '" + ((SpeedoField) newElem).name
						+ "' into the fetchgroup '" + ((SpeedoFetchGroup) graph).name +"'");
				}
				((SpeedoFetchGroup) graph).addField(newElem);
        	}
		} else if (tag.equals("version")) { //the new element is a version
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						"add version into the class '" + ((SpeedoClass) graph).name +"'");
			}
			((SpeedoClass) graph).addVersion(newElem);
        } else if (tag.equals("collection")) { //the new element is a collection
            logger.log(BasicLevel.DEBUG,
                    "add collection: '" + ((SpeedoField) graph).name + "'");
            ((SpeedoField) graph).jdoTuple = (SpeedoTuple) newElem;
            ((SpeedoTuple) newElem).jdoField = (SpeedoField) graph;

        } else if (tag.equals("map")) { //the new element is a map
            logger.log(BasicLevel.DEBUG,
                    "add map: '" + ((SpeedoField) graph).name + "'");
            ((SpeedoField) graph).jdoTuple = (SpeedoTuple) newElem;
            ((SpeedoTuple) newElem).jdoField = (SpeedoField) graph;

        } else if (tag.equals("array")) { //the new element is an array
            logger.log(BasicLevel.DEBUG,
                    "add array: '" + ((SpeedoField) graph).name + "'");
            ((SpeedoField) graph).jdoTuple = (SpeedoTuple) newElem;
            ((SpeedoTuple) newElem).jdoField = (SpeedoField) graph;

        } else if (tag.equals("extension")) { //the new element is an extension
            logger.log(BasicLevel.DEBUG, "add extension");
            ((SpeedoElement) graph).addExtension((SpeedoExtension) newElem);
            ((SpeedoExtension) newElem).jdoElement = (SpeedoElement) graph;
            
		} else if (tag.equals("fetch-group")) { //the new element is a fetch-group
			if(graph instanceof SpeedoClass){
				if (logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG,
						"add fetchgroup '" +  ((SpeedoFetchGroup) newElem).name +
						"' into the class '" + ((SpeedoClass) graph).name +"'");
				}
				((SpeedoClass) graph).addFetchGroup(newElem);
			}
			else{
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
						"add fetchgroup '" +  ((SpeedoFetchGroup) newElem).name +
						"' into the class '" + ((SpeedoClass) graph).name +"'");
				}
				((SpeedoFetchGroup) graph).addFetchGroup(newElem);
			}
        }
        return graph;
    }
}
