/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.metadata;

import org.objectweb.asm.Constants;
import org.objectweb.asm.Type;
import org.objectweb.speedo.generation.enhancer.Util;

/**
 * This class corresponds to the description of persistent capable fields.
 * @author	S.Chassande-Barrioz
 */
public class SpeedoField extends SpeedoElement {

    public final static byte NO_RELATION = 0;
    public final static byte ONE_ONE_RELATION = 1;
    public final static byte ONE_MANY_RELATION = 2;
    public final static byte MANY_ONE_RELATION = 3;
    public final static byte MANY_MANY_RELATION = 4;

    /**
     * name of the persistent field
     */
    public String name;

    /**
     * modifier of the field (public | protected | private)
     */
    public int access;

    /**
     * Type of the persistent field
     */
    public String desc;

    /**
     * is the index of the persistent field. This field is computed through a 
     * MI visitor.
     */
    public int number = -1;

    /**
     * Attribute persistence-modifier in the XML file.
     */
    public byte persistenceModifier = SpeedoModifier.missing;

    /**
     * Boolean indicating if the field is a primary key.
     */
    public boolean primaryKey = false;

    /**
     * Attribute null-value in the XML file.
     */
    public byte nullValue = SpeedoNullValue.NONE;

    /**
     * This boolean indicates if the field is in the default fetch group.
     * (not managed)
     */
    public boolean defaultFetchGroup = true;

	/**
 	 * Recursive fetchgroups are controlled by the depth attribute. A depth of 0 (the default)
 	 * will fetch the whole graph of instances reachable from this field. 
 	 */
	public int depth = 0;
	
	public String fetchGroup;
	
    /**
     * Attribute embedded of the XML file. (not managed)
     */
    public boolean embedded = true;

    /**
     * Type of the relation containinig the persistent field
     * @see NO_RELATION, ONE_ONE_RELATION, ONE_MANY_RELATION, MANY_ONE_RELATION,
     * MANY_MANY_RELATION
     */
    public byte relationType = NO_RELATION;

    /**
     * Type of the tuple represented by this SpeedoField if it is a tuple
     */
    public SpeedoTuple jdoTuple;

    /**
     * Description of the class containing the field.
     */
    public SpeedoClass jdoClass;

	
    /**
     * Transforms a SpeedoField into a String.
     * @return the String corresponding to the SpeedoField.
     */
    public String toString() {
        String s = "\n field name : " + name + ", persistenceModifier : "
                + persistenceModifier + ", primaryKey : " + primaryKey
                + ", nullValue : " + nullValue + ", defaultFetchGroup : "
                + defaultFetchGroup + ", embedded : " + embedded 
				+ ", fetchGroup : " + fetchGroup + ", depth : " + depth;
        if (jdoTuple != null)
            s = s + jdoTuple.toString();
        return s;
    }

    /**
     * Returns the field's signature with a public modifier.
     * @return the String containing the field' signature with a public modifier.
     */
    public String publicSignature() {
        int publicAccess = access;
        publicAccess &= ~(Constants.ACC_PRIVATE | Constants.ACC_PROTECTED);
        publicAccess |= Constants.ACC_PUBLIC;
        return Util.modifier(publicAccess) + type() + " " + name;
    }

    /**
     * Returns the field's signature with a private modifier.
     * @return the String containing the field' signature with a private modifier.
     */
    public String privateSignature() {
        int privateAccess = access;
        privateAccess &= ~(Constants.ACC_PUBLIC | Constants.ACC_PROTECTED);
        privateAccess |= Constants.ACC_PRIVATE;
        return Util.modifier(privateAccess) + type() + " " + name;
    }

    /**
     * Returns the field's modifier
     * @return the String representing the modifier (<code>public</code>,
     * <code>private</code>, <code>protected</code>)
     */
    public String modifier() {
        return Util.modifier(access);
    }

    /**
     * Returns the type of the field.
     * @return type of the field.
     */
    public String type() {
        return Util.type(Type.getType(desc));
    }
}
