/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.mim.api;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pm.api.ProxyManagerFactory;
import org.objectweb.speedo.mim.api.SpeedoAccessor;

import javax.jdo.spi.PersistenceCapable;

import java.util.Collection;
import java.util.Map;

/**
 * This interface what is a Proxy in Speedo. A proxy a PersistentCapable
 * instance because it is the user object. It is the jorm PBinding because it is
 * unique. There are several fields: <UL>
 * <LI>The field 'isActive' indicates if the proxy is active (linked to a ProxyManager)</LI>
 * <LI>The fields 'pmf' represents the data source where
 * the current object is persistent.</LI>
 * <LI>The fields 'tpm' is the perseus TransactionalPersistenceManager managing
 * this instance.</LI>
 * </UL>
 *
 * @author	S.Chassande-Barrioz
 */
public interface SpeedoProxy
        extends PersistenceCapable, PBinding, ReplaceableCacheEntry {

    /**
     * Tests whether the Proxy infrastructure has already been set.
     * A proxy has a correct PClassMapping linked to a PBinder,
     * a PName obtained thanks to this PClassMapping.
     *
     * @return true if this object is a Proxy.
     */
    boolean jdoIsActive();

    /**
     * Sets the new value of the proxy flags
     *
     * @see #jdoIsActive()
     * @param newvalue the new value of the proxy flag
     */
    void jdoIsActive(boolean newvalue);
	
	/**
	 * The versioning strategy.
	 * @see SpeeedoVersion.TIMESTAMP, STATE_COMPARISON, VERSION_NUMBER, NO_VERSION
	 */
	byte jdoGetVersioningStrategy();
		
    /**
     * It retrieves the SpeedoAccessor instance used in the current context.
     * If the proxy is not active then the reference state is returned.
     * otherwise a $classNameFields is returned
     */
    SpeedoAccessor getSpeedoAccessor();

    SpeedoAccessor getReferenceAccessor();

    void setReferenceAccessor(SpeedoAccessor refAcc);

	SpeedoAccessor createAccessor();

	void copyAccessor(SpeedoAccessor src, SpeedoAccessor dest);

	/**
	 * Set the encoded PName of the detach copy to allow serialization.
	 * @param encodedPName : the encoded PName of the persistent object
	 */
	void setEncodedPName(Object encodedPName);
	
	/**
	 * @return the encoded PName of the detached copy
	 */
	Object getEncodedPName();
	
    /**
     * @return the ProxyManagerFactory which represents the data support
     * inside which the proxy is persistent.
     */
	ProxyManagerFactory jdoGetProxyManagerFactory();

    /**
     * It assignes a ProxyManagerFactory
     * @param pmf is the ProxyManagerFactory which represents the data
     * support inside which the proxy is persistent.
     */
    void jdoSetProxyManagerFactory(ProxyManagerFactory pmf);

    /**
     * @return the TransactionalPersistenceManager which manage the
     * concurrency, loading, ... of the proxy.
     */
    TransactionalPersistenceManager jdoGetTransactionalPersistenceManager();

    /**
     * It assignes a TransactionalPersistenceManager
     * @param tpm is the TransactionalPersistenceManager which manage the
     * concurrency, loading, ... of the proxy.
     */
    void jdoSetTransactionalPersistenceManager(TransactionalPersistenceManager tpm);

	void jdoMakePersistent(ProxyManager pm, SpeedoAccessor sa);

    void jdoDeletePersistent(ProxyManager pm, SpeedoAccessor sa);

	/**
	 * @return the detached copy of this SpeedoProxy.
	 */
	SpeedoProxy jdoDetachCopy(ProxyManager pm, Map map, Object clone, Collection fgHints);
	
	
	/**
	 * Attach the detached clone to the datastore.
	 */
	void jdoAttachCopy(ProxyManager pm, Map map, Object clone, SpeedoAccessor sa, boolean makeTransactional);
	
	/**
	 * Refresh the fields to be fetched according to the active fetch groups  
	 */
	void jdoRefresh(ProxyManager pm, Map map, Collection fgHints);
	
    /**
     * @return return the mapper name by which the instance is persistent.
     */
    String jdoGetProjectName();
    
    /**
     * @return return if the class has been declared detachable in the jdo file
     */
    boolean jdoIsDetachable();

    /**
     * @return The file name of the jdo descriptor. This file must be availlable
     * in the class at runtime.
     */
    String jdoGetJdoFileName();

    /**
     * Retrieves the hints needed to build the PName of the persistent object
     * during the export action.
     */
    Object jdoGetPNameHints();

    /**
     * Adds an element in the gen class field whose number is given. This method
     * must call the speedoAdd method of the SpeedoGenClassCoherence interface,
     * or the Collection.add method (if the field contains a collection that
     * does implement the previous interface). If the field is null, a new
     * (speedo) collection must be created first.
     */
    void speedoAdd (Object elem, int gcFieldNumber);
    
    /**
     * @param fgName The name of the fetchgroup
     * @return true if fetchgroup identified by the name fgName has the postload attribute to true, else false
     */
    boolean fgIsPostload(String fgName);
    
    /**
     * @param fgName The name of the fetchgroup
     * @return the collection of fields to load when using the fetchgroup identified by the name fgName 
     */
    Collection fgGetFieldsToLoad(String fgName);
    
    /**
     * Same as fgGetFieldsToLoad(String fgName) but if onlyDirectRef is set to true, all the fields
     * of type a.b.c or a#element or a/fg are removed from the returned collection.
     * @param fgName
     * @param onlyDirectRef
     * @return
     */
    Collection fgGetFieldsToLoad(String fgName, boolean onlyDirectRef);
    
    /**
     * @param fgName The name of the fetchgroup
     * @return the collection of the nested fetchgroups when using the fetchgroup identified by the name fgName 
     */
    Collection fgGetNestedFetchGroups(String fgName);
    
    /**
     * @param fgName The name of the fetchgroup
     * @return true if this fetchgroup has been defined in the jdo file of this class.
     */
    boolean fgIsDefined(String fgName); 
	
}
