/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S. Chassande-Barrioz
 *
 */

package org.objectweb.speedo.query.lib;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.eval.lib.MedorEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBufferFactory;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.optim.api.QueryTransformer;
import org.objectweb.medor.optim.jorm.JormFilteredInheritance;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.optim.lib.QueryNodeReplacer;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.perseus.cache.api.UnFixProtocolException;
import org.objectweb.perseus.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateFilter;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.mapper.lib.DelegatePMapper;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.query.api.CompiledQuery;
import org.objectweb.speedo.query.api.QueryDefinition;
import org.objectweb.speedo.query.parser.ASTSpeedoQL;
import org.objectweb.speedo.query.parser.ParseException;
import org.objectweb.speedo.query.parser.SimpleNode;
import org.objectweb.speedo.query.parser.SpeedoQL;
import org.objectweb.speedo.query.parser.SpeedoQLQueryFilterVisitor;
import org.objectweb.speedo.query.parser.SpeedoQLVariableVisitor;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;
import java.io.CharArrayReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * SpeedoCompiledQuery object represents a query. This object is created
 * when a new query is created, and can be used several times. A list of
 * SpeedoCompiledQuery is managed with a SpeedoQueryManager object.
 * When a user creates a new JDO Query object (SpeedoQuery), a
 * SpeedoCompiledQuery object is associated to the SpeedoQuery object which is
 * used to delegate some methods.
 */
public class SpeedoCompiledQuery
        implements CompiledQuery, StateFilter, ReplaceableCacheEntry {

    private short status;
	private ClassLoader classLoader;

    /**
     * Fields from javax.jdo.Query
     */
	private QueryDefinitionImpl qd = null;

    /**
     * vparams, hparams and hvars are internal objects to manipulate the list
     * of parameters, and the list of variables.
     */
    private Map hparams = null;
    private Map paramName2paramClass = null;
    private Map hvars = null;
    private QueryEvalContext[] qecs = null;

    /**
     * The query tree which is the result from the parsing of the given filter
     */
    //private QueryTree qt = null;

    private Collection pncParams = null;
    /**
     * Logger for monolog
     */
    private Logger logger = null;
	private Logger varParserlogger = null;
	private Logger filterParserLogger = null;

    private DelegatePMapper mapper = null;
    private JormFactory jf;
	private PrefetchBufferFactory prefetchBufferFactory;


	private long age = 0;
	private int fixCount = 0;


    public void init(Logger l,
					 Logger logParserVar,
					 Logger logParserFil,
					 PMapper m,
					 PrefetchBufferFactory pbf,
					 JormFactory jf) {
        logger = l;
		varParserlogger = logParserVar;
		filterParserLogger = logParserFil;
        mapper = new DelegatePMapper(m);
		this.jf = jf;
        mapper.setJormFactory(jf);
		this.prefetchBufferFactory = pbf;
    }

    public PMapper getMapper() {
        return mapper;
    }

    public void setMapper(PMapper m) {
        mapper = new DelegatePMapper(m);
        if (jf != null) {
            mapper.setJormFactory(jf);
        }
    }

    public void setJormFactory(JormFactory jf) {
        this.jf = jf;
        if (mapper != null) {
            mapper.setJormFactory(jf);
        }
    }

	public Logger getLogger() {
		return logger;
	}

	public QueryEvalContext[] getQueryEvalContext() {
		return qecs;
	}

	public PrefetchBufferFactory getPrefetchBufferFactory() {
		return prefetchBufferFactory;
	}

	public boolean isPrefetchResult() {
		return qd.withPrefetch;
	}

    public void defineQuery(QueryDefinition _qd) {
	    this.qd = new QueryDefinitionImpl(_qd);
        classLoader = _qd.getCandidateClass().getClassLoader();
        if (classLoader == null) {
            classLoader = getClass().getClassLoader();
			if (classLoader == null) {
				classLoader = ClassLoader.getSystemClassLoader();
				logger.log(BasicLevel.DEBUG, "The system classLoader of the " +
						"class is assigned to the query: " + classLoader);
			} else {
				logger.log(BasicLevel.DEBUG, "The classLoader of Speedo is" +
						" assigned the query: " + classLoader);
			}
		} else {
			logger.log(BasicLevel.DEBUG, "The classLoader of the class is " +
					"assigned to the query: " + classLoader);
		}
        mapper.setClassLoader(classLoader);
        status = DEFINED;
    }

    // IMPLEMENTATION OF ReplaceableCacheEntry INTERFACE //
    //---------------------------------------------------//
	public long getCeAge() {
		return age;
	}

	public void setCeAge(long age) {
		this.age = age;
	}

	public void fixCe() {
		fixCount++;
	}

	public void unfixCe() throws UnFixProtocolException {
		fixCount--;
	}

	public int getCeFixCount() {
		return fixCount;
	}

	public Object getCeObject() {
		return this;
	}

	public Object getCeIdentifier() {
		return qd;
	}

    // IMPLEMENTATION OF CompiledQuery INTERFACE //
    //-------------------------------------------//
	public synchronized QueryDefinition getDefinition() {
		switch(status) {
		case DEFINED:
		case COMPILED:
			return qd;
		case UNDEFINED:
		default:
			return null;
		}
	}

    /**
     * compile the current SpeedoCompiledQuery.
     * The query is prepared to be executed.
     * The PersistenceManager is set (even if there was a previous definition
     * of a PersistenceManager.
     */
    public synchronized void compile() throws SpeedoException, MedorException, ExpressionException {
        if (status == UNDEFINED)
            throw new SpeedoException("Impossible to compile an undefined query");
        if (status == COMPILED)
            return;
		long timeToCompile = System.currentTimeMillis();
        boolean debug = logger.isLoggable(BasicLevel.DEBUG);
        // create a speedoQL object with a filter string
	    String filter = qd.filter;
	    filter = '(' + filter + ')';
        SimpleNode node = null;
        try {
            node = new SpeedoQL(new CharArrayReader(filter.toCharArray())).SpeedoQL();
        } catch (ParseException e) {
            throw new SpeedoException(
                    "Impossible to parse the filter and to create AST", e);
        }

        Manager miManager = mapper.getMetaInfoManager();
        if (miManager == null)
            throw new SpeedoException(
                    "A non null Meta information manager is needed");

        // create a new SpeedoQLVariableVisitor object with the previous
        // node and a Schema (jorm metainformation) array
        SpeedoQLVariableVisitor sqvv = new SpeedoQLVariableVisitor(node, mapper);
        sqvv.setLogger(varParserlogger);

        // create representations of the parameters list and the variable
        // list
        toHashtableParams(qd.parameters, ";,");
        toHashtableVars(qd.variables, ";,");

        // set these representations to the SpeedoQLVariableVisitor object
        sqvv.setParams(hparams);
        sqvv.setVars(hvars);
        sqvv.setOrders(qd.order);

        // set up the nameof the current class
        // could it be null ??? to verify it ... !!!
        sqvv.setCurrentClass(qd.candidateClass.getName());

        // start the variable visitor to catch all variables an build a
        // first tree of them
        try {
            sqvv.startVisiting();
        } catch (Exception e) {
            throw new SpeedoException(
                    "Impossible to visit the variable of the filter", e);
        }

        // get the query tree to set up the query filter visitor
		QueryTree qt = sqvv.getQueryTree();

        if (debug) {
            Field[] fields = qt.getTupleStructure().getFields();
            for (int i = 0; i < fields.length; i++) {
                logger.log(BasicLevel.DEBUG,
                        "i=" + i + ", fieldname=<" + fields[i].getName() + ">");
            }
        }

        //Assign mappers and project and project name on the JormExtent nodes
        Collection extents = JormQueryTreeHelper.getJormExtents(qt);
        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Extent nodes: " + extents.size());
        try {
            for (Iterator it = extents.iterator(); it.hasNext();) {
                JormExtent je = (JormExtent) it.next();
                if (debug)
                    logger.log(BasicLevel.DEBUG, "Looking the extent of the class " + je.getJormName());
                PClassMapping pcm = jf.getPClassMapping(
                        JormPathHelper.getOriginClass(je.getJormName()),
                        classLoader);
                if (debug)
                    logger.log(BasicLevel.DEBUG,
                            "JormExtent: " + je + " / pcm=" + pcm);
                je.setPMapper(pcm.getPMapper(), pcm.getProjectName());
            }
        } catch (PException e) {
            throw new SpeedoException("Error while fetching PClassPMapping", e);
        }

        // Build the root QueryNode which defines the filter and the projected
        // fields
        QueryNode root = new SelectProject("");
        if (!filter.equals("(true)") && !filter.equals("true")) {
            if (debug) {
                logger.log(BasicLevel.DEBUG, "filter = " + qd.filter);
            }
            SpeedoQLQueryFilterVisitor qfilter =
                    new SpeedoQLQueryFilterVisitor(qt, (ASTSpeedoQL) node);

			qfilter.setLogger(filterParserLogger);

            // set the representations of the variables and parameters to the
            // SpeedoQLQueryFilterVisitor object
            qfilter.setParams(hparams);
            qfilter.setVars(hvars);

            // set up the nameof the current class
            // could it be null ??? to verify it ... !!!
            qfilter.setCurrentClass(qd.candidateClass.getName());

            // start the query filter visitor, to build and expression tree of
            // the filter expression
            try {
                qfilter.startVisiting();
            } catch (Exception e) {
                throw new SpeedoException(
                        "Impossible to built the queryTree during the filter visit", e);
            }

            // this expression is the filter of the SelectProject
            Expression exprFilter = qfilter.getQueryFilter();
            if (exprFilter != null) {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "filter = " + exprFilter.getClass().getName());
                }
                root.setQueryFilter(exprFilter);
            }
        }

        logger.log(BasicLevel.INFO, "QueryTree built");
        if (debug) {
            QueryTreePrinter.printQueryTree(qt, logger);
        }

		qecs = getQueryTreeForInheritanceTree(qt, root, sqvv.classExtent);
		for(int i=0; i<qecs.length; i++) {
			// Optimize the queryTree
			optimize(qecs[i], debug);
			// Creates an evaluator associated to the QueryTree
			qecs[i].evaluator = new MedorEvaluator(qecs[i].query, 0);
		}
		timeToCompile = System.currentTimeMillis() - timeToCompile;
        status = COMPILED;
        logger.log(BasicLevel.INFO, "Query compiled in " + timeToCompile + "ms");
    }

	/**
	 * @return
	 * result[i][0] is the queryTree
	 * result[i][1] is the root queryTree where the projection and prefetch
	 * fields must be added. Root is a sub node of result[i][0].
	 * result[i][2] is the classExtent
	 * @throws MedorException
	 */
	private QueryEvalContext[] getQueryTreeForInheritanceTree(QueryTree qt,
														 QueryNode root,
													   ClassExtent ce) throws MedorException {
		org.objectweb.jorm.metainfo.api.Class clazz =
			(org.objectweb.jorm.metainfo.api.Class) ce.getMetaObject();
		ArrayList qts = new ArrayList();
		getQueryTreeForInheritanceTree(qt, root, ce, clazz, qts);
		return (QueryEvalContext[]) qts.toArray(new QueryEvalContext[qts.size()]);
	}

	private void getQueryTreeForInheritanceTree(QueryTree qt,
												QueryNode root,
												ClassExtent ce,
												org.objectweb.jorm.metainfo.api.Class clazz,
												List qts) throws MedorException {
		// Recall on sub classes. Must be done in first otherwise the cloning
		// is done on a modified queryTree
		for(Iterator it = clazz.getSubClasses().iterator(); it.hasNext();) {
			getQueryTreeForInheritanceTree(qt, root, ce,
				(org.objectweb.jorm.metainfo.api.Class) it.next(), qts);
		}
		if (clazz.isAbstract()) {
			//No query for an abstract class
			return;
		}

		QueryTree newQt = qt;
		ClassExtent newCe = ce;
		QueryEvalContext qec = new QueryEvalContext(root,
			qd.withPrefetch,// in case of filtered mapping ==> TODO: take in account horizontal mapping
			mapper.lookup(clazz.getFQName()),
			this,
				//TODO replace null by a Map for EvaluationMetaData
				null);


		if(!qec.withPrefetch) {
			//Nothing to do: execute the query in one shoot
		} else if(clazz.getFQName().equals(ce.getJormName())) {
			//The extent represents the right persistent class
			ce.setWithSubClasses(false);
		} else {
			//Clone the QueryTree
			HashMap obj2clone = new HashMap();
			boolean rootIsLinked = root.getChildren().length > 0;
			try {
				qec.query = (QueryNode) root.clone(null, obj2clone);
				newQt = (QueryTree) qt.clone(null, obj2clone);
			} catch (CloneNotSupportedException e) {
				throw new MedorException(e);
			}

			//Create a new ClassExtent for the current class
			newCe = new ClassExtent(clazz, clazz.getFQName(), Field.PNAMENAME, true);
			newCe.setPMapper(mapper, qec.pcm.getProjectName());

			//Fetch the cloned ClassExtent from the cloning context (Map)
			ClassExtent clonedCe = (ClassExtent) obj2clone.get(ce);
			if (clonedCe == null) {
				throw new MedorException("No clone found for the ClassExtent " + ce);
			}
			//Garbage the cloning context
			obj2clone.clear();
			obj2clone = null;

			// add the same field (the extent represent a sub class therefore,
			// the fields exist on the sub class)
			Field[] ceFields = clonedCe.getFields();
			//Start to 1 for avoiding the PName field which is already created
			for(int i=1; i<ceFields.length; i++) {
				String fn = ceFields[i].getName();
				if (fn.startsWith(ce.getName() + '.')) {
					fn = fn.substring(ce.getName().length() + 1);
				}
				if (!newCe.contains(newCe.getFieldName(newCe.getName(),fn))) {
					newCe.addField(fn);
				}
			}
			//Replace the ClassExtent of the ancestor by the one of the current class
			if (rootIsLinked) {
				qec.query = new QueryNodeReplacer().replaceNode(clonedCe, newCe, qec.query);
				if (newQt == clonedCe) {
					newQt = newCe;
				}
			} else {
				newQt = new QueryNodeReplacer().replaceNode(clonedCe, newCe, newQt);
			}
		}
		addProjectedAndOrderFields(newQt, (QueryNode) qec.query, newCe, clazz.getFQName(), qec);

		//Add the root of the queryTree to the list
		qts.add(qec);

		if (!qec.withPrefetch) {
			return;
		}
	}

	private void addProjectedAndOrderFields(QueryTree qt,
											QueryNode root,
											ClassExtent ce,
											String className,
											QueryEvalContext qec)
		throws MedorException {

		if (qd.withPrefetch) {
			JormQueryTreeHelper.addPrefetchFields(ce, qt, root);
			qec.pcm = mapper.lookup(className);
			qec.pnIndex= root.getTupleStructure().getSize() + 1;
		}

		// Project the pname
		QueryTreeField qtf = (QueryTreeField) qt.getTupleStructure()
			.getField(className + "." + Field.PNAMENAME);
		root.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

		if (qd.order != null && qd.order.size() > 0) {
			OrderField[] ofs = new OrderField[qd.order.size()];
			for(int i=0; i<ofs.length; i++) {
				String o = (String) qd.order.get(i);
				int idx = o.indexOf(' ');
				boolean desc = false;
				if (idx != -1) {
					desc = o.substring(idx + 1).trim().equals("descending");
					o = o.substring(0, idx);
				}
				o = className + '.' + o;
				ofs[i] = new BasicOrderField((QueryTreeField)
					qt.getTupleStructure().getField(o), desc);
			}
			root.setOrderBy(ofs);
		}
	}

	private void optimize(QueryEvalContext pqc, boolean debug)
		throws MedorException, ExpressionException {
		ArrayList rules = new ArrayList();
		rules.add(new JormFilteredInheritance());
		rules.add(new FlattenQueryTreeRule());
		//TODO: remove the dependency to JormFlatten2Rdb
		rules.add(new JormFlatten2Rdb());
		QueryTransformer queryTransformer = new BasicQueryRewriter(rules);
		IndexesGenerator indexesGenerator = new IndexesGenerator();
		try {
			pqc.query = queryTransformer.transform(pqc.query);
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Query optimized");
				QueryTreePrinter.printQueryTree(pqc.query, logger);
			}
			pqc.query = indexesGenerator.transform(pqc.query);
		} catch (Exception e) {
			throw new MedorException("Impossible to optimize the query", e);
		}

		pncParams = JormQueryTreeHelper.getRequiredPNameManagers(pqc.query);
		for (Iterator it = pncParams.iterator(); it.hasNext();) {
			ParameterOperand po = (ParameterOperand) it.next();
			String name = po.getName();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "ParameterOperand " + name + " is expected.");
			}
			po.setValue(JormPathHelper.getPNameCoder(name, mapper));
		}
	}


    /**
     * evaluate the query with a single parameter which is a array of object
     * parameters.
     * @param a    the array parameter of the query
     * @param pm    the persistence manager object
     * @return  a Collection of result objects
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
    public Collection execute(Object[] a, ProxyManager  pm)
            throws SpeedoException, MedorException, ExpressionException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos = null;
        if (a != null) {
            pos = new ParameterOperand[a.length + pncParams.size()];
            for(Iterator it = hparams.values().iterator(); it.hasNext();) {
				Object[] o = (Object[]) it.next();
				int idx = ((Integer) o[0]).intValue();
                pos[idx] = new BasicParameterOperand(
                    (BasicParameterOperand) o[1]);
				treatParameter(pos[idx], a[idx]);
            }
        } else {
            pos = new ParameterOperand[pncParams.size()];
        }
		if (pncParams.size() >0) {
			int i = (a == null ? 0 : a.length);
			for (Iterator it = pncParams.iterator(); it.hasNext(); i++) {
				pos[i++] = (ParameterOperand) it.next();
			}
		}
        return executeQT(pm, pos);
    }


    /**
     * evaluate the query with a single parameter which is a Map of object parameters.
     * @param m    the map parameter of the query
     * @param pm    the persistence manager object
     * @return  a Collection of result objects
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
    public Collection execute(Map m, ProxyManager  pm)
            throws SpeedoException, MedorException, ExpressionException {
        if (status != COMPILED)
            throw new EvaluationException(
                    "Impossible to execute a query if it has not been defined and compiled before");
        ParameterOperand[] pos =
                new BasicParameterOperand[m.size() + pncParams.size()];
		for(Iterator it = hparams.values().iterator(); it.hasNext();) {
			Object[] o = (Object[]) it.next();
			int idx = ((Integer) o[0]).intValue();
			pos[idx] = new BasicParameterOperand(
				(BasicParameterOperand) o[1]);
			treatParameter(pos[idx], m.get(pos[idx].getName()));
			}
		if (pncParams.size() >0) {
			int i = (m == null ? 0 : m.size());
			for (Iterator it = pncParams.iterator(); it.hasNext(); i++) {
				pos[i++] = (ParameterOperand) it.next();
			}
		}
        return executeQT(pm, pos);
    }

	private void treatParameter(ParameterOperand po, Object value)
		throws SpeedoException, ExpressionException {
		Object val = value;
		if (po.getType().getTypeCode() == QType.TYPECODE_PNAME) {
			//Assign the pname of the parameter
			if (val == null) {
				try {
					val =
						jf.getPClassMapping(
						(Class) paramName2paramClass.get(po.getName()))
						.getPBinder().getNull();
				} catch (Exception e) {
					throw new SpeedoException(
						"Impossible to find the null PName of the parameter (name= "
						+ po.getName()
						+ " / type= " + po.getType().getJormName(),
						e);
				}
			} else if (val instanceof SpeedoProxy) {
				val = ((SpeedoProxy) value).getPName();
			} else {
				throw new JDOUserException("The parameter '"
					+ po.getName()
					+ "' must be a persistent object: " + val);
			}
		}
		po.setValue(val);
	}

    public short getStatus() {
        return status;
    }

    // Private Methods //
    //-----------------//

    /**
     * Hash a String, and compute a Hashtable
     * example:
     *   ("String name, Float salary, Employee boss", ",")
     *   keys     | values
     *   ---------------------
     *   "name"   | "String"
     *   "salary" | "Float"
     *   "boss"   | "Employee"
     *
     * @param stringToHash	the String to hash
     * @param separator		the separator to tokenize
     */
    private void toHashtableParams(String stringToHash, String separator) {
        hparams = new HashMap();
	    paramName2paramClass = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok = new StringTokenizer(stringToHash, separator);
			int idx = 0;
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String name = tuple.substring(i + 1, tuple.length());
	            String paramType = tuple.substring(0, i).trim();
	            PType type = getPType(paramType);
                hparams.put(name, new Object[]{
					new Integer(idx),
					new BasicParameterOperand(type, name)});
				idx ++;
	            if (type == PTypeSpaceMedor.PNAME) {
	                paramName2paramClass.put(name, getClass(paramType));
                }
            }
        }
    }

    private void toHashtableVars(String stringToHash, String separator) {
        hvars = new HashMap();
        if (stringToHash != null) {
            StringTokenizer tok = new StringTokenizer(stringToHash, separator);
            while (tok.hasMoreTokens()) {
                String tuple = tok.nextToken().trim();
                int i = tuple.indexOf(' ');
                String type = tuple.substring(0, i);
                hvars.put(tuple.substring(i + 1, tuple.length()), type);
            }
        }
    }

	private PType getPType(String name) {
		if ("String".equals(name)) {
			return PTypeSpace.STRING;
		}
		if ("Integer".equals(name)) {
			return PTypeSpace.OBJINT;
		}
		for (int i = 0; i < PTypeSpace.PREDEFINEDPTYPES.length; i++) {
			PType type = PTypeSpace.PREDEFINEDPTYPES[i];
			if (type.getJavaName().equals(name)
			    || type.getJormName().equals(name)) {
				return type;
			}
		}
		if ("Collection".equals(name) 
		        || PTypeSpaceMedor.COLLECTION.getJavaName().equals(name)) {
		    return PTypeSpaceMedor.COLLECTION;
		}
		return PTypeSpaceMedor.PNAME;
	}
    private Class getClass(String classname) {
        if (classLoader != null) {
            try {
                return classLoader.loadClass(classname);
            } catch (ClassNotFoundException e) {
                try {
                    return classLoader.loadClass(
                            qd.candidateClass.getPackage().getName() + "." + classname);
                } catch (ClassNotFoundException e2) {
                    return null;
                }
            }
        } else {
            return null;
        }
    }

    /**
     * executes a the current query, and returns a Collection of object.
     * @return  a new Collection of objects.
     * @throws org.objectweb.medor.api.EvaluationException
     * @throws org.objectweb.medor.api.MedorException
     */
	private Collection executeQT(ProxyManager pm, ParameterOperand[] pos)
        throws EvaluationException, MedorException {
		if (!qd.ignoreCache || !pm.getIgnoreCache()) {
			try {
				pm.getTransactionalPersistenceManager()
						.flush((Transaction) pm.currentTransaction(), this);
			} catch (PersistenceException e) {
				throw new JDOException("Impossible to use the cache into a " +
						"query: Error during the flushing of data on the support", e);
			}
		}
		Object connection = ((Transaction) pm.currentTransaction())
				.getConnectionHolder();
		TupleCollection queryResult = null;
		if (qecs.length == 1) {
			//optimization: avoid to use an intermediate class

			queryResult = qecs[0].eval(pm, pos, connection);
			if (queryResult == null || queryResult.isEmpty()) {
				logger.log(BasicLevel.INFO,
						"Be careful, the result collection is empty");
				return Collections.EMPTY_SET;
			}
		} else {
			//Use a multiplex of query
			queryResult = new QueriesUnion(pos, pm, connection, qecs);
		}
		return new QueryResultCollection(
			queryResult, pm, new Object[]{connection},
			qecs.length == 1, logger);
    }

	// IMPLEMENTATION OF THE CacheEntryFilter INTERFACE //
	//--------------------------------------------------//

	public boolean accept(State ce) {
		return true;
	}


}