/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S. Chassande-Barrioz
 *
 */

package org.objectweb.speedo.query.parser;

import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.jorm.lib.NavigatorOperator;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;

/**
 * Implementation of a visitor that creates a map of pairs [id,JormExtent] for all defined
 * identifiers of the query.
 */
public class SpeedoQLVariableVisitor extends SpeedoQLAbstractVisitor {

    private PMapper mapper;

    /**
     * ids is a temporary structure to store variables, parameters, name
     * definitions, and build corresponding QueryTree(s)
     * key: the name to resolve
     * value: an IdValue object which contains the name, the paths, and the
     * corresponding QueryTree (see the definition of IdValue object in
     * SpeedoQLAbstractVisitor object)
     */
    private Map ids = new HashMap();

    /**
     * qt is the built QueryTree from the current object.
     */
    private QueryTree qt = null;

    /**
     * speedoql is the result of the filter parsing
     */
    private SimpleNode speedoql = null;

    public ClassExtent classExtent = null;

    private int nbNot = 0;

    private List orders;


    /**
     * constructor
     * @param speedoql root of the lexical tree
     * @param mapper is the JORM mapper
     */
    public SpeedoQLVariableVisitor(SimpleNode speedoql, PMapper mapper) {
        this.speedoql = speedoql;
        this.mapper = mapper;
    }

    public void setOrders(List orders) {
        this.orders = orders;
    }

    /**
     * The visit of the tree starts here.
     * Please setup current class, params and vars hashtable before calling
     * this method.
     */
    public void startVisiting() throws Exception {
        debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
        nbNot = 0;

        //Add the initial class as the first IdValue. The key is 'this'
        if (curClass != null) {
            logger.log(BasicLevel.DEBUG, "create a new IdValue object " +
                    "with the current class (" + curClass + ")");

            SpeedoQLAbstractVisitor.IdValue iv = new SpeedoQLAbstractVisitor.IdValue();
            iv.name = new String[1];
            iv.name[0] = curClass;
            iv.nameType = SpeedoQLAbstractVisitor.JORM_NAME;
            ids.put("this", iv);
        }
        //visit the tree
        visit(speedoql);
        if (orders != null) {
            for(int i=0; i<orders.size(); i++) {
                String fieldName = (String) orders.get(i);
                int idx = fieldName.indexOf(' ');
                if (idx != -1) {
                    fieldName = fieldName.substring(0, idx).trim();
                }
                manageName(null, fieldName, splitPath(fieldName));
            }
        }

        // Build the QueryTree instances
        for (Iterator i = ids.keySet().iterator(); i.hasNext();) {
            String id = (String) i.next();
            getQueryTree(id);
        }

        qt = ((SpeedoQLAbstractVisitor.IdValue) ids.get("this")).qt;
    }

    private QueryTree getQueryTree(String id) throws ParseException, MedorException {
        logger.log(BasicLevel.DEBUG, "getQueryTree(" + id + "): begin");
        SpeedoQLAbstractVisitor.IdValue iv = (SpeedoQLAbstractVisitor.IdValue) (id.equals(curClass)
                ? ids.get("this") : ids.get(id));
        if (iv == null || iv.name == null)
            throw new ParseException("Undeclared identifier '" + id + "'");

        if (iv.qt != null)
            return iv.qt;

        if (iv.name.length == 1) {
            // Persistent class
            if (mapper.lookup(iv.name[0]) == null) {
                throw new ParseException("Class '" + iv.name[0] + "' is not mapped");
            }
            Class theClass = mapper.getMetaInfoManager().getClass(iv.name[0]);
            if (theClass == null) {
                throw new ParseException("Undeclared persistent class '"
                        + iv.name[0] + "'");
            }
            classExtent = new ClassExtent(
                theClass, curClass, Field.PNAMENAME, false);
			classExtent.setWithSubClasses(!theClass.getSubClasses().isEmpty());
            iv.qt = classExtent;
            if (debug) {
                logger.log(BasicLevel.DEBUG, "getQueryTree(" + id + "): extent=" + classExtent);
            }

        } else {
            // The query tree depends on the defintion of another id
            // and is obtain after a navigation
            iv.qt = getQueryTree(iv.name[0]);
            if (debug) {
                QueryTreePrinter.printQueryTree(iv.qt, logger);
            }
            iv.qt = getNavNode(iv.qt, iv.name, id, iv.nameType, false).getQueryTree();
        }

        //Build the nodes depending on this definition
        for (int i = 0; i < iv.getDeclaredPathLength(); i++) {
            String[] p = iv.getSplitedPath(i);
            if (debug) {
                QueryTreePrinter.printQueryTree(iv.qt, logger);
            }
            iv.qt = getNavNode(iv.qt, p, mergePath(p, 0, p.length - 1), iv.nameType, true).getQueryTree();
        }

        if (iv.name.length > 1) {
            setQueryTree(iv.name[0], iv.qt);
        }
        if (debug) {
            logger.log(BasicLevel.DEBUG, "getQueryTree(" + id + "): end");
            QueryTreePrinter.printQueryTree(iv.qt, logger);
        }
        return iv.qt;
    }

    private QueryTreeField getNavNode(QueryTree qt,
                                 String[] pathName,
                                 String alias,
                                 int name_t,
                                 boolean isPath) throws ParseException, MedorException  {
        if (debug) {
            logger.log(BasicLevel.DEBUG, "getNavNode(" + alias
                + ", " + buildStringwithout(pathName, -1, ".")
                + ", " + operationToString(name_t)
                + ", isPath=" + isPath + ")");
        }
        NavigatorOperator nav = null;
        boolean baseOnThis;
        Field f;
        PNameField pnf = null;
        switch (name_t) {
        case NAVIGATION:
        case JORM_NAME:
        case IS_EMPTY:
        case IS_NOT_EMPTY:
            //Fetch the pname field of the initial class
            baseOnThis = pathName[0].equals("this");

            f = qt.getTupleStructure().getField(baseOnThis
                        ? curClass + "." + Field.PNAMENAME : pathName[0]);
            if (debug) {
                logger.log(BasicLevel.DEBUG, "Starting field: " + f.getName());
            }
            if (f instanceof PNameField) {
                pnf = (PNameField) f;
            } else if (f instanceof PropagatedField) {
                pnf = (PNameField) ((PropagatedField) f).getOriginFields()[0];
            } else {
                throw new JDOException("Internal error: Unsupported field: " + f);
            }
            //build NavigatorOperator tree in according to the name
            nav = new NavigatorOperator(new BasicFieldOperand(f), pathName[1]);
            for (int i = 2; i < pathName.length; i++) {
                if (i < (pathName.length-1) ||  isMethodOperator(pathName[i]) == -1) {
                    nav = new NavigatorOperator(nav, pathName[i]);
                }
            }
            int idx = alias.indexOf("this");
            if (idx != -1) {
                alias = curClass + alias.substring(idx + 4);
            }
            byte op;
            switch(name_t) {
            case NAVIGATION:
            case JORM_NAME:
                op = NavigatorNodeFactory.NAVIGATION;
                break;
            case IS_EMPTY:
                op = NavigatorNodeFactory.IS_EMPTY;
                break;
            case IS_NOT_EMPTY:
                op = NavigatorNodeFactory.IS_NOT_EMPTY;
                break;
            default:
                throw new JDOException("Internal error !");
            }
            return NavigatorNodeFactory.navigate(nav, alias, op);
        case CONTAINS_MEMBEROF:
            return getNavNode(qt, pathName, null, CONTAINS_IN, isPath);
        case CONTAINS_IN:
            if (isPath) {
                return getNavNode(qt, pathName, alias, NAVIGATION, isPath);
            }
            baseOnThis = pathName[0].equals("this");
            f = qt.getTupleStructure().getField(baseOnThis
                    ? curClass + "." + Field.PNAMENAME : pathName[0]);
            nav = new NavigatorOperator(new BasicFieldOperand(f), pathName[1]);
            for (int i = 2; i < pathName.length; i++) {
                nav = new NavigatorOperator(nav, pathName[i]);
            }
            return NavigatorNodeFactory.navigate(nav, alias,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST);

        default:
            throw new JDOException("Unsupported type of identifier: "
                    + name_t + " / " + operationToString(name_t));
        }
    }

    private void setQueryTree(String id, QueryTree aqt) {
        SpeedoQLAbstractVisitor.IdValue iv = null;
        if (id.compareTo(curClass) == 0)
            iv = (SpeedoQLAbstractVisitor.IdValue) ids.get("this");
        else
            iv = (SpeedoQLAbstractVisitor.IdValue) ids.get(id);

        iv.qt = aqt;
        if (iv.name.length > 1)
            setQueryTree(iv.name[0], iv.qt);
    }

    /**
     * get the Map that was built from visiting the lexical query tree
     */
    public QueryTree getQueryTree() {
        return qt;
    }


    /**
     * ********************* VISITOR METHODS ***********************************
     */

    public Object visit(ASTSpeedoPrimary node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTSpeedoQL node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTPrimary node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTRelationalExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTAdditiveExpression node, Object data) {

        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTUnaryExpression node, Object data) {
        boolean hasNot = node.ops.size() > 0
                && ((Integer) node.ops.get(0)).intValue() == SpeedoQLConstants.NOT;
        if (hasNot) {
            nbNot ++;
            if (debug) {
                logger.log(BasicLevel.DEBUG, "remember a Not: "+ nbNot);
            }
        }
        visit((SimpleNode) node, data);
        if (hasNot && nbNot> 0) {
            nbNot--;
            if (debug) {
                logger.log(BasicLevel.DEBUG, "forget a Not: "+ nbNot);
            }
        }
        //logger.log(BasicLevel.DEBUG, "End of Visit UnaryExpression");
        return null;
    }

    public Object visit(ASTCastExpression node, Object data) {
        return null;
    }

    public Object visit(ASTArgumentList node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTLiteral node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTType node, Object data) {
        //voir cast , donc on s'en fout
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTQualifiedName node, Object data) {
        Stack stack = (Stack) data;
        String name = (String) node.value;
        if (debug) {
            logger.log(BasicLevel.DEBUG, "variable visitor: qualifiedname=<" + name + ">");
        }

        String[] splitted = splitPath(name);
        SpeedoQLAbstractVisitor.IdValue iv = null;
        if (params != null && params.get(splitted[0]) != null) {
            if (stack.size() > 0 && stack.peek() instanceof String) {
                String setpath = (String) stack.pop();
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "The parameter '" + splitted[0]
                        + "' is used in a contains expression: "
                        + setpath + ".contains(" + splitted[0] + ")");
                }
                iv = (SpeedoQLAbstractVisitor.IdValue) ids.get(splitted[0]);
                if (iv == null) {
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "Create a new IdValue id="
                            + splitted[0] + " / name=" + setpath);
                    }
                    iv = new SpeedoQLAbstractVisitor.IdValue(setpath, CONTAINS_MEMBEROF);
                    ids.put(splitted[0], iv);
                } else {
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "Use the IdValue " + splitted[0]);
                    }
                }
            } else {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Use of the parameter " + splitted[0]);
                }
                //nothing to do
            }

        } else if (vars != null && vars.get(splitted[0]) != null) {
            if (debug) {
                logger.log(BasicLevel.DEBUG, "Use of the variable " + splitted[0]
                    + " : " + name);
            }
            //fetch or create an IdValue
            iv = (SpeedoQLAbstractVisitor.IdValue) ids.get(splitted[0]);
            if (iv == null) {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Create a new IdValue for the variable " + splitted[0]);
                }
                // The definition of the variable is not yet knwon
                iv = new SpeedoQLAbstractVisitor.IdValue();
                iv.nameType = CONTAINS_IN;
                ids.put(splitted[0], iv);
            } else {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Use the IdValue of the variable " + splitted[0]);
                }
            }
            if (stack.size() > 0 && stack.peek() instanceof String) {
                // Define the IdValue from previous (contains)
                String path = (String) stack.pop();
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "Define the variable " + splitted[0]
                        + " to " + path);
                }
                if (iv.name != null) {
                    throw new JDOUserException("Variable '"
                            + splitted[0] + "' defined several times : \n\t"
                            + iv.name + "\n\t" + path);
                }
                iv.name = splitPath(path);
            }
            String last = splitted[splitted.length - 1];
            int operatorId = isMethodOperator(last);
            if (operatorId != -1) {
                //current expression is ended by an operator
                String trunked = buildStringwithout(splitted, splitted.length-1, ".");
                if (operatorId == CONTAINS_OPERATOR) {
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "The variable '" + splitted[0] + " is used in a contain expression");
                    }
                    //push in the stack a mean to find the contains stuff next time...
                    if (stack != null) {
                        stack.push(trunked);
                    }
                }
                // Add the use in the path list
                iv.addPath(trunked);
            } else if (splitted.length > 1) {
                // Add the use in the path list
                iv.addPath(name);
            }
        } else { // current class
            manageName(stack, name, splitted);
        }
        if (debug) {
            logger.log(BasicLevel.DEBUG, "variable visitor: qualifiedname=<" + name + ">-end");
        }
        return null;
    }

    public void manageName(Stack stack, String name, String[] splitted) {
        String begin = buildStringwithout(splitted, splitted.length-1, ".");
        String last = splitted[splitted.length - 1];
        int operatorId = isMethodOperator(last);
        if (operatorId != -1) {
            switch(operatorId) {
            case CONTAINS_OPERATOR: 
                    String trunked = "this."
                        + buildStringwithout(splitted, splitted.length-1, ".");
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "contains case, path: " + trunked);
                    }
                //stock in stack a mean to find the contains stuff...
                if (stack != null) {
                    stack.push(trunked);
                }
                break;
            case IS_EMPTY_OPERATOR: 
                    String setpath = "this."
                        + buildStringwithout(splitted, splitted.length-1, ".");
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "isempty case, path: " + setpath);
                    }
                //stock in stack a mean to find the contains stuff...
                SpeedoQLAbstractVisitor.IdValue iv = (SpeedoQLAbstractVisitor.IdValue) ids.get("this");
                iv = (SpeedoQLAbstractVisitor.IdValue) ids.get(splitted[0]);
                if (iv == null) {
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "Create a new IdValue id="
                            + splitted[0] + " / name=" + setpath);
                    }
                    iv = new SpeedoQLAbstractVisitor.IdValue(setpath, ((nbNot % 2) == 0 ? IS_EMPTY : IS_NOT_EMPTY));
                    ids.put(splitted[0], iv);
                } else {
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "Use the IdValue " + splitted[0]);
                    }
                }
                break;
            }
            return;
        } else {
            String path = (name.startsWith("this")
                    ? name : "this." + name);
            if (debug) {
                logger.log(BasicLevel.DEBUG, "Navigation case, path:" + path);
            }
            SpeedoQLAbstractVisitor.IdValue iv = (SpeedoQLAbstractVisitor.IdValue) ids.get("this");
            iv.nameType = NAVIGATION;
            iv.addPath(path);
        }
    }

}
