/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.workingset.lib;

import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Loggable;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.speedo.api.TransactionListener;

import java.util.Map;
import java.util.HashMap;
import java.util.GregorianCalendar;
import java.util.Calendar;

/**
 * This class is a simple implementation of the TransactionListener interface
 * permitting to listen events concerning the transaction life cycle.
 *
 * @author S.Chassande-Barrioz
 */
public class WorkingSetStatistic
		implements TransactionListener, Loggable {

	public static int LOG_LEVEL = BasicLevel.INFO;

	public static WorkingSetStatistic instance = null;
	public int maxWSSize;
	public int minWSSize;
	public int nbWS;
	public int averageWSSize;
	public long averageWSduration;
	public int committedWS;
	public Logger logger;
	private Calendar cal;

	public Map currentCtx;



	public WorkingSetStatistic() {
		instance = this;
		cal = GregorianCalendar.getInstance();
		init();
	}

	public void init() {
		maxWSSize = 0;
		minWSSize = Integer.MAX_VALUE;
		nbWS = 0;
		averageWSSize = 0;
		committedWS = 0;
		if (currentCtx == null) {
			currentCtx = new HashMap(10);
		} else {
			currentCtx.clear();
		}
	}


	public void printStatistic() {
		printStatistic(logger);
	}

	public void printStatistic(Logger logger) {
		logger.log(LOG_LEVEL, "Transaction statistic:"
				+ "\n\t- Max size: " + maxWSSize
				+ "\n\t- Min size: " + minWSSize
				+ "\n\t- Average size: " + averageWSSize
				+ "\n\t- Average duration: " + averageWSduration
				+ "\n\t- Transaction number: " + nbWS
				+ "\n\t- Commited Transaction: " + ((committedWS * 100) / nbWS) + "%"
				+ "\n\t- current Transaction: " + currentCtx.size());
	}

	public void transactionEnded(Object tx, int s, boolean validate) {
		long d2 = cal.getTime().getTime();
		Long l = (Long) currentCtx.get(tx);
		if (l != null) {
			currentCtx.remove(tx);
			averageWSduration = ((averageWSduration * nbWS) + d2 - l.longValue()) / (nbWS + 1);
		}
		maxWSSize = Math.max(maxWSSize, s);
		minWSSize = Math.min(minWSSize, s);
		averageWSSize = ((averageWSSize * nbWS) + s) / (nbWS + 1);
		nbWS++;
		if (validate)
			committedWS++;
	}


	// IMPLEMENTATION OF THE TransactionListener INTERFACE //
	//-----------------------------------------------------//

	public synchronized void transactionBegun(Object tx) {
		Long l = new Long(cal.getTime().getTime());
		currentCtx.put(tx, l);
	}

	public synchronized void transactionCommitted(Object tx, int s) {
		transactionEnded(tx, s, true);
		printStatistic();
	}

	public synchronized void transactionAborted(Object tx, int s) {
		transactionEnded(tx, s, false);
		printStatistic();
	}


	// IMPLEMENTATION OF THE Loggable INTERFACE //
	//------------------------------------------//

	public Logger getLogger() {
		return logger;
	}

	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	public LoggerFactory getLoggerFactory() {
		return null;
	}

	public void setLoggerFactory(LoggerFactory loggerFactory) {
		if (loggerFactory != null) {
			logger = loggerFactory.getLogger("org.objectweb.speedo.txStatistic");
		}
	}
}
