/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 */

package org.objectweb.speedo.runtime.detach;


import java.sql.Connection;
import java.util.ArrayList;
import java.util.Collection;

import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;

import org.objectweb.jorm.api.PMapper;
import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.pobjects.detach.Car;
import org.objectweb.speedo.pobjects.detach.Coach;
import org.objectweb.speedo.pobjects.detach.FormulaOne;
import org.objectweb.speedo.pobjects.detach.Player;
import org.objectweb.speedo.pobjects.detach.Team;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * @author Y.Bersihand
 */
public class TestDetach extends SpeedoTestHelper {

	public TestDetach(String s) {
		super(s);
		deleteDatabase();
	}

	protected String getLoggerName() {
		return LOG_NAME + ".rt.detach.TestDetach";
	}
	
	/**
	 * Test the detach method: make an object persistent, and detach it out of an active transaction.
	 */
	public void testDetachCopy() {
		logger.log(BasicLevel.DEBUG, "***************testDetachCopy*****************");
		Team t = new Team("Bordeaux",null,null);
		Collection players = new ArrayList();
		Player p1 = new Player("p1", t, 25);
		players.add(p1);
		Player p2 = new Player("p2", t, 32);
		players.add(p2);
		t.setPlayers(players);
		Coach c = new Coach("c1", 5, t);
		t.setCoach(c);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		logger.log(BasicLevel.DEBUG, "make persistent the team " + t.toString());
		pm.makePersistent(t);
		pm.currentTransaction().commit();
		//detach the team t
		Team copyOfT = (Team) ((ProxyManager)pm).detachCopy((Object)t);
		//print the team out
		logger.log(BasicLevel.DEBUG, copyOfT.toString());
		pm.close();
	}
	
	
	/**
	  * Test the detach method with a non persistent object: the object is made persistent before being detached. 
	  */
	public void testDetachNonPersistentCopy() {
		logger.log(BasicLevel.DEBUG, "*************testDetachNonPersistentCopy*****************");
		Team t = new Team("Marseille",null,null);
		Coach c = new Coach("c2", 5, t);
		t.setCoach(c);
		Player p = new Player("p3", t, 25);
		t.addPlayer(p);
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		//		detach the team t while it is not persistent
		Team copyOfT = (Team) ((ProxyManager)pm).detachCopy((Object)t);
		pm.currentTransaction().commit();
		//print the team out
		logger.log(BasicLevel.DEBUG, copyOfT.toString());
		pm.close();
	}
	
	/**
	  * Test the detach method with a persistent-new object: the object is flushed before being detached. 
	  */
	public void testDetachPersistentNew() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentNew****************");
		Team t = new Team("Monaco",null,null);
		Coach c = new Coach("c3", 5, t);
		t.setCoach(c);
		Player p = new Player("p4", t, 25);
		t.addPlayer(p);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		pm.makePersistent(c);
		//detach the team while it is persistent-new: the object must be flushed before being detached
		Team copyOfT = (Team) ((ProxyManager)pm).detachCopy((Object)t);
		pm.currentTransaction().commit();
		pm.close();
		//	print the team out
		logger.log(BasicLevel.DEBUG, copyOfT.toString());
	}
	
	/**
	  * Test the detach method with a persistent-dirty object: the object is flushed before being detached. 
	  */
	public void testDetachPersistentDirty() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentDirty****************");
		Team t = new Team("Nantes",null,null);
		Coach c = new Coach("c4", 5, t);
		t.setCoach(c);
		Player p = new Player("p5", t, 28);
		t.addPlayer(p);
				
		PersistenceManager pm = pmf.getPersistenceManager();
		//make persistent
		pm.currentTransaction().begin();
		pm.makePersistent(c);
		pm.currentTransaction().commit();
		//update and detach
		pm.currentTransaction().begin();
		c.setExperience(10);
		//detach the team while it is persistent-dirty: the object must be flushed before being detached
		Team copyOfT = (Team) ((ProxyManager)pm).detachCopy((Object)t);
		pm.currentTransaction().commit();
		pm.close();
		//	print the team out
		logger.log(BasicLevel.DEBUG, copyOfT.toString());
	}

	/**
	  * Test the detach method with a persistent-deleted object: an exception must be thrown. 
	  */
	public void testDetachPersistentDeleted() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentDeleted****************");
		Team t = new Team("Paris",null, null);
		Coach c = new Coach("c5", 5, t);
		t.setCoach(c);
		Player p = new Player("p6", t, 28);
		t.addPlayer(p);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		//make persistent
		pm.currentTransaction().begin();
		pm.makePersistent(c);
		pm.currentTransaction().commit();
		//delete
		pm.currentTransaction().begin();
		pm.deletePersistent(c);
		try{
			//detach the coach c while it is persistent-deleted: an exception must be thrown
			Coach copyOfC = (Coach) ((ProxyManager)pm).detachCopy((Object)c);
		}
		catch(Exception e){
			if(e.getClass() == JDOUserException.class)
				logger.log(BasicLevel.DEBUG, "Correct exception type is thrown: " + e);
			else
				logger.log(BasicLevel.DEBUG, "Wrong exception type is thrown: " + e + ". It should be a JDOUserException exception.");
		}
		pm.currentTransaction().commit();
		pm.close();
	}
	
	/**
	  * Test the detach method with a persistent-deleted object within a collection: an exception must be thrown. 
	  */
	public void testDetachPersistentDeletedCollection() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentDeletedCollection****************");
		Team t = new Team("Auxerre",null, null);
		Coach c = new Coach("c6", 10, t);
		t.setCoach(c);
		Player p = new Player("p7", t, 28);
		t.addPlayer(p);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		//make persistent
		pm.currentTransaction().begin();
		pm.makePersistent(c);
		pm.currentTransaction().commit();
		//delete
		pm.currentTransaction().begin();
		pm.deletePersistent(p);
		try{
			//detach the coach c while it is persistent-deleted: an exception must be thrown
			Coach copyOfC = (Coach) ((ProxyManager)pm).detachCopy((Object)c);
			//Team copyOfT = (Team) ((ProxyManager)pm).detachCopy(t);
		}
		catch(Exception e){
			if(e.getClass() == JDOUserException.class)
				logger.log(BasicLevel.DEBUG, "Correct exception type is thrown: " + e);
			else
				logger.log(BasicLevel.DEBUG, "Wrong exception type is thrown: " + e + ". It should be a JDOUserException exception.");
		}
		pm.currentTransaction().commit();
		pm.close();
	}

	/**
	 * Test the detach method with inheritance.
	 */
	public void testDetachInherited() {
		logger.log(BasicLevel.DEBUG, "***************testDetachInherited*****************");
		Car c = new Car("r5", 4, "red");
		FormulaOne f = new FormulaOne("williams", 4, "green", 262);
	
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		logger.log(BasicLevel.DEBUG, "make persistent the car " + c.toString() + 
			" and the formula one " + f.toString());
		pm.makePersistent(c);
		pm.makePersistent(f);
		pm.currentTransaction().commit();
		//detach the car c
		Car copyOfC = (Car) ((ProxyManager)pm).detachCopy((Object)c);
		//print the car out
		logger.log(BasicLevel.DEBUG, copyOfC.toString());
		//detach the formula one f
		FormulaOne copyOfF = (FormulaOne) ((ProxyManager)pm).detachCopy((Object)f);
		//print the formula one out
		logger.log(BasicLevel.DEBUG, copyOfF.toString());
		pm.close();		
	}	

	/**
	  * Test the detach method with a non persistent object which is inherited from a super class: the object is made persistent before being detached. 
	  */
	public void testDetachNonPersistentInherited() {
		logger.log(BasicLevel.DEBUG, "*************testDetachNonPersistentInherited*****************");
		FormulaOne f = new FormulaOne("sauber", 4, "blue", 274);
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		//		detach the formula one f while it is not persistent
		FormulaOne copyOfF = (FormulaOne) ((ProxyManager)pm).detachCopy((Object)f);
		pm.currentTransaction().commit();
		//print the team out
		logger.log(BasicLevel.DEBUG, copyOfF.toString());
		pm.close();
	}	

	/**
	  * Test the detach method with a persistent-deleted inherited object: an exception must be thrown. 
	  */
	public void testDetachPersistentDeletedInherited() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentDeletedInherited****************");
		FormulaOne f = new FormulaOne("honda", 4, "yellow", 240);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		//make persistent
		pm.currentTransaction().begin();
		pm.makePersistent(f);
		pm.currentTransaction().commit();
		//delete
		pm.currentTransaction().begin();
		pm.deletePersistent(f);
		try{
			//detach the coach c while it is persistent-deleted: an exception must be thrown
			FormulaOne copyOfF = (FormulaOne) ((ProxyManager)pm).detachCopy((Object)f);
		}
		catch(Exception e){
			if(e.getClass() == JDOUserException.class)
				logger.log(BasicLevel.DEBUG, "Correct exception type is thrown: " + e);
			else
				logger.log(BasicLevel.DEBUG, "Wrong exception type is thrown: " + e + ". It should be a JDOUserException exception.");
		}
		pm.currentTransaction().commit();
		pm.close();
	}
	
	
	/**
	  * Test the detach method with a persistent-dirty inherited object: the object is flushed before being detached. 
	  */
	public void testDetachPersistentDirtyInherited() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentDirtyInherited****************");
		FormulaOne f = new FormulaOne("ferrari", 4, "yellow", 240);
		PersistenceManager pm = pmf.getPersistenceManager();
		//make persistent
		pm.currentTransaction().begin();
		pm.makePersistent(f);
		pm.currentTransaction().commit();
		//update and detach
		pm.currentTransaction().begin();
		f.setNbOfWheels(3);
		f.setSpeedMax(300);
		//detach the formula one while it is persistent-dirty: the object must be flushed before being detached
		FormulaOne copyOfF = (FormulaOne) ((ProxyManager)pm).detachCopy((Object)f);
		pm.currentTransaction().commit();
		pm.close();
		//	print the formula one out
		logger.log(BasicLevel.DEBUG, copyOfF.toString());
	}
	
	/**
	  * Test the detach method with a persistent-new inherited object: the object is flushed before being detached. 
	  */
	public void testDetachPersistentNewInherited() {
		logger.log(BasicLevel.DEBUG, "******************testDetachPersistentNewInherited****************");
		FormulaOne f = new FormulaOne("renault", 4, "yellow", 260);
		
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		pm.makePersistent(f);
		//detach the formula one while it is persistent-new: the object must be flushed before being detached
		FormulaOne copyOfF = (FormulaOne) ((ProxyManager)pm).detachCopy((Object)f);
		pm.currentTransaction().commit();
		pm.close();
		//	print the formula one out
		logger.log(BasicLevel.DEBUG, copyOfF.toString());
	}	
	
	/*----------------- PRIVATE METHODS----------------------------*/
	
	/**
	 * Delete all the entries from the table tableName
	 */
	private void deleteQuery(String tableName){
		try{
			 PMapper mapper =  getMapper(getPMF());
			Connection connection = (Connection) mapper.getConnection();
			connection.createStatement().execute("DELETE FROM "+ tableName);
			mapper.closeConnection(connection);
		}
		catch (Exception e){
			logger.log(BasicLevel.DEBUG, e.toString());
		}
	}
	
	private void deleteDatabase(){
		deleteQuery("SE_COACH");
		deleteQuery("SE_TEAM");
		deleteQuery("SE_PLAYER");
		deleteQuery("SE_VEHICLE");
	}
}
