/*
 * Speedo: an implementation of JDO compliant personality on top of JORM
 * generic I/O sub-system. Copyright (C) 2001-2004 France Telecom R&D
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Release: 1.0
 * 
 * Created on Apr 8, 2004 @author fmillevi@yahoo.com
 *  
 */
package org.objectweb.speedo.j2eedo.bo;

import java.util.ArrayList;
import java.util.Vector;

import javax.transaction.Synchronization;

/**
 * The J2eedo application maintains three static lists storing the primary key
 * of existing
 * {@link org.objectweb.speedo.j2eedo.database.Department departments},
 * {@link org.objectweb.speedo.j2eedo.database.Employee employees}and
 * {@link org.objectweb.speedo.j2eedo.database.Project projects}. This class
 * <ul>
 * <li>Each time a new instance of this object is created the factory class
 * has to add the id in the dedicated poll</li>
 * <li>Each time an existing object is deleted, its id have to be remove from
 * the poll</li>
 * </ul>
 * <p>
 * This class allows the <b>synchronization</b> the <b>static pools</b> id
 * content update and the <b>validation of the transaction</b>.
 * 
 * @author fmillevi@yahoo.com
 * @see DatabaseImpl
 */
public class PollsSynchronizations implements Synchronization {
	private ArrayList pools;
	private ArrayList actionADDLists;
	private ArrayList actionREMLists;
	private boolean processed;

	/**
	 * Do nothing
	 * 
	 * @see javax.transaction.Synchronization#beforeCompletion()
	 */
	public void beforeCompletion() {
		// do nothing
	}

	/**
	 * The afterCompletion method has to update the pool content according the
	 * action list.
	 * 
	 * @see javax.transaction.Synchronization#afterCompletion(int)
	 */
	public void afterCompletion(int arg0) {
		// allready processed
		if (this.processed) 
			return;
		this.processed = true;
		// no pool defined
		if (0 == this.pools.size())
			return;
		for (int i = 0; i < this.pools.size(); i++) {
			// check if some action must be done
			if (0 != ((Vector) this.actionADDLists.get(i)).size()
				|| 0 != ((Vector) this.actionREMLists.get(i)).size()) {
				for (int j = 0;
					j < ((Vector) this.actionADDLists.get(i)).size();
					j++) {
					((Vector) this.pools.get(i)).add(
						((Vector) this.actionADDLists.get(i)).get(j));
				}
				for (int j = 0;
					j < ((Vector) this.actionREMLists.get(i)).size();
					j++) {
					((Vector) this.pools.get(i)).remove(
						((Vector) this.actionREMLists.get(i)).get(j));
				}
			}
		}
	}

	/**
	 * Remember the new id created
	 * 
	 * @param id
	 */
	public void addInPool(Vector pool, long id) {
		int i = 0;
		// search if the pool is in the current list
		for (i = 0; i < this.pools.size(); i++) {
			if (pool == (Vector) (this.pools.get(i)))
				break;
		}
		// the pool does not exist yet
		if (i == this.pools.size()) {
			this.pools.add(pool);
			this.actionADDLists.add(new Vector());
			this.actionREMLists.add(new Vector());
		}

		if (!((Vector) this.actionADDLists.get(i)).contains(new Long(id)))
			 ((Vector) this.actionADDLists.get(i)).add(new Long(id));
	}

	/**
	 * Remember the new id deleted
	 * 
	 * @param id
	 */
	public void removeFromPool(Vector pool, long id) {
		int i = -1;
		// search if the pool is in the current list
		for (i = 0; i < this.pools.size(); i++) {
			if (pool == (Vector) (this.pools.get(i)))
				break;
		}
		// the pool does not exist yet
		if (i == this.pools.size()) {
			this.pools.add(pool);
			this.actionADDLists.add(new Vector());
			this.actionREMLists.add(new Vector());
		}

		if (!((Vector) this.actionREMLists.get(i)).contains(new Long(id)))
			 ((Vector) this.actionREMLists.get(i)).add(new Long(id));
	}

	/**
	 * The constructor initialize the list of add and remove actions to be
	 * treated while performing the {@link #afterCompletion afterCompletion}
	 * 
	 * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfDepartmentId
	 * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfEmployeeId
	 * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfProjectId
	 */
	public PollsSynchronizations() {
		super();
		this.pools = new ArrayList();
		this.actionADDLists = new ArrayList();
		this.actionREMLists = new ArrayList();
		this.processed = false;
	}
}
