/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

import junit.framework.Assert;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.pobjects.userid.LinkedIntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Stresses the reference iteration of Speedo.
 *
 * @author M. Guillemin
 */
public class TestNavigation extends LinkedIntUserIdHelper {

	public TestNavigation(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestNavigation";
	}

	protected void perform(Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		LIUICtx nctx = (LIUICtx) ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			// Execute one transaction: Iterate nbRead from first
			int first = Alea.rand(0, nctx.dbSize - 1);
			int nbRead = Alea.rand(1, 1000);
			int name = -1;
			if (debug) {
				logger.log(BasicLevel.DEBUG, "first= " + first);
				logger.log(BasicLevel.DEBUG, "nbRead= " + nbRead);
			}
			LinkedIntUserId current = (LinkedIntUserId) pm.getObjectById(nctx.oids[first], false);
			for (int no = 0; no < nbRead; no++) {
				Assert.assertNotNull("first=" + first + " nbRead=" + nbRead, current);
				name = current.getName();
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Obj " + name);
				}
				current = current.getNext();
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	private void performNavigation(int nbThread, int dbsize, int nbTx, int to) {
		perform(nbThread, nbTx, to, new LIUICtx(dbsize));
	}


	/**
	 * Tests the reference iteration of a lot of persistent objects, with
	 * interactive setting of test parameteres (see file
	 * userconf/project.properties).
	 */
	public void testNavigation() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(),
			        new LIUICtx(
			            Integer.getInteger(DBSIZE, 100000).intValue()));
		}
	}

	/**
	 * Tests 100 transactions where each of them iterates nbread from first using 1 thread.
	 */
	public void testNavigationTh1Tx100De1() {
		if (!interactive) {
			performNavigation(1, 100, 100, 1000000);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them iterates nbread from first using 1 thread.
	 */

	public void testNavigationTh1Tx1000De1() {
		if (!interactive) {
			performNavigation(1, 1000, 1000, 1000000);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them iterates nbread from first using 1 thread.
	 * thread.
	 */

	public void testNavigationTh1Tx10000De1() {
		if (!interactive) {
			performNavigation(1, 10000, 10000, 1000000);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them iterates nbread from first using 1 thread.
	 */

	public void testNavigationTh1Tx100000De1() {
		if (!interactive) {
			performNavigation(1, 100000, 100000, 1000000);
		}
	}

	/**
	 * Tests 100 transactions where each of them iterates nbread from first using 10 thread.
	 */
	public void testNavigationTh10Tx100De1() {
		if (!interactive) {
			performNavigation(10, 100, 100, 1000000);
		}
	}

	/**
	 * Tests 1.000 transactions where each of them iterates nbread from first using 10 thread.
	 */

	public void testNavigationTh10Tx1000De1() {
		if (!interactive) {
			performNavigation(10, 1000, 1000, 1000000);
		}
	}

	/**
	 * Tests 10.000 transactions where each of them iterates nbread from first using 10 threads.
	 * thread.
	 */

	public void testNavigationTh10Tx10000De1() {
		if (!interactive) {
			performNavigation(10, 10000, 10000, 1000000);
		}
	}

	/**
	 * Tests 100.000 transactions where each of them iterates nbread from first using 10 threads.
	 */

	public void testNavigationTh10Tx100000De1() {
		if (!interactive) {
			performNavigation(10, 100000, 100000, 1000000);
		}
	}
}
