/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.pobjects.userid.LinkedIntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.Extent;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;
import javax.jdo.Query;

import java.util.Collection;
import java.util.Iterator;

import junit.framework.TestSuite;
import junit.textui.TestRunner;
import junit.framework.Assert;

/**
 * Stresses the query execution of Speedo.
 *
 * @author M. Guillemin
 */
public class TestQuery extends LinkedIntUserIdHelper {
	public static int RESULT_MAX = 1000;

	public TestQuery(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestQuery";
	}

	public static void main(String[] args) {
		TestRunner.run(new TestSuite(TestQuery.class));
	}

	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object _ctx,
	                       StressHelper.PerformResult res) {
		LIUICtx ctx = (LIUICtx) _ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			int min = Alea.rand(0, ctx.dbSize - RESULT_MAX);
			if (min > ctx.dbSize - 1) {
				min = 0;
			}
			int max = Alea.rand(min + 1, min + RESULT_MAX);
			if (max > ctx.dbSize - 1) {
				max = ctx.dbSize - 1;
			}
			if (debug) {
				logger.log(BasicLevel.DEBUG, "Thread " + threadId
				        + " min = " + min + " / max = " + max);
			}
			Extent liui = pm.getExtent(LinkedIntUserId.class, false);
			String filter = "((name>min) && (name<max))";
			Query query = pm.newQuery(liui, filter);
			query.declareParameters("Integer min, Integer max");
			Collection cLinkedUsers = (Collection) query.execute(new Integer(min), new Integer(max));
			Iterator iterator = cLinkedUsers.iterator();
			while (iterator.hasNext()) {
				LinkedIntUserId lLinkedIntUserId = (LinkedIntUserId) iterator.next();
				int i = lLinkedIntUserId.getName();
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Thread " + threadId
					        + " Name = " + i);
				}
			}
			Assert.assertEquals("Bad query result size", max - min - 1, cLinkedUsers.size());
			query.closeAll();
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	protected void performQuery(int nbThread,
	                            int dbSize,
	                            int nbTx,
	                            int threadTimeout) {
		perform(nbThread, nbTx, threadTimeout, new LIUICtx(dbSize));
	}


	/**
	 * Tests the creation of a lot of persistent objects, with interactive
	 * setting of test parameteres (see file userconf/project.properties).
	 */
	public void testQuery() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(),
			        new LIUICtx(
			                Integer.getInteger(DBSIZE, 100000).intValue()));
		}
	}

	/**
	 * Tests 100 transactions where each of them execute 1 query on 100 objects using 1 thread.
	 */
	public void testQueryTh1Tx100Qu1() {
		if (!interactive) {
			performQuery(1, 10000, 100, 1000000);
		}
	}

	/**
	 * Tests 1000 transactions where each of them execute 1 query on 1000 objects using 1 thread.
	 */

	public void testQueryTh1Tx1000Qu1() {
		if (!interactive) {
			performQuery(1, 10000, 1000, 1000000);
		}
	}

	/**
	 * Tests 10000 transactions where each of them execute 1 query on 10000 objects using 1 thread.
	 */

	public void testQueryTh1Tx10000Qu1() {
		if (!interactive) {
			performQuery(1, 10000, 10000, 1000000);
		}
	}

	/**
	 * Tests 100000 transactions where each of them execute 1 query on 100000 objects using 1 thread.
	 */

	public void testQueryTh1Tx100000Qu1() {
		if (!interactive) {
			performQuery(1, 10000, 100000, 1000000);
		}
	}

	/**
	 * Tests 100 transactions where each of them execute 1 query on 100 objects using 10 thread.
	 */
	public void testQueryTh10Tx100Qu1() {
		if (!interactive) {
			performQuery(10, 10000, 100, 1000000);
		}
	}

	/**
	 * Tests 1000 transactions where each of them execute 1 query on 1000 objects using 10 thread.
	 */

	public void testQueryTh10Tx1000Qu1() {
		if (!interactive) {
			performQuery(10, 1000, 1000, 1000000);
		}
	}

	/**
	 * Tests 10000 transactions where each of them execute 1 query on 10000 objects using 10 thread.
	 */

	public void testQueryTh10Tx10000Qu1() {
		if (!interactive) {
			performQuery(10, 10000, 10000, 1000000);
		}
	}

	/**
	 * Tests 100000 transactions where each of them execute 1 query on 100000 objects using 10 thread.
	 */

	public void testQueryTh10Tx100000Qu1() {
		if (!interactive) {
			performQuery(10, 10000, 100000, 1000000);
		}
	}
}
