/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

import junit.framework.TestSuite;
import junit.textui.TestRunner;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.pobjects.basic.BasicA;
import org.objectweb.speedo.pobjects.collection.AMMB;
import org.objectweb.speedo.pobjects.collection.BMMB;
import org.objectweb.speedo.pobjects.collection.Ref2Ref2AMMB;
import org.objectweb.speedo.pobjects.ref.Department;
import org.objectweb.speedo.pobjects.ref.Employee;
import org.objectweb.speedo.pobjects.userid.AutoIncFieldId;
import org.objectweb.speedo.pobjects.userid.Ref2AutoIncFieldId;
import org.objectweb.speedo.runtime.query.TestQueries;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Stresses the query execution of Speedo.
 *
 * @author M. Guillemin
 */
public class TestQueryConcurrency extends QueryHelper {
	
	public static int RESULT_MAX = 1000;
	
	protected String QUERYSUBSET = getLoggerName() + ".querysubset";
	
	private static TestQueries test = new TestQueries("TestQueryConcurrency");
	
	public TestQueryConcurrency(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{Employee.class.getName(),
							Department.class.getName(),
							AMMB.class.getName(),
							BMMB.class.getName(),
							Ref2Ref2AMMB.class.getName(),
							Ref2AutoIncFieldId.class.getName(),
							AutoIncFieldId.class.getName(),
							BasicA.class.getName()};
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestQueryConcurrency";
	}

	public static void main(String[] args) {
		TestRunner.run(new TestSuite(TestQueryConcurrency.class));
	}

	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object _ctx,
	                       StressHelper.PerformResult res) {
		QueryCtx queryCtx = (QueryCtx) _ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			
			int query = queryCtx.queries[Alea.rand(0, queryCtx.queries.length-1)];
			//for (query=0; query<35; query++)
			launchTest(query, threadId);

			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	/**
	 * Tests the execution of different queries, with interactive
	 * setting of test parameteres (see file userconf/stress.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 10).intValue(),
			        Integer.getInteger(TX, 100).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(),
					new QueryCtx(System.getProperty(QUERYSUBSET)));
		}
	}
	
	/**
	 * Switch method
	 */
	private void performQuery(
			int nbThread,			
			int nbTx,
			int threadTimeout) {
		perform(nbThread, nbTx, threadTimeout, new QueryCtx(System.getProperty(QUERYSUBSET)));
	}
	
	/**
	 * Tests 100 transactions where each of them execute 1 query on 100 objects using 1 thread.
	 */
	public void testQueryConcurrencyTh1Tx100Qu1() {
		if (!interactive) {
			performQuery(1, 100, 1000000);
		}
	}

	/**
	 * Tests 1000 transactions where each of them execute 1 query on 1000 objects using 1 thread.
	 */

	public void testQueryConcurrencyTh1Tx1000Qu1() {
		if (!interactive) {
			performQuery(1, 1000, 1000000);
		}
	}

	/**
	 * Tests 10000 transactions where each of them execute 1 query on 10000 objects using 1 thread.
	 */

	public void testQueryConcurrencyTh1Tx10000Qu1() {
		if (!interactive) {
			performQuery(1, 10000, 1000000);
		}
	}

	/**
	 * Tests 100000 transactions where each of them execute 1 query on 100000 objects using 1 thread.
	 */

	public void testQueryConcurrencyTh1Tx100000Qu1() {
		if (!interactive) {
			performQuery(1, 100000, 1000000);
		}
	}

	/**
	 * Tests 100 transactions where each of them execute 1 query on 100 objects using 10 thread.
	 */
	public void testQueryConcurrencyTh10Tx100Qu1() {
		if (!interactive) {
			performQuery(10, 100, 1000000);
		}
	}

	/**
	 * Tests 1000 transactions where each of them execute 1 query on 1000 objects using 10 thread.
	 */

	public void testQueryConcurrencyTh10Tx1000Qu1() {
		if (!interactive) {
			performQuery(10, 1000, 1000000);
		}
	}

	/**
	 * Tests 10000 transactions where each of them execute 1 query on 10000 objects using 10 thread.
	 */

	public void testQueryConcurrencyTh10Tx10000Qu1() {
		if (!interactive) {
			performQuery(10, 10000, 1000000);
		}
	}

	/**
	 * Tests 100000 transactions where each of them execute 1 query on 100000 objects using 10 thread.
	 */

	public void testQueryConcurrencyTh10Tx100000Qu1() {
		if (!interactive) {
			performQuery(10, 100000, 1000000);
		}
	}


	private void launchTest(int queryidx, int threadId) {
		String msg = "Thread " + threadId + " method: ";
		try {
			switch(queryidx) {
			case 0:
				msg += "test0Parameter";
				logger.log(BasicLevel.DEBUG, msg);
				test.test0Parameter();
				break;
			case 1:
				msg += "test10rder";
				logger.log(BasicLevel.DEBUG, msg);
				test.test10rder();
				break;
			case 2:
				msg += "test20rder";
				logger.log(BasicLevel.DEBUG, msg);
				test.test20rder();
				break;
			case 3:
				msg += "test1Parameter";
				logger.log(BasicLevel.DEBUG, msg);
				test.test1Parameter();
				break;
			case 4:
				msg += "testFieldParameter";
				logger.log(BasicLevel.DEBUG, msg);
				test.testFieldParameter();
				break;
			case 5:
				msg += "test2Parameters";
				logger.log(BasicLevel.DEBUG, msg);
				test.test2Parameters();
				break;
			case 6:
				msg += "test3Parameters";
				logger.log(BasicLevel.DEBUG, msg);
				test.test3Parameters();
				break;
			case 7:
				msg += "testMapParameters";
				logger.log(BasicLevel.DEBUG, msg);
				test.testMapParameters();
				break;
			case 8:
				msg += "testArrayParameters";
				logger.log(BasicLevel.DEBUG, msg);
				test.testArrayParameters();
				break;
			case 9:
				msg += "testFieldRef";
				logger.log(BasicLevel.DEBUG, msg);
				test.testFieldRef();
				break;
			case 10:
				msg += "testThis";
				logger.log(BasicLevel.DEBUG, msg);
				test.testThis();
				break;
			case 11:
				msg += "testParamRef";
				logger.log(BasicLevel.DEBUG, msg);
				test.testParamRef();
				break;
			case 12:
				msg += "testBsContainsParamA0B0";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsParamA0B0();
				break;
			case 13:
				msg += "testBsContainsParamA0B2";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsParamA0B2();
				break;
			case 14:
				msg += "testBsContainsVarB3EmptyShortPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB3EmptyShortPath();
				break;
			case 15:
				msg += "testBsContainsVarB0ShortPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB0ShortPath();
				break;
			case 16:
				msg += "testBsContainsVarB1ShortPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB1ShortPath();
				break;
			case 17:
				msg += "testBsContainsVarB3EmptyLongPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB3EmptyLongPath();
				break;
			case 18:
				msg += "testBsContainsVarB0LongPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB0LongPath();
				break;
			case 19:
				msg += "testBsContainsVarB1LongPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB1LongPath();
				break;
			case 20:
				msg += "testBsIsEmptyA0";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsEmptyA0();
				break;
			case 21:
				msg += "testBsIsEmptyA3";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsEmptyA3();
				break;
			case 22:
				msg += "testBsIsEmptyA0WithNot";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsEmptyA0WithNot();
				break;
			case 23:
				msg += "testBsIsEmptyA3WithNot";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsEmptyA3WithNot();
				break;
			case 24:
				msg += "testBsContainsVarB3EmptyLongPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB3EmptyLongPath();
				break;
			case 25:
				msg += "testBsContainsVarB3EmptyLongPath";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsContainsVarB3EmptyLongPath();
				break;
			case 26:
				msg += "testBsIsNotEmptyA0Global";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsNotEmptyA0Global();
				break;
			case 27:
				msg += "testBsIsNotEmptyA0Local";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsNotEmptyA0Local();
				break;
			case 28:
				msg += "testBsIsNotEmptyA3Global";
				logger.log(BasicLevel.DEBUG, msg);
				test.testBsIsNotEmptyA3Global();
				break;
			case 29:
				msg += "testNotA0";
				logger.log(BasicLevel.DEBUG, msg);
				test.testNotA0();
				break;
			case 30:
				msg += "testExtentAMMBfalse";
				logger.log(BasicLevel.DEBUG, msg);
				test.testExtentAMMBfalse();
				break;
			case 31:
				msg += "testExtentBMMBfalse";
				logger.log(BasicLevel.DEBUG, msg);
				test.testExtentBMMBfalse();
				break;
			case 32:
				msg += "testQueryBasedOnExtent";
				logger.log(BasicLevel.DEBUG, msg);
				test.testQueryBasedOnExtent();
				break;
			case 33:
				msg += "testSequenceIdNavigateToPrimitive";
				logger.log(BasicLevel.DEBUG, msg);
				test.testSequenceIdNavigateToPrimitive();
				break;
			case 34:
				msg += "testSequenceIdNavigateToPrimitive2";
				logger.log(BasicLevel.DEBUG, msg);
				test.testSequenceIdNavigateToPrimitive2();
				break;
			}
		} catch (Exception e) {
			Exception ie = ExceptionHelper.getNested(e);
			logger.log(BasicLevel.ERROR, msg, ie);
			fail(msg + ie.getMessage());
		}
	}
	

	
}
