/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import org.objectweb.speedo.pobjects.relations.C;
import org.objectweb.speedo.pobjects.relations.D;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

/**
 * Stress the relation one-many implementation by Speedo.
 *
 * @author M. Guillemin
 */
public class TestRelation1_NConcurrency extends CDRelationHelper {

	public TestRelation1_NConcurrency(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{
			org.objectweb.speedo.pobjects.relations.C.class.getName(),
			org.objectweb.speedo.pobjects.relations.D.class.getName()
		};
	}
	
	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestRelation1_NConcurrency";
	}

	protected int getNbSucc() {
		return 4; // because d1, d2, d3, d4
	}
	
	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       StressHelper.PerformResult res) {
		CDCtx cdctx = (CDCtx) ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);

			int oid = getOid(cdctx, (task.txToExecute.length==0), getNbSucc());			
			
			C c1 = (C) pm.getObjectById(cdctx.oidsC[oid], false);
			C c2 = (C) pm.getObjectById(cdctx.oidsC[oid+1], false);
			D d1 = (D) pm.getObjectById(cdctx.oidsD[oid], false);
			D d2 = (D) pm.getObjectById(cdctx.oidsD[oid+1], false);
			D d3 = (D) pm.getObjectById(cdctx.oidsD[oid+2], false);
			D d4 = (D) pm.getObjectById(cdctx.oidsD[oid+3], false);
			
			c1.getNds().add(d1);
			assertTrue(d1.getNc() == c1);
			c1.getNds().add(d2);
			assertTrue(d2.getNc() == c1);

			c2.getNds().add(d3);
			assertTrue(d3.getNc() == c2);
			c2.getNds().add(d4);
			assertTrue(d4.getNc() == c2);

			c1.getNds().add(d3);
			assertTrue(d3.getNc() == c1);
			assertTrue(!c2.getNds().contains(d3));

			c1.getNds().remove(d1);
			assertTrue(d1.getNc() == null);

			d3.setNc(c2);
			assertTrue(!c1.getNds().contains(d3));
			assertTrue(c2.getNds().contains(d3));

			c1.setNds(c2.getNds());
			
			commitTx(pm, task, threadId, txId);
			
			assertTrue(!c1.getNds().contains(d1));
			assertTrue(!c1.getNds().contains(d2));
			assertTrue(c1.getNds().contains(d3));
			assertTrue(c1.getNds().contains(d4));
			assertTrue(c1.getNds().size() == 2);

			assertTrue(!c2.getNds().contains(d1));
			assertTrue(!c2.getNds().contains(d2));
			assertTrue(!c2.getNds().contains(d3));
			assertTrue(!c2.getNds().contains(d4));

			assertTrue(d1.getNc() == null);
			assertTrue(d2.getNc() == null);
			assertTrue(d3.getNc() == c1);
			assertTrue(d4.getNc() == c1);
			
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}


	/**
	 * Tests the read of a lot of persistent objects, with interactive setting
	 * of test parameteres (see file userconf/project.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			logger.log(BasicLevel.INFO, "testInteractive");
			perform(Integer.getInteger(THREAD, 1).intValue(),
			        Integer.getInteger(TX, 1).intValue(),
			        Integer.getInteger(TIMEOUT, 1).intValue(), 
					new CDCtx(Integer.getInteger(DBSIZE, 1000).intValue()));
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testInteractive");
		}				
	}

	/**
	 * Methode d'aiguillage
	 */
	private void perform(
			int nbThread,
			int dbSize,
			int nbTx,
			int threadTimeout) {
		perform(nbThread, nbTx, threadTimeout, new CDCtx(dbSize));
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_NConcurrencyTh1Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh1Tx100");
			perform(1, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh1Tx100");
		}				
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_NConcurrencyTh1Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh1Tx1000");
			perform(1, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh1Tx1000");
		}				
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_NConcurrencyTh1Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh1Tx10000");
			perform(1, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh1Tx10000");
		}				
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_NConcurrencyTh2Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh2Tx100");
			perform(2, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh2Tx100");
		}				
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_NConcurrencyTh2Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh2Tx1000");
			perform(2, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh2Tx1000");
		}		
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_NConcurrencyTh2Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh2Tx10000");
			perform(2, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh2Tx10000");
		}		
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 10 thread.
	 */
	public void testRelation1_NConcurrencyTh10Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh10Tx100");
			perform(10, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh10Tx100");
		}		
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 10 thread.
	 */
	public void testRelation1_NConcurrencyTh10Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh10Tx1000");
			perform(10, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh10Tx1000");
		}						
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_NConcurrencyTh10Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_NConcurrencyTh10Tx10000");
			perform(10, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testRelation1_NConcurrencyTh10Tx10000");
		}				
	}
}
