/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

import junit.framework.Assert;

import org.objectweb.speedo.Alea;
import org.objectweb.speedo.pobjects.userid.LinkedIntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Stresses the write access function of Speedo.
 * @author M. Guillemin
 */
public class TestSetObject extends LinkedIntUserIdHelper {

	private String NBWRITE = getLoggerName() + ".nbwrite";
	private String NBWRITEPERSISTENT = getLoggerName() + ".nbwritepersistent";

	public TestSetObject(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestSetObject";
	}

	class SetCtx extends LIUICtx{
		public int nbWrite;
		public int nbWritePersistent;

		public SetCtx(int dbSize, int nbWrite, int nbWritePersistent) {
			super(dbSize);
			this.nbWrite = nbWrite;
			this.nbWritePersistent = nbWritePersistent;
		}
	}

	protected void perform(Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		SetCtx sctx = (SetCtx) ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			// Execute one transaction
			// Do nbWrite access to a persistent
			LinkedIntUserId oUserId;
			int oid;
			String f1 = Alea.randomstring();
			for (int no = 0; no < sctx.nbWrite; no++) {
				oid = Alea.rand(0, sctx.dbSize - 1);
				oUserId = (LinkedIntUserId) pm.getObjectById(sctx.oids[oid], false);
				if (debug) {
					logger.log(BasicLevel.DEBUG, "Writing object " + oid);
				}
				oUserId.writeF1(sctx.nbWritePersistent, f1);
				Assert.assertEquals("Write error", f1, oUserId.getF1());
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	private void perform(int nbThread,
	                     int dbSize,
	                     int nbTx,
	                     int nbWrite,
	                     int nbWritePersistent,
	                     int threadTimeout) {
		if (!interactive) {
			perform(nbThread, nbTx, threadTimeout,
			        new SetCtx(dbSize, nbWrite, nbWritePersistent));
		}
	}

	/**
	 * Tests the read of a lot of persistent objects, with interactive
	 * setting of test parameteres (see file userconf/project.properties).
	 */
	public void testInteractive() {
		if (interactive) {
			logger.log(BasicLevel.INFO, "testInteractive");
			perform(Integer.getInteger(THREAD, 0).intValue(),
			        Integer.getInteger(TX, 0).intValue(),
			        Integer.getInteger(TIMEOUT, 120000).intValue(),
			        new SetCtx(
			                Integer.getInteger(DBSIZE, 10000).intValue(),
			                Integer.getInteger(NBWRITE, 0).intValue(),
			                Integer.getInteger(NBWRITEPERSISTENT, 0).intValue()));
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testInteractive");
		}
	}

	/**
	 * Tests 100 transactions where each of them do 5 write access on 10 objects among 100 using 1 thread.
	 */
	public void testSetObjectTh1Tx100nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh1Tx100nbW5");
			perform(1, 10000, 100, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh1Tx100nbW5");
		}
	}

	/**
	 * Tests 1000 transactions where each of them do 5 write access on 10 objects among 1000 using 1 thread.
	 */
	public void testSetObjectTh1Tx1000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh1Tx1000nbW5");
			perform(1, 10000, 1000, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh1Tx1000nbW5");
		}
	}

	/**
	 * Tests 10.000 transactions where each of them do 5 write access on 10 objects among 10.000 using 1 thread.
	 */
	public void testSetObjectTh1Tx10000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh1Tx10000nbW5");
			perform(1, 10000, 10000, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh1Tx10000nbW5");
		}
	}

	/**
	 * Tests 100.000 transactions where each of them do 5 write access on 10 objects among 100.000 using 1 thread.
	 */
	public void _testSetObjectTh1Tx100000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "_testSetObjectTh1Tx100000nbW5");
			perform(1, 10000, 100000, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore _testSetObjectTh1Tx100000nbW5");
		}
	}

	/**
	 * Tests 100 transactions where each of them do 5 write access on 10 objects among 100 using 100 thread.
	 */
	public void testSetObjectTh100Tx100nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh100Tx100nbW5");
			perform(100, 10000, 100, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh100Tx100nbW5");
		}
	}

	/**
	 * Tests 1000 transactions where each of them do 5 write access on 100 objects among 1000 using 100 thread.
	 */
	public void testSetObjectTh100Tx1000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh100Tx1000nbW5");
			perform(100, 10000, 1000, 100, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh100Tx1000nbW5");
		}
	}

	/**
	 * Tests 10.000 transactions where each of them do 5 write access on 1000 objects among 10.000 using 100 thread.
	 */
	public void testSetObjectTh100Tx10000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh100Tx10000nbW5");
			perform(100, 10000, 10000, 1000, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh100Tx10000nbW5");
		}
	}

	/**
	 * Tests 100.000 transactions where each of them do 5 write access on 10000 objects among 100.000 using 100 thread.
	 */
	public void _testSetObjectTh100Tx100000nbW5() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "_testSetObjectTh100Tx100000nbW5");
			perform(100, 10000, 100000, 10000, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore _testSetObjectTh100Tx100000nbW5");
		}
	}

	/**
	 * Tests 100 transactions where each of them do 10 write access on 10 objects among 100 using 5 thread.
	 */
	public void testSetObjectTh5Tx100nbW10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh5Tx100nbW10");
			perform(5, 10000, 100, 10, 10, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh5Tx100nbW10");
		}
	}

	/**
	 * Tests 1000 transactions where each of them do 10 write access on 100 objects among 1000 using 5 thread.
	 */
	public void testSetObjectTh5Tx1000nbW10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh5Tx1000nbW10");
			perform(5, 10000, 1000, 100, 10, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh5Tx1000nbW10");
		}
	}

	/**
	 * Tests 10.000 transactions where each of them do 5 write access on 10 objects among 10.000 using 5 thread.
	 */
	public void testSetObjectTh5Tx10000nbW10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testSetObjectTh5Tx10000nbW10");
			perform(5, 10000, 10000, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore testSetObjectTh5Tx10000nbW10");
		}
	}

	/**
	 * Tests 100.000 transactions where each of them do 5 write access on 10 objects among 100.000 using 5 thread.
	 */
	public void _testSetObjectTh5Tx100000nbW10() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "_testSetObjectTh5Tx100000nbW10");
			perform(5, 10000, 100000, 10, 5, 120000);
		} else {
			logger.log(BasicLevel.INFO, "Interactive mode: ignore _testSetObjectTh5Tx100000nbW10");
		}
	}
}
